  #include <stdio.h>
  #include <stdlib.h>
  #include <math.h>
  #include <dos.h>
  #include <malloc.h>
  #include <string.h>

  #include "g3def.h"
  #include "g3malloc.h"
  #include "g3lib.h"


  G3_SKL       *globalptrskl;
  G3_VERTEX    *globalptrvertex;

  typedef struct
    {
    SD nbuse;        /* nb use in polygons       */
    float x;            /* x normal comp            */
    float y;            /* y normal comp            */
    float z;            /* z normal comp            */
    float nx;           /* x normal comp            */
    float ny;           /* y normal comp            */
    float nz;           /* z normal comp            */
    }
  G3_3DSVERTEX;

  typedef struct
    {
    SD pta;        /* point a              */
    SD ptb;        /* point b              */
    SD ptc;        /* point c              */
    SD ab;
    SD bc;           /* segment's visibility */
    SD ca;
    char  *material;      /* used material        */
    SD error;        /* error on this poly   */
    SD used;         /* ever used ?          */
    SD coupled;      /* coupled with another */
    }
  G3_3DSTRIANGLE;

  typedef struct
    {
    SD recurse;                /* recursion level          */
    char name[32];               /* name of the chain        */
    G3_VECTOR translation;  /* translation              */
    G3_ROTATION     rotation;     /* rotation                 */
    G3_VECTOR  target;       /* target of sklitem */
    }
  G3_3DSSKLITEM;


  G3_3DSSKLITEM sklitem[128];

  float globalzoom;
  char  *globalfilename;
  int globalenv;
  int globalskl;
  int globalvtx;
  float globalxsize;
  float globalysize;
  float globalzsize;
  int   globalsingleskl;

  void object_swapvertex(G3_OBJECT *ptrobject,SD src,SD dst);
  void freadword(FILE *handle,char *ptr);
  int ffindword(FILE *handle,char *w);
  void relocate_skl_vertex(SD recurse,G3_VECTOR *parenttranslation);
  SD object_loadasc(char *filename,G3_OBJECT *ptrobject);

/**************************************************************************/
/*              EDITOR FUNCTION                                           */
/**************************************************************************/

void replacevertex(VERTEXINDEX *ptr,VERTEXINDEX src,VERTEXINDEX dst)
  {
  if(*ptr==src)
    {
    *ptr=dst;
    }
  else
    {
    if(*ptr==dst)
      {
      *ptr=src;
      }
    }
  }


void object_swapvertex(G3_OBJECT *ptrobject,VERTEXINDEX src,VERTEXINDEX dst)
  {
  G3_POLYGON *ptrpoly;
  G3_VERTEX *ptrvertex;
  G3_VERTEX tmpvertex;
  ptrpoly=ptrobject->ptrpoly;
  ptrvertex=ptrobject->ptrvertex;
  /*---------- swap vertex in vertex buffer ----------*/
  tmpvertex=ptrvertex[src];
  ptrvertex[src]=ptrvertex[dst];
  ptrvertex[dst]=tmpvertex;
  /*---------- swap vertex in polybuffer ----------*/
  while(ptrpoly->type!=G3_ENDMARK)
    {
    replacevertex(&ptrpoly->pts1,src,dst);
    replacevertex(&ptrpoly->pts2,src,dst);
    replacevertex(&ptrpoly->pts3,src,dst);
//    replacevertex(&ptrpoly->pts4,src,dst);
//    if(ptrpoly->pts5!=G3_TRIANGLEMARK)
//      {
//      replacevertex(&ptrpoly->pts5,src,dst);
//      }
    ptrpoly++;
    }
  }


void relocate_skl_vertex(SD recurse,G3_VECTOR *parenttranslation)
  {
  G3_VECTOR translation;
  /*-------- relocate vertex for each child ----------*/
  while(globalptrskl->recurse==recurse)
    {
    /*----------- relocate current skl vertex list ----------*/
    do
      {
      globalptrvertex->x -= globalptrskl->translation.x;
      globalptrvertex->y -= globalptrskl->translation.y;
      globalptrvertex->z -= globalptrskl->translation.z;
      globalptrvertex++;
      }
    while(globalptrvertex->flag>=0);
    /*----------- goto next skl ----------*/
    translation.x=globalptrskl->translation.x;
    translation.y=globalptrskl->translation.y;
    translation.z=globalptrskl->translation.z;
    globalptrskl->translation.x -= parenttranslation->x;
    globalptrskl->translation.y -= parenttranslation->y;
    globalptrskl->translation.z -= parenttranslation->z;
    //globalptrskl->translation.x <<= G3_FPSHIFT;
    //globalptrskl->translation.y <<= G3_FPSHIFT;
    //globalptrskl->translation.z <<= G3_FPSHIFT;
    globalptrskl++;
    if(globalptrskl->recurse>recurse)  relocate_skl_vertex(globalptrskl->recurse,&translation);
    }
  }


void freadword(FILE *handle,char *ptr)
  {
  char c;
  *ptr=0;
  c=(char)(fgetc(handle));
  while(!feof(handle) && (c<=' ' || c==':' || c=='\"') )
    {
    c=(char)(fgetc(handle));
    };
  if(!feof(handle))
    {
    while (!feof(handle) && (c>' ' && c!=':' && c!='\"'))
      {
      *(ptr++)=c;
      c=(char)(fgetc(handle));
      };
    *ptr='\000';
    };
  }

int ffindword(FILE *handle,char *w)
  {
  char word[256];
  freadword(handle,word);
  while(!feof(handle) && strcmp(word,w))
    {
    freadword(handle,word);
    };
  return(!strcmp(word,w));
  }



SD object_loadasc(char *filename,G3_OBJECT *ptrobject)
  {
  FILE *handle;
  char word[512];                         /* used to read a word */
  SD i,j;                     /* used for cycling */
  SD vertexcount;                      /* count of vertex in the .ASC file */
  SD polysize;                         /* size of the polybuffer */
  SD nbvertex,nbpolygon;
  SD pta,ptb,ptc;
  SD ab,bc,ca;
  SD tmp,nbfinalpoly;
  SD len;
  G3_POLYGON *ptrpoly;
  G3_VERTEX *ptrvertex;
  G3_3DSVERTEX *ptr3dsvertex;
  G3_3DSVERTEX tmp3dsvertex;
  G3_3DSTRIANGLE *ptrtriangle;
  char **vertexmaterial;
  char *tmpvertexmaterial;
  SD nvertex;
  G3_3DSSKLITEM tmpsklitem;
  SD nbsklitem;
  G3_SKL  *ptrskl;
  G3_VECTOR translation;

  float x,y,z;
  float xmax,ymax,zmax;
  float xmin,ymin,zmin;
  float xsize,ysize,zsize;


    handle=fopen(filename,"r");
    if(handle!=NULL)
    {

    /*---------------------------------------------------------------------
          ANALYSE AND STORE MESHE DATA (VERTEX & POLY)
     ----------------------------------------------------------------------*/
    /*---------- search for the vertice count ----------*/
    ffindword(handle,"Vertices");
    freadword(handle,word);
    nbvertex=atoi(word);
    printf("Nb vertex : %d\n",nbvertex);
    /*---------- alloc the object vertex buffer for storing ----------*/
    vertexcount=nbvertex;
    ptrobject->ptrvertex=(G3_VERTEX *)(malloc(sizeof(G3_VERTEX)*nbvertex+sizeof(SD)));  /* allocate one more SD to put ENDFLAG */
    if(ptrobject->ptrvertex==NULL)
      {
      G3_ERROR3(G3_TRUE,"Unable to alloc %d vertex with object %s\n",nbvertex,filename);
      fclose(handle);
      return(G3_FALSE);
      }

    /*---------- alloc the 3DS vertex buffer for computing ----------*/
    ptr3dsvertex=(G3_3DSVERTEX *)(malloc(sizeof(G3_3DSVERTEX)*nbvertex));
    if(ptr3dsvertex==NULL)
      {
      G3_ERROR3(G3_TRUE,"Unable to alloc %d 3dsvertex with object %s\n",nbvertex,filename);
      free(ptrobject->ptrvertex);
      fclose(handle);
      return(G3_FALSE);
      }
    ptrvertex=ptrobject->ptrvertex;
    /*---------- search for the polygons count ----------*/
    ffindword(handle,"Faces");
    freadword(handle,word);
    nbpolygon=atoi(word);
    printf("Nb polygon : %d\n",nbpolygon);
    /*---------- alloc a triangle buffer for storing 3DS triangles----------*/
    ptrtriangle=(G3_3DSTRIANGLE *)(malloc(sizeof(G3_3DSTRIANGLE)*nbpolygon));
    if(ptrtriangle==NULL)
      {
      G3_ERROR3(G3_TRUE,"Unable to alloc %d 3ds triangle buffer %s\n",nbvertex,filename);
      free(ptr3dsvertex);
      free(ptrobject->ptrvertex);
      fclose(handle);
      return(G3_FALSE);
      }
    /*---------- search for the vertex list beginning ----------*/
    ffindword(handle,"list");
    /*---------- read each vertex coordinates ----------*/
    for(i=0;i<nbvertex;i++)
      {
      /*---------- search for the vertex X coordinate ----------*/
      ffindword(handle,"X");
      freadword(handle,word);
      x=atof(word);
      /*---------- search for the vertex Y coordinate ----------*/
      ffindword(handle,"Y");
      freadword(handle,word);
      y=atof(word);
      /*---------- search for the vertex Z coordinate ----------*/
      ffindword(handle,"Z");
      freadword(handle,word);
      z=atof(word);
      /*---------- store x,y,z coordinates to 3DS vertex buffer ----------*/
      ptr3dsvertex[i].nbuse=0;
      ptr3dsvertex[i].x=x;
      ptr3dsvertex[i].y=z;
      ptr3dsvertex[i].z=y;
      ptr3dsvertex[i].nx=0;
      ptr3dsvertex[i].ny=0;
      ptr3dsvertex[i].nz=0;

      if(ptr3dsvertex[i].x>xmax) xmax=ptr3dsvertex[i].x;
      if(ptr3dsvertex[i].y>ymax) ymax=ptr3dsvertex[i].y;
      if(ptr3dsvertex[i].z>zmax) zmax=ptr3dsvertex[i].z;

      if(ptr3dsvertex[i].x<xmin) xmin=ptr3dsvertex[i].x;
      if(ptr3dsvertex[i].y<ymin) ymin=ptr3dsvertex[i].y;
      if(ptr3dsvertex[i].z<zmin) zmin=ptr3dsvertex[i].z;

      }

    xsize=(xmax-xmin);
    ysize=(ymax-ymin);
    zsize=(zmax-zmin);

    /*---------- store x,y,z coordinates to object vertex buffer ----------*/

    if(globalxsize!=0.0)
      {
      globalzoom=globalxsize/xsize;
      }
    if(globalysize!=0.0)
      {
      globalzoom=globalysize/ysize;
      }
    if(globalzsize!=0.0)
      {
      globalzoom=globalzsize/zsize;
      }

    for(i=0;i<nbvertex;i++)
      {
      ptrvertex[i].x=(SD)(ptr3dsvertex[i].x*globalzoom*((float)(1<<G3_FPSHIFT))+0.5);
      ptrvertex[i].y=(SD)(ptr3dsvertex[i].y*globalzoom*((float)(1<<G3_FPSHIFT))+0.5);
      ptrvertex[i].z=(SD)(ptr3dsvertex[i].z*globalzoom*((float)(1<<G3_FPSHIFT))+0.5);
      }
    /*---------- mark the end of vertex buffer ---------------*/
    ptrvertex[i].flag=G3_ENDMARK;
    /*---------- search for the poly list beginning ----------*/
    ffindword(handle,"list");
    /*---------- scan each polygon of the list ----------*/

    for(i=0;i<nbpolygon;i++)
      {
      /*---------- search for face header ----------*/
      ffindword(handle,"Face");
      /*---------- search for point A ----------*/
      ffindword(handle,"A");
      freadword(handle,word);
      pta=atoi(word);
      /*---------- search for point B ----------*/
      ffindword(handle,"B");
      freadword(handle,word);
      ptb=atoi(word);
      /*---------- search for point C ----------*/
      ffindword(handle,"C");
      freadword(handle,word);
      ptc=atoi(word);
      /*---------- search for edge AB ----------*/
      ffindword(handle,"AB");
      freadword(handle,word);
      ab=atoi(word);
      /*---------- search for edge BC ----------*/
      ffindword(handle,"BC");
      freadword(handle,word);
      bc=atoi(word);
      /*---------- search for edge CA ----------*/
      ffindword(handle,"CA");
      freadword(handle,word);
      ca=atoi(word);
      /*---------- reorder hiden edges ----------*/
      if(ab==0 || bc==0 || ca==0)   /* if one edge set to 0 then this must be the CA edge */
        {
        while(ca!=0)
          {
          tmp=pta;                  /* swap edges */
          pta=ptb;
          ptb=ptc;
          ptc=tmp;
          tmp=ab;
          ab=bc;
          bc=ca;
          ca=tmp;
          }
        }
      /*---------- store the triangle ----------*/
      ptrtriangle[i].pta=pta;                               /* store vertex used */
      ptrtriangle[i].ptb=ptb;
      ptrtriangle[i].ptc=ptc;
      ptrtriangle[i].ab=1;//ab;                                 /* store edges visibility flags */
      ptrtriangle[i].bc=1;//bc;
      ptrtriangle[i].ca=1;//ca;
      ptrtriangle[i].error=0;                               /* no error for this thiangle */
      ptrtriangle[i].used=0;                                /* this triangle is not used for the moment */
      ptrtriangle[i].coupled=0;                             /* this triangle has no 'friendly' triangle */
      /*---------- find the material ----------*/
      if(globalsingleskl)
        {
        ptrtriangle[i].material=(char *)malloc(2);          /* store the material name */
        strcpy(ptrtriangle[i].material,"A");
        }
      else
        {
        ffindword(handle,"Material");
        freadword(handle,word);
        len=strlen(word);
        ptrtriangle[i].material=(char *)malloc(len+1);     /* store the material name */
        strcpy(ptrtriangle[i].material,word);
        }
    }


    /*---------- verify 'integrity' of each triangle ----------*/
    for(i=0;i<nbpolygon;i++)
      {
      if( (ptrtriangle[i].pta==ptrtriangle[i].ptb) ||
        (ptrtriangle[i].ptb==ptrtriangle[i].ptc) ||
        (ptrtriangle[i].pta==ptrtriangle[i].ptc) )
        {
        ptrtriangle[i].error=1;                               /* if one triangle has same vertex then error */
        G3_ERROR2(G3_TRUE,"Triangle %d has same points\n",i);
        exit(-1);
        }
      if( (ptrtriangle[i].ab==1 && ptrtriangle[i].bc==1 && ptrtriangle[i].ca==1))
        {
        /* no errors, its an isolated triangle */
        }
      else
        {
        /* this is a triangle with multiple hiden edge -> force it to an isolated triangle */
        if(!(ptrtriangle[i].ab==1 && ptrtriangle[i].bc==1 && ptrtriangle[i].ca==0))
          {
          printf("Triangle %d has multiple hiden edges\n",i);
          ptrtriangle[i].ab=1;
          ptrtriangle[i].bc=1;
          ptrtriangle[i].ca=1;
          }
        }
      }

    /*---------- try to associate couple of triangle ------*/
    for(i=0;i<nbpolygon-1;i++)                                              /* for each triangle... */
      {
      if(ptrtriangle[i].coupled==0 && ptrtriangle[i].error==0)              /* if this 1st triangle has no errors an is not yet maried to another...*/
        {
        for(j=i+1;j<nbpolygon;j++)                                          /* for each other triangles...*/
          {
          if(ptrtriangle[j].coupled==0 && ptrtriangle[j].error==0)          /* if this second triangle has no errors an is not yet maried to another...*/
            {
            if( (ptrtriangle[i].ca==0) &&                                   /* if triangles can be maried... then do it */
              (ptrtriangle[j].ca==0) &&
              (ptrtriangle[i].pta==ptrtriangle[j].ptc) &&
              (ptrtriangle[i].ptc==ptrtriangle[j].pta))
              {
              ptrtriangle[i].coupled=j+1;
              ptrtriangle[j].coupled=i+1;
              j=nbpolygon;                                                  /* force loop to end (not very good)*/
              }
            }
          }
        }
      }


    /*---------- compute size of polybuffer ----------*/
    nbfinalpoly=0;
    for(i=0;i<nbpolygon;i++)
      {
      if(ptrtriangle[i].error==0)
        {
        if(ptrtriangle[i].coupled!=0)
          nbfinalpoly+=1;
        else
          nbfinalpoly+=2;
        }
      }
    nbfinalpoly /= 2;



    polysize=sizeof(G3_POLYGON)*nbfinalpoly;
    polysize+=sizeof(SD);
    printf("nbfinalpoly=%d size=%d\n",nbfinalpoly,polysize);
    /*---------- allocate object polybuffer ----------*/
    ptrobject->ptrpoly=(G3_POLYGON *)(malloc(polysize));
    if(ptrobject->ptrpoly==NULL)
      {
      G3_ERROR3(G3_TRUE,"Unable to alloc %d 3ds triangle buffer %s\n",nbvertex,filename);
      free(ptrobject->ptrvertex);
      free(ptrtriangle);
      fclose(handle);
      return(G3_FALSE);
      }

    ptrpoly=ptrobject->ptrpoly;
    /*---------- fill the object polybuffer ----------*/
    for(i=0;i<nbpolygon;i++)
      {
      if(ptrtriangle[i].error==0 && ptrtriangle[i].used==0)
        {
        j=ptrtriangle[i].coupled;
        if(j!=0)
          {
          j-=1;
          ptrpoly->type=G3_GRAPHCALL_LPOLYF;
          ptrpoly->color=((ptrtriangle[i].material[0]-'A')<<4)+(ptrtriangle[i].material[1]-'A');
          ptrpoly->color=((ptrpoly->color & 0xFF)<<8)+(ptrpoly->color &0xFF);
          ptrpoly->color=((ptrpoly->color & 0xFF)<<8)+(ptrpoly->color &0xFF);
          ptrpoly->color=((ptrpoly->color & 0xFF)<<8)+(ptrpoly->color &0xFF);
          ptrpoly->bank=0;
          ptrpoly->pts1=ptrtriangle[i].pta;
          ptrpoly->pts2=ptrtriangle[j].ptb;
          ptrpoly->pts3=ptrtriangle[i].ptc;
//          ptrpoly->pts4=ptrtriangle[i].ptb;
//          ptrpoly->pts5=ptrtriangle[i].pta;
          ptrpoly->bank=0;
          ptrpoly->xa=0;
          ptrpoly->ya=0;
          ptrpoly->xb=255<<G3_FPSHIFT;
          ptrpoly->yb=0;
          ptrpoly->xc=255<<G3_FPSHIFT;
          ptrpoly->yc=255<<G3_FPSHIFT;

          ptrtriangle[i].used=1;
          ptrtriangle[j].used=1;
          }
        else
          {
          ptrpoly->type=G3_GRAPHCALL_LPOLYF;
          ptrpoly->color=((ptrtriangle[i].material[0]-'A')<<4)+(ptrtriangle[i].material[1]-'A');
          ptrpoly->color=((ptrpoly->color & 0xFF)<<8)+(ptrpoly->color &0xFF);
          ptrpoly->color=((ptrpoly->color & 0xFF)<<8)+(ptrpoly->color &0xFF);
          ptrpoly->color=((ptrpoly->color & 0xFF)<<8)+(ptrpoly->color &0xFF);
          ptrpoly->bank=0;
          ptrpoly->pts1=ptrtriangle[i].pta;
          ptrpoly->pts2=ptrtriangle[i].ptc;
          ptrpoly->pts3=ptrtriangle[i].ptb;
//          ptrpoly->pts4=ptrtriangle[i].pta;
//          ptrpoly->pts5=G3_TRIANGLEMARK;
          ptrpoly->bank=0;
          ptrpoly->xa=0;
          ptrpoly->ya=0;
          ptrpoly->xb=255<<G3_FPSHIFT;
          ptrpoly->yb=0;
          ptrpoly->xc=255<<G3_FPSHIFT;
          ptrpoly->yc=255<<G3_FPSHIFT;
          ptrtriangle[i].used=1;
          }
        ptrpoly++;
        }
      }
    ptrpoly->type=G3_ENDMARK;

    /*---------------------------------------------------------------------
          ANALYSE AND STORE SKL DATA
     ----------------------------------------------------------------------*/
    vertexmaterial=(char **)malloc(sizeof(char **)*nbvertex);

    for(i=0;i<nbvertex;i++)
      {
      vertexmaterial[i]=NULL;
      }
    /*---------- compute each vertex group name ----------*/
    for(i=0;i<nbpolygon;i++)    /* for each polygon */
      {
      for(j=0;j<3;j++)          /* for each vertex used in polygon */
        {
        /*----------------if material not equal to "0"----------------*/
        if(strcmp(ptrtriangle[i].material,"0"))
          {
          switch(j)
            {
            case 0:
              nvertex=ptrtriangle[i].pta;
              break;
            case 1:
              nvertex=ptrtriangle[i].ptb;
              break;
            case 2:
              nvertex=ptrtriangle[i].ptc;
              break;
            }
          if(nvertex<0 || nvertex>=nbvertex)
            {
            printf("Error with nvertex\n");
            }
          if(vertexmaterial[nvertex]==NULL)       /* if vertex not named for the moment */
            {

            vertexmaterial[nvertex]=(char *)malloc(strlen(ptrtriangle[i].material)+1);

            strcpy(vertexmaterial[nvertex],ptrtriangle[i].material);

            printf("%s\n",vertexmaterial[nvertex]);
            }
          else                                    /* vertex has an assigned material */
            {
            if(strlen(ptrtriangle[i].material)<strlen(vertexmaterial[nvertex]))
              {

              free(vertexmaterial[nvertex]);

              vertexmaterial[nvertex]=(char *)malloc(strlen(ptrtriangle[i].material)+1);

              strcpy(vertexmaterial[nvertex],ptrtriangle[i].material);

              printf("%s\n",vertexmaterial[nvertex]);
              }
            }

          }
        }
      }

    /*---------- verify if each vertex owns to a group ----------*/
    for(i=0;i<nbvertex;i++)
      {
      if(vertexmaterial[i]==NULL)
        {
        printf("Warning, vertex %d not used in a group\n",i);
        vertexmaterial[i]=(char *)malloc(2);
        strcpy(vertexmaterial[i],"0");
        }
      }
    /*---------- now sort vertex by ascending owner group name------------*/

    for(i=0;i<nbvertex-1;i++)
      {
      for(j=i+1;j<nbvertex;j++)
        {
        if(strcmp(vertexmaterial[i],vertexmaterial[j])>0)
          {
          tmpvertexmaterial=vertexmaterial[i];
          vertexmaterial[i]=vertexmaterial[j];
          vertexmaterial[j]=tmpvertexmaterial;
          tmp3dsvertex=ptr3dsvertex[i];
          ptr3dsvertex[i]=ptr3dsvertex[j];
          ptr3dsvertex[j]=tmp3dsvertex;
          object_swapvertex(ptrobject,i,j);
          }
        }
      }
    /*---------- compute normals of each vertex -----------------------------*/

    ptrpoly=ptrobject->ptrpoly;
    while(ptrpoly->type!=G3_ENDMARK)
      {
      float ax,ay,az;
      float bx,by,bz;
      float nx,ny,nz;
      float norm;
      /* calculate normal of face */
      pta=ptrpoly->pts1;
      ptb=ptrpoly->pts2;
      ptc=ptrpoly->pts3;
      ax=(ptr3dsvertex[ptb].x-ptr3dsvertex[pta].x);
      ay=(ptr3dsvertex[ptb].y-ptr3dsvertex[pta].y);
      az=(ptr3dsvertex[ptb].z-ptr3dsvertex[pta].z);
      bx=(ptr3dsvertex[ptc].x-ptr3dsvertex[ptb].x);
      by=(ptr3dsvertex[ptc].y-ptr3dsvertex[ptb].y);
      bz=(ptr3dsvertex[ptc].z-ptr3dsvertex[ptb].z);
      nx=(ay*bz)-(az*by);
      ny=(az*bx)-(ax*bz);
      nz=(ax*by)-(ay*bx);
      norm=sqrt((nx*nx)+(ny*ny)+(nz*nz));
      if(norm==0)
        {
        printf("SOMETHING WRONG WHEN COMPUTING NORM : OBJECT TOO SCREETCHED\n");
        }
      nx/=norm;
      ny/=norm;
      nz/=norm;

      nvertex=ptrpoly->pts1;
      ptr3dsvertex[nvertex].nx+=nx;
      ptr3dsvertex[nvertex].ny+=ny;
      ptr3dsvertex[nvertex].nz+=nz;
      ptr3dsvertex[nvertex].nbuse+=1;

      nvertex=ptrpoly->pts2;
      ptr3dsvertex[nvertex].nx+=nx;
      ptr3dsvertex[nvertex].ny+=ny;
      ptr3dsvertex[nvertex].nz+=nz;
      ptr3dsvertex[nvertex].nbuse+=1;

      nvertex=ptrpoly->pts3;
      ptr3dsvertex[nvertex].nx+=nx;
      ptr3dsvertex[nvertex].ny+=ny;
      ptr3dsvertex[nvertex].nz+=nz;
      ptr3dsvertex[nvertex].nbuse+=1;

//      if(ptrpoly->type==G3_GRAPHCALL_LPOLYF)
//        {
//        nvertex=ptrpoly->pts4;
//        ptr3dsvertex[nvertex].nx+=nx;
//        ptr3dsvertex[nvertex].ny+=ny;
//        ptr3dsvertex[nvertex].nz+=nz;
//        ptr3dsvertex[nvertex].nbuse+=1;
//        }
      ptrpoly++;
      }


    for(i=0;i<nbvertex;i++)
      {
      float nx,ny,nz;
      float norm;
      nx=ptr3dsvertex[i].nx/ptr3dsvertex[i].nbuse;
      ny=ptr3dsvertex[i].ny/ptr3dsvertex[i].nbuse;
      nz=ptr3dsvertex[i].nz/ptr3dsvertex[i].nbuse;
      norm=sqrt((nx*nx)+(ny*ny)+(nz*nz));
      if(norm==0)
        {
        printf("SOMETHING WRONG WHEN COMPUTING NORM 2\n");
        norm=1.0;
        }
      nx/=norm;
      ny/=norm;
      nz/=norm;
      ptrvertex[i].nx=(SD)(nx*(1<<G3_FPSHIFT));
      ptrvertex[i].ny=(SD)(ny*(1<<G3_FPSHIFT));
      ptrvertex[i].nz=(SD)(nz*(1<<G3_FPSHIFT));
      }

    /*---------- put vertex group change indicator ----------*/


    tmpvertexmaterial=vertexmaterial[0];
    for(i=0;i<nbvertex;i++)
      {
        if(strcmp(tmpvertexmaterial,vertexmaterial[i]))
          {
          ptrvertex[i].flag=G3_ENDMARK-1;
          tmpvertexmaterial=vertexmaterial[i];
          }
        else
          {
          ptrvertex[i].flag=0;
          }
      }
    ptrvertex[i].flag=G3_ENDMARK;



    /*---------- now, read the skletron ------------------------------------*/
    /* each chain of skletron is supposed to point in ascending z direction */
    if(globalsingleskl)
      {
      nbsklitem=0;
      strcpy(sklitem[nbsklitem].name,"A");
      sklitem[nbsklitem].translation.x= (((xmax+xmin)/2)+0.5);
      sklitem[nbsklitem].translation.y= (((zmax+zmin)/2)+0.5);
      sklitem[nbsklitem].translation.z= (((ymax+ymin)/2)+0.5);
      sklitem[nbsklitem].target.x= (((xmax+xmin)/2)*globalzoom*((float)(1<<G3_FPSHIFT))+0.5)-sklitem[nbsklitem].translation.x;
      sklitem[nbsklitem].target.y= (((zmax+zmin)/2)*globalzoom*((float)(1<<G3_FPSHIFT))+0.5)-sklitem[nbsklitem].translation.y;
      sklitem[nbsklitem].target.z= (((ymax+ymin)/2)*globalzoom*((float)(1<<G3_FPSHIFT))+0.5)-sklitem[nbsklitem].translation.z;
      sklitem[nbsklitem].rotation.x=0;
      sklitem[nbsklitem].rotation.y=0;
      sklitem[nbsklitem].rotation.z=0;
      nbsklitem++;
      }
    else
    {
    nbsklitem=0;
    ffindword(handle,"object");
    while(!feof(handle))
      {

      sklitem[nbsklitem].recurse=0;

      /*---------- search for chain info ----------*/
      freadword(handle,word);
      strcpy(sklitem[nbsklitem].name,word);
      printf("sklitem %d %s\n",nbsklitem,sklitem[nbsklitem].name);

      /*---------- search for chain start position ----------*/
      ffindword(handle,"Position");
      ffindword(handle,"X");
      freadword(handle,word);
      x=atof(word);
      ffindword(handle,"Y");
      freadword(handle,word);
      y=atof(word);
      ffindword(handle,"Z");
      freadword(handle,word);
      z=atof(word);
      sklitem[nbsklitem].translation.x= (x*globalzoom*((float)(1<<G3_FPSHIFT))+0.5);
      sklitem[nbsklitem].translation.y= (z*globalzoom*((float)(1<<G3_FPSHIFT))+0.5);
      sklitem[nbsklitem].translation.z= (y*globalzoom*((float)(1<<G3_FPSHIFT))+0.5);

      /*---------- search for chain target position ----------*/
      ffindword(handle,"Target");
      ffindword(handle,"X");
      freadword(handle,word);
      x=atof(word);
      ffindword(handle,"Y");
      freadword(handle,word);
      y=atof(word);
      ffindword(handle,"Z");
      freadword(handle,word);
      z=atof(word);
      sklitem[nbsklitem].target.x= (x*globalzoom*((float)(1<<G3_FPSHIFT))+0.5)-sklitem[nbsklitem].translation.x;
      sklitem[nbsklitem].target.y= (z*globalzoom*((float)(1<<G3_FPSHIFT))+0.5)-sklitem[nbsklitem].translation.y;
      sklitem[nbsklitem].target.z= (y*globalzoom*((float)(1<<G3_FPSHIFT))+0.5)-sklitem[nbsklitem].translation.z;
      sklitem[nbsklitem].rotation.x= 0;
      sklitem[nbsklitem].rotation.y= 0;
      sklitem[nbsklitem].rotation.z= 0;
      nbsklitem++;
      ffindword(handle,"object");
      }
    }

    /*---------- sort sklitem by name ---------------*/


    for(i=0;i<nbsklitem-1;i++)
      {
      for(j=i+1;j<nbsklitem;j++)
        {
        if(strcmp(sklitem[i].name,sklitem[j].name)>0)
          {
          memcpy(&tmpsklitem,&sklitem[i],sizeof(G3_3DSSKLITEM));
          memcpy(&sklitem[i],&sklitem[j],sizeof(G3_3DSSKLITEM));
          memcpy(&sklitem[j],&tmpsklitem,sizeof(G3_3DSSKLITEM));
          }
        }
      }

    /*---------- compute skl item recurse level ----------*/

    printf("NBSKLITEM:%d\n",nbsklitem);
    for(i=0;i<nbsklitem;i++)
      {
      sklitem[i].recurse=strlen(sklitem[i].name);
      printf("SI %d %s %d\n",i,sklitem[i].name,sklitem[i].recurse);
      }

    /*---------- allocate object sklitem list ----------*/

    ptrobject->ptrskl=(G3_SKL *)malloc(sizeof(G3_SKL)*nbsklitem+sizeof(SD));
    ptrskl=ptrobject->ptrskl;
    for(i=0;i<nbsklitem;i++)
      {
      ptrskl[i].recurse=sklitem[i].recurse;
      ptrskl[i].translation=sklitem[i].translation;
      ptrskl[i].rotation=sklitem[i].rotation;
      ptrskl[i].target=sklitem[i].target;
      }
    ptrskl[i].recurse=G3_ENDMARK;

    /*---------- compute skl ------------------*/

    translation.x=0;
    translation.y=0;
    translation.z=0;
    globalptrskl=ptrobject->ptrskl;
    globalptrvertex=ptrobject->ptrvertex;
    relocate_skl_vertex(1,&translation);



    /*---------- recompute vertex in polybuffer to suit g3_object_envsave (it divide) ---------*/
    ptrpoly=ptrobject->ptrpoly;
    while(ptrpoly->type!=G3_ENDMARK)
      {
      ptrpoly->pts1*=sizeof(G3_DVERTEX);
      ptrpoly->pts2*=sizeof(G3_DVERTEX);
      ptrpoly->pts3*=sizeof(G3_DVERTEX);
//      ptrpoly->pts4*=sizeof(G3_DVERTEX);
//      if(ptrpoly->pts5!=G3_TRIANGLEMARK) ptrpoly->pts5*=sizeof(G3_DVERTEX);
      ptrpoly++;
      }
    /*---------- free vertex's name ----------*/
    for(i=0;i<nbvertex;i++)
      {
      if(vertexmaterial[i]!=NULL) free(vertexmaterial[i]);
      }

    /*---------- free triangle's name ----------*/
    for(i=0;i<nbpolygon;i++)
      {
      free(ptrtriangle[i].material);
      }
    /*---------- free array of triangle ----------*/

    free(ptr3dsvertex);
    free(ptrtriangle);
    fclose(handle);
    return(G3_TRUE);
    }
  else
    {
    G3_ERROR2(G3_TRUE,"Sorry : unable to open file %s\n",filename);
    return(G3_FALSE);
    };
  }


int extract_args(int argc,char **argv)
  {
  SD i;
  SD file;

  if(argc<=1)
    {
    printf("Type 'CASC -h' for help\n");
    return(G3_FALSE);
    }
  else
    {
    file=G3_FALSE;
    globalsingleskl=G3_FALSE;
    globalzoom=1.0;
    globalxsize=0.0;
    globalysize=0.0;
    globalzsize=0.0;
    globalenv=G3_TRUE;
    globalskl=G3_TRUE;
    globalvtx=G3_TRUE;
    for(i=1;i<argc;i++)
      {
      if(argv[i][0]=='-')
        {
        switch(argv[i][1])
          {
          case 'h':
            printf("CASC file[.ASC] [option]\n");
            printf("option are :\n");
            printf("-xn.n     : adjust maximum x range to n \n");
            printf("-yn.n     : adjust maximum y range to n \n");
            printf("-zn.n     : adjust maximum z range to n \n");
            printf("-e        : don't erase previous *.ENV file\n");
            printf("-s        : don't erase previous *.SKL file\n");
            printf("-v        : don't erase previous *.VTX file\n");
            printf("-c        : generate a single chain skeletron\n");
            printf("-Zn.n     : global zoom (default:1.0)\n");
            printf("-h        : this help screen\n");
            return(G3_FALSE);
          case 'Z':
            globalzoom=atof(argv[i]+2);
            if(globalzoom==0) globalzoom=1.0;
            break;
          case 'x':
            globalxsize=atof(argv[i]+2);
            break;
          case 'y':
            globalysize=atof(argv[i]+2);
            break;
          case 'z':
            globalzsize=atof(argv[i]+2);
            break;
          case 'v':
            globalvtx=G3_FALSE;
            break;
          case 'e':
            globalenv=G3_FALSE;
            break;
          case 's':
            globalskl=G3_FALSE;
            break;
          case 'c':
            globalsingleskl=G3_TRUE;
            break;
          }
        }
      else
        {
        globalfilename=argv[i];
        file=G3_TRUE;
        }
      }
    if(!file)
      {
      printf("You should specifie a file to compute...\n");
      return(G3_FALSE);
      }
    }
  return(G3_TRUE);
  }

main(int argc,char *argv[])
  {
  char drive[ _MAX_DRIVE ];
  char dir[ _MAX_DIR ];
  char fname[ _MAX_FNAME ];
  char ext[ _MAX_EXT ];
  char filename[_MAX_PATH];
  if(extract_args(argc,argv))
    {
    strcpy(filename,globalfilename);
    _splitpath(filename,drive,dir,fname,ext);
    _makepath(filename,drive,dir,fname,".ASC");

    object_loadasc(filename,&g3_object[0]);
    _splitpath(filename,drive,dir,fname,ext);
    _makepath(filename,drive,dir,fname,"");

    if(globalvtx)
      {
      printf("SAVING %s.VTX WITH %d VERTEX\n",fname,g3_object_vtxcount(&g3_object[0]));
      g3_object_vtxsave(&g3_object[0],filename);
      }
    if(globalenv)
      {
      printf("SAVING %s.ENV WITH %d POLYGONS\n",fname,g3_object_envcount(&g3_object[0]));
      g3_object_envsave(&g3_object[0],filename);
      }
    if(globalskl)
      {
      printf("SAVING %s.SKL WITH %d LINKS\n",fname,g3_object_sklcount(&g3_object[0]));
      g3_object_sklsave(&g3_object[0],filename);
      }
    return(0);
    }
  else
    {
    return(-1);
    }
  }




