
/************************************************************************/
/*              STANDARD INCLUDES                                       */
/************************************************************************/


/************************************************************************/
/*              SPECIAL  INCLUDES                                       */
/************************************************************************/

  #include "g3iff.h"

/************************************************************************/
/*              PRIVATE DEFINITIONS                                     */
/************************************************************************/

  #define G3_ID(d,c,b,a) ((UD)(a)<<24 | (UD)(b)<<16 | (UD)(c)<<8 | (UD)(d) )

  #define G3_ID_FORM   G3_ID('F','O','R','M')
  #define G3_ID_PROP   G3_ID('P','R','O','P')
  #define G3_ID_LIST   G3_ID('L','I','S','T')
  #define G3_ID_CAT    G3_ID('C','A','T',' ')
  #define G3_ID_FILLER G3_ID(' ',' ',' ',' ')
  #define G3_ID_PBM    G3_ID('P','B','M',' ')
  #define G3_ID_ILBM   G3_ID('I','L','B','M')
  #define G3_ID_BMHD   G3_ID('B','M','H','D')
  #define G3_ID_CMAP   G3_ID('C','M','A','P')
  #define G3_ID_GRAB   G3_ID('G','R','A','B')
  #define G3_ID_DEST   G3_ID('D','E','S','T')
  #define G3_ID_SPRT   G3_ID('S','P','R','T')
  #define G3_ID_CAMG   G3_ID('C','A','M','G')
  #define G3_ID_BODY   G3_ID('B','O','D','Y')

  typedef struct
    {
    UW w, h;                     /* raster width & height in pixels */
    UW x, y;                     /* position for this image */
    UB nplanes;                  /* # source bitplanes */
    UB masking;                  /* masking technique */
    UB compression;              /* compression algoithm */
    UB pad1;                     /* UNUSED.  For consistency, put 0 here.*/
    UW transparentcolor;         /* transparent "color number" */
    UB xaspect, yaspect;         /* aspect ratio, a rational number x/y */
    UW pagewidth, pageheight;    /* source "page" size in pixels */
    }
    G3_BITMAPHEADER;

  typedef struct
    {
    SD type;
    SD cksize;
    SD subtype;
    }
    G3_FORMCHUNK;

  typedef struct
    {
    SD ckid;
    SD cksize;
    }
    G3_CHUNKHEADER;

  typedef struct
    {
    SD ckid;
    SD cksize;
    UB ckdata[1];
    }
    G3_CHUNK;

/************************************************************************/
/*              PRIVATE VARIABLES                                       */
/************************************************************************/


/************************************************************************/
/*              FUNCTIONS                                               */
/************************************************************************/

UD g3_iff_byteflipdword(UD number)
   {
   UD y,t,i;
   t=number;
   y=0;for(i=0;i<4;i++){y=y|(t&0xFF);if(i<3){y=y<<8;t=t>>8;}}
   return(y);
   }

UW g3_iff_byteflipword(UW number)
   {
   UW xx1,xx2;
   UW result;
   UW t;
   t=number;
   xx1 = (t<<8);
   xx2 = (t>>8);
   result=xx1|xx2;
   return(result);
   }

UB *g3_iff_load(char *picname)
   {
   UB cplanes[8][80];
   UB *pplanes;
   G3_FILE *pic;
   G3_FORMCHUNK fchunk;
   G3_CHUNKHEADER  chunk;
   G3_BITMAPHEADER bmhd;
   SB  value;
   SD sofar;
   SD height;
   SD i,j;
   SW *wordptr;
   SB  *destx,*savedestx;
   pplanes=(UB *)&cplanes[0][0];
  if((pic=g3_fopen(picname,"rb"))==NULL)
     {
     printf("Error while opening IFF file\n");
     return((UB *)(0));
     }
   g3_fread(&fchunk,1,sizeof(fchunk),pic);
   if (fchunk.type!=G3_ID_FORM)
      {
      g3_fclose(pic);
      return((UB *)(0));
      }
   if (fchunk.subtype != G3_ID_PBM)
      {
      printf("Error: Not form PBM!");
      g3_fclose(pic);
      return((UB *)(0));
      }

   while(1)
      {
      g3_fread(&chunk,1,sizeof(G3_CHUNKHEADER),pic);
      chunk.cksize=g3_iff_byteflipdword(chunk.cksize);
      if(chunk.cksize&1) chunk.cksize++;
      if(chunk.ckid == G3_ID_BMHD)
          {
          g3_fread(&bmhd,1,sizeof(G3_BITMAPHEADER),pic);
          bmhd.w=g3_iff_byteflipword(bmhd.w);
          bmhd.h=g3_iff_byteflipword(bmhd.h);
          destx = (SB *)g3_malloc((bmhd.w*bmhd.h+sizeof(SW)*2));
          if (destx==NULL)
            {
            g3_fclose(pic);
            return((UB *)0L);
            }
          wordptr=(SW *)destx;
          *wordptr++=bmhd.w;
          *wordptr++=bmhd.h;
          savedestx = destx;
          destx=(SB *)wordptr;
          continue;
          }
      if(chunk.ckid==G3_ID_CMAP)
          {
          g3_fread(g3_colorregister,1,chunk.cksize,pic);
          for (i=0;i<256;i++)
              {
              g3_colorregister[i].red>>=2;
              g3_colorregister[i].green>>=2;
              g3_colorregister[i].blue>>=2;
              }
          continue;
          }
      if(chunk.ckid==G3_ID_BODY)
          {
          for(height = 0; height<bmhd.h; height ++)
            {
            unsigned char *dest;
            dest = (unsigned char *)&(pplanes[0]);
            sofar = bmhd.w;
            if (sofar&1) sofar++;
            while (sofar)
          {
          if (bmhd.compression)
            {
            value=fgetc(pic);
            #if 0
            // if (value == 128) continue; /* NOP..just get another*/
            #endif
            if (value>0)
                {
                j=value+1;
                sofar-=j;
                if(!(g3_fread(dest,j,1,pic)))
                  {
                  g3_fclose(pic);
                  g3_free(&savedestx);
                  return((UB *)0L);
                  }
                dest +=j;
                }
            else
                {
                i = -value;
                i ++;
                sofar -= i;
                value=fgetc(pic);
                while (--i >= 0) *dest++ = value;
                }
            }
          else
            {
            g3_fread(dest,sofar,1,pic); /* just throw on plane */
            sofar = 0;
            }
          }
          if (sofar < 0)
            {
            g3_fclose(pic);
            }
            _fmemcpy(destx,pplanes,bmhd.w);
            destx += bmhd.w;
            }
          break; /* leave if we've unpacked the BODY*/
          }
      fseek(pic,chunk.cksize,SEEK_CUR);
      }
   g3_fclose(pic);
   return((UB *)savedestx);
   }


