

/************************************************************************/
/*              RAW MANAGEMENT FUNCTION                                 */
/************************************************************************/
  #define _G3_RAW_C_


  #include "g3raw.h"

/************************************************************************/
/*              PRIVATE DEFINITIONS                                     */
/************************************************************************/

/* Format: RAW

   offset 0     : 0x6d6877616168 "mhwanh" : entete
   offset 6     : no de version (0x0004 only supported)
   offset 8     : taille en X
   offset 10    : taille en Y
   offset 12    : nb de couleur (2-255) (0 ou -24 en true color)
   offset 14    : taille Horizontale en DPI
   offset 16    : taille Vertivale en DPI
   offset 18    : Correction Gamma * 100 (2.2 => 220)
   offset 20-31 : reserver
   offset 32    : palette
   offset ???   : image (1 octet par point ou 3 en true color)
*/


#define SWAPUW(A) ((UW)((A & (UW)0x00FF)<<8)|(A>>8))

typedef struct
  {
  char id[6];
  UW noversion;
  UW w;
  UW h;
  UW nbcolor;
  UW dpiwitdh;
  UW dpiheight;
  UW gamma;
  char  reserved[12];
  }
G3_RAWHEADER;


/************************************************************************/
/*              PRIVATE VARIABLES                                       */
/************************************************************************/


/************************************************************************/
/*              FUNCTIONS VARIABLES                                     */
/************************************************************************/


PIXEL *g3_raw_load(char *filename,SD *w,SD *h)
  {
  SD size;
  G3_FILE *handle;
  SD i;
  SD nbcolor;
  G3_RAWHEADER rawheader;
  PIXEL *data;
  handle=g3_fopen(filename,"rb");
  if(handle!=NULL)
    {
    g3_fread(&rawheader,sizeof(G3_RAWHEADER),1,handle);
    rawheader.w=SWAPUW(rawheader.w);
    rawheader.h=SWAPUW(rawheader.h);
    rawheader.nbcolor=SWAPUW(rawheader.nbcolor);
    *w=(SD)rawheader.w;
    *h=(SD)rawheader.h;
    if(rawheader.nbcolor==0)
    {
    size=(*w)*(*h)*sizeof(PIXEL);
    data=(PIXEL *)g3_malloc(size);
    if(data!=NULL)
      {
      PIXEL *ptrdata;
      int i;  
      ptrdata=data;
      for(i=0;i<(*w)*(*h);i++)
      {
        g3_fread(ptrdata,sizeof(PIXEL),3,handle);
        ptrdata++;
      }
      }
    else
      {
      printf("Can't allocate data in %s",__FILE__);
      }
    }
    else
    {
    for(i=0;i<G3_MAXCOLOR;i++)
          {
          g3_fread(&g3_colorregister[i],sizeof(G3_COLORREGISTER),1,handle);
          g3_colorregister[i].red>>=2;
          g3_colorregister[i].green>>=2;
          g3_colorregister[i].blue>>=2;
          }
    size=(*w)*(*h)*sizeof(PIXEL);
    data=(PIXEL *)g3_malloc(size);
    if(data!=NULL)
      {
      g3_fread(data,size,1,handle);
      }
    else
      {
      printf("Can't allocate data in %s",__FILE__);
      }
    }
    g3_fclose(handle);
    }
  else
    {
    printf("Error while opening RAW file %s\n",filename);
    return(NULL);
    }
  return(data);
  }

void g3_raw_save(char *filename)
  {
  G3_FILE *handle;
  SD i;
  G3_RAWHEADER rawheader;
  handle=g3_fopen(filename,"wb");
  if(handle!=NULL)
    {
    memset(&rawheader,0,sizeof(G3_RAWHEADER));
    rawheader.id[0]='m';
    rawheader.id[1]='h';
    rawheader.id[2]='w';
    rawheader.id[3]='a';
    rawheader.id[4]='n';
    rawheader.id[5]='h';
    rawheader.noversion=SWAPUW(4);
    rawheader.w=SWAPUW(256);
    rawheader.h=SWAPUW(256);
    rawheader.nbcolor=SWAPUW(256);
    g3_fwrite(&rawheader,sizeof(G3_RAWHEADER),1,handle);
    {
      SD r,g,b,i;
      i=16;
      for(r=(256<<15)/12;r<(256<<15);r+=(256<<15)/6)
      for(g=(256<<15)/10;g<(256<<15);g+=(256<<15)/5)
      for(b=(256<<15)/16;b<(256<<15);b+=(256<<15)/8)
      {
        g3_colorregister[i].red=(r>>15);
        g3_colorregister[i].green=(g>>15);
        g3_colorregister[i].blue=(b>>15);
        i++;
      }
    }
    for(i=0;i<256;i++)
          {
          g3_fwrite(&g3_colorregister[i],sizeof(G3_COLORREGISTER),1,handle);
          }


    for(i=0;i<256*256;i++)
      g3_fwrite(&i,1,1,handle);


    g3_fclose(handle);
    }
  else
    {
    printf("Error while opening RAW file %s\n",filename);
    }
  }



