

/************************************************************************/
/*  VOBJECT  : 3D OBJECT VIEWER                                           */
/************************************************************************/

  #include <stdio.h>
  #include <stdlib.h>
  #include <math.h>
  #include <dos.h>
  #include <bios.h>
  #include <malloc.h>
  #include <string.h>

  #include "g3def.h"
  #include "g3malloc.h"
  #include "g3mouse.h"
  #include "g3joy.h"
  #include "g3iff.h"
  #include "g3lib.h"
  #include "g3timer.h"

/************************************************************************/
/*  MAIN                                                                */
/************************************************************************/

#define SCREENW 640
#define SCREENH 480


#define WIN_MAIN    0
#define WIN_BANK    1
#define WIN_PAL     2
#define MAXWIN      3


typedef struct
  {
  SD  x,y,w,h,refresh;
  }
WIN;

WIN win[MAXWIN]=
  {
   {1,1,SCREENW-260,480-2-10,G3_TRUE},          // main win
   {SCREENW-260+3,1,256,256,G3_TRUE},         // bank win
   {SCREENW-260+3,256+3,16*8,16*8,G3_TRUE},   // pal win
  };



G3_POLYGON    *ptrpolygon;
G3_POLYGON    *newptrpolygon;
SD         currentcolor=0;
SD         backgroundcolor=0;
SD         currentbank=0;
SD         currentwin=0;
PIXEL       *font;

SD getwindow(SD x,SD y)
{
  SD i;
  for(i=0;i<MAXWIN;i++)
  {
    if( x>=win[i].x && x<=win[i].x+win[i].w-1 &&
        y>=win[i].y && y<=win[i].y+win[i].h-1)
        return(i);
  }
  return(-1);
}

void display_bank(void)
  {
  g3_window_set(win[WIN_BANK].x,win[WIN_BANK].y,win[WIN_BANK].w,win[WIN_BANK].h);
  g3_sprite(g3_bank_getaddr(currentbank),0,0,256,256,0,0);
  g3_vesa_rectcopy(g3_screen_getw(),g3_screen_geth(),g3_screen_getaddr(),g3_window_getx(),g3_window_gety(),win[WIN_BANK].w,win[WIN_BANK].h,win[WIN_BANK].x,win[WIN_BANK].y);
  g3_window_set(0,0,SCREENW,SCREENH);
  }

void display_palette(void)
  {
  SD i,j;
  g3_window_set(win[WIN_PAL].x,win[WIN_PAL].y,win[WIN_PAL].w,win[WIN_PAL].h);
  for(j=0;j<16;j++)
    for(i=0;i<16;i++)
    {
      g3_rectfill(i*8,j*8,i*8+7,j*8+7,i*16+j);
    }
  j=currentcolor % 16;
  i=currentcolor / 16;
  g3_rectfill(i*8,j*8,i*8+7,j*8+7,16);
  g3_vesa_rectcopy(g3_screen_getw(),g3_screen_geth(),g3_screen_getaddr(),g3_window_getx(),g3_window_gety(),win[WIN_PAL].w,win[WIN_PAL].h,win[WIN_PAL].x,win[WIN_PAL].y);
  g3_window_set(0,0,SCREENW,SCREENH);
  }


void display_object(void)
  {
    char string[256];
    g3_window_set(win[WIN_MAIN].x,win[WIN_MAIN].y,win[WIN_MAIN].w,win[WIN_MAIN].h);
    g3_window_clear(0);
    g3_ot_clear();
    g3_object_prepare(&g3_object[0]);
    g3_object_compute(&g3_object[0]);
    g3_ot_display();
    if(ptrpolygon!=NULL)
      {
      SD oldcolor;
      oldcolor=g3_polygon_colorget(ptrpolygon);
      g3_polygon_colorput(ptrpolygon,15);
      G3_POLYW(ptrpolygon);
      g3_polygon_colorput(ptrpolygon,oldcolor);
      }
    if(ptrpolygon!=NULL)
      {
      sprintf(string,"TYPE : %s",g3_polygon_typenameget(ptrpolygon));
      g3_string_put(font,0,0,string,7);
      }
    g3_vesa_rectcopy(g3_screen_getw(),g3_screen_geth(),g3_screen_getaddr(),g3_window_getx(),g3_window_gety(),win[WIN_MAIN].w,win[WIN_MAIN].h,win[WIN_MAIN].x,win[WIN_MAIN].y);
    g3_window_set(0,0,SCREENW,SCREENH);
  }

main(SD argc,char *argv[])
  {
  SD shift;
  SD oldmode;
  SD key,mbutton,mx,my;
  G3_VECTOR camtranslation;
  G3_ROTATION    camrotation;
  char filename[_MAX_PATH];
  char string[_MAX_PATH];
  PIXEL *addr;

  if(argc!=2) exit(-1);
  g3_path_set(argv[0]);


  if (!g3_object_vtxload(argv[1],&g3_object[0])) exit(-1);
  if (!g3_object_envload(argv[1],&g3_object[0])) exit(-1);
  if (!g3_object_sklload(argv[1],&g3_object[0])) exit(-1);


  sprintf(string,"%sFNT\\FONTLOW.DAT",g3_path_get());
  font=g3_font_load(string);
  if(font==NULL)
    {
    printf("Can t find %s\n",string);
    return(-1);
    }


  sprintf(string,"%sLND\\LAND",g3_path_get());
  g3_bank_load(filename,g3_vesa_getredfieldposition(),g3_vesa_getgreenfieldposition(),g3_vesa_getbluefieldposition(),g3_vesa_getredmasksize(),g3_vesa_getgreenmasksize(),g3_vesa_getbluemasksize());

  ptrpolygon=NULL;


  oldmode=g3_vesa_getmode();
  g3_vesa_setmode(G3_MODE640480);
  g3_init(g3_vesa_getwidth(),g3_vesa_getheight(),g3_vesa_getbytesperline(),G3_ZOOMSHIFT,G3_FPVALUE(g3_window_getw())/320,G3_FPVALUE(g3_window_geth())/240,(g3_vesa_getbitsperpixel()!=8));
  addr=g3_malloc(g3_vesa_getbytesperline()*g3_vesa_getheight());
  g3_screen_setaddr(addr);
  sprintf(string,"%sLND\\LAND",g3_path_get());
  g3_land_load(string);

  g3_nearplane_set(G3_FPVALUE(16));

  g3_mouse_init(g3_screen_getw(),g3_screen_geth());
  sprintf(filename,"%sBNK\\BANK",g3_path_get());

  g3_bank_load(filename,g3_vesa_getredfieldposition(),g3_vesa_getgreenfieldposition(),g3_vesa_getbluefieldposition(),g3_vesa_getredmasksize(),g3_vesa_getgreenmasksize(),g3_vesa_getbluemasksize());
  g3_setallcolorregister();

  g3_object[0].rotation.x=G3_FPANGLE(0);
  g3_object[0].rotation.y=G3_FPANGLE(0);
  g3_object[0].rotation.z=G3_FPANGLE(0);
  g3_object[0].translation.x=G3_FPVALUE(0);
  g3_object[0].translation.y=G3_FPVALUE(0);
  g3_object[0].translation.z=G3_FPVALUE(0);

  camrotation.x=G3_FPANGLE(0);
  camrotation.y=G3_FPANGLE(0);
  camrotation.z=G3_FPANGLE(0);
  camtranslation.x=G3_FPVALUE(0);
  camtranslation.y=G3_FPVALUE(0);
  camtranslation.z=G3_FPVALUE(512);
  g3_setpolarcamera(&camrotation,&camtranslation);

  /*---------- init events -------------------------*/
  key=0;
  while(key!=27)
    {
    /*-------------- display ------------------------------*/
    if(win[WIN_BANK].refresh)
      {
      display_bank();
      win[WIN_BANK].refresh=G3_FALSE;
      }
    if(win[WIN_PAL].refresh)
      {
      display_palette();
      win[WIN_PAL].refresh=G3_FALSE;
      }
    if(win[WIN_MAIN].refresh)
      {
      display_object();
      win[currentwin].refresh=G3_FALSE;
      }
    /*-------------- wait mouse events ------------------------------*/
    g3_mouse_waitevent();
    shift=(_bios_keybrd(_KEYBRD_SHIFTSTATUS) & 0x3);
    mx=g3_mouse_getx();
    my=g3_mouse_gety();
    mbutton=g3_mouse_getbutton();
    key=g3_mouse_getkey();
    currentwin=getwindow(mx,my);
    mx-=win[currentwin].x;
    my-=win[currentwin].y;
    /*--------------- manage mouse events -----------------------------------------------*/
    if(mbutton==1)
      {
        switch(currentwin)
        {
          case 0:
            g3_window_set(win[WIN_MAIN].x,win[WIN_MAIN].y,win[WIN_MAIN].w,win[WIN_MAIN].h);
            g3_ot_clear();
            g3_object_prepare(&g3_object[0]);
            g3_object_compute(&g3_object[0]);
            newptrpolygon=g3_polygon_selectinot(mx,my);
            if(newptrpolygon) ptrpolygon=newptrpolygon;
            win[WIN_MAIN].refresh=G3_TRUE;
            break;
          case 1:
            break;
          case 2:
            currentcolor=(my/8)+(mx/8)*16;
            win[WIN_PAL].refresh=G3_TRUE;
            break;
        }
      }
    else
    if(mbutton==2)
      {
        switch(currentwin)
        {
          case 0:
            if(ptrpolygon!=NULL)
              {
              g3_polygon_colorput(ptrpolygon,currentcolor);
              win[WIN_MAIN].refresh=G3_TRUE;
              }
            break;
          case 1:
            break;
          case 2:
            break;
        }
      }

    /*--------------- manage key's events -----------------------------------------------*/
    switch(key)
      {
      case G3_KEY_PGUP:
        currentbank++;
        if(currentbank>=G3_MAXBANK) currentbank=0;
        win[WIN_BANK].refresh=G3_TRUE;
        break;
      case G3_KEY_PGDN:
        currentbank--;
        if(currentbank<0) currentbank=G3_MAXBANK-1;
        win[WIN_BANK].refresh=G3_TRUE;
        break;
      case '+':
        camtranslation.z+=G3_FPVALUE(16);
        g3_setpolarcamera(&camrotation,&camtranslation);
        win[WIN_MAIN].refresh=G3_TRUE;
        break;
      case '-':
        camtranslation.z-=G3_FPVALUE(16);
        g3_setpolarcamera(&camrotation,&camtranslation);
        win[WIN_MAIN].refresh=G3_TRUE;
        break;
      case G3_KEY_LEFT:
        g3_object[0].rotation.y=G3_ANGLE(g3_object[0].rotation.y+G3_FPANGLE(G3_NBANGLE/16));
        win[WIN_MAIN].refresh=G3_TRUE;
        break;
      case G3_KEY_RIGHT:
        g3_object[0].rotation.y=G3_ANGLE(g3_object[0].rotation.y-G3_FPANGLE(G3_NBANGLE/16));
        win[WIN_MAIN].refresh=G3_TRUE;
        break;
      case G3_KEY_UP:
        g3_object[0].rotation.x=G3_ANGLE(g3_object[0].rotation.x+G3_FPANGLE(G3_NBANGLE/16));
        win[WIN_MAIN].refresh=G3_TRUE;
        break;
      case G3_KEY_DOWN:
        g3_object[0].rotation.x=G3_ANGLE(g3_object[0].rotation.x-G3_FPANGLE(G3_NBANGLE/16));
        win[WIN_MAIN].refresh=G3_TRUE;
        break;
      case ' ':
        if(ptrpolygon!=NULL)
          {
          G3_POLYGON *ptrpoly;
          SD oldcolor;
          oldcolor=g3_polygon_colorget(ptrpolygon);
          g3_polygon_colorput(ptrpolygon,currentcolor);
          ptrpoly=g3_object_envaddr(&g3_object[0]);
          if(ptrpoly!=NULL)
            {
            while(g3_polygon_typeget(ptrpoly)!=G3_ENDMARK)
              {
              if(g3_polygon_colorget(ptrpoly)==oldcolor)
                {
                g3_polygon_colorput(ptrpoly,currentcolor);
                }
              ptrpoly++;
              }
            }
          win[WIN_MAIN].refresh=G3_TRUE;
          }
        break;
      case 'r':
      case 'R':
        if(ptrpolygon!=NULL)
          {
          g3_polygon_rotate(ptrpolygon);
          win[WIN_MAIN].refresh=G3_TRUE;
          }
        break;
      case 'i':
      case 'I':
        if(ptrpolygon!=NULL)
          {
          g3_polygon_invert(ptrpolygon);
          win[WIN_MAIN].refresh=G3_TRUE;
          }
        break;
      case 'c':
      case 'g':
        if(ptrpolygon!=NULL)
          {
          SD inc;
          if(key=='c')
            inc=1;
          else
            inc=16;
          g3_polygon_colorput(ptrpolygon,g3_polygon_colorget(ptrpolygon)+inc);
          win[WIN_MAIN].refresh=G3_TRUE;
          }
        break;
      case 'D':
      case 'd':
          g3_graphcall_setlevel(g3_graphcall_getlevel()+1);
          win[WIN_MAIN].refresh=G3_TRUE;
          break;
      case 'C':
      case 'G':
        if(ptrpolygon!=NULL)
          {
          G3_POLYGON *ptrpoly;
          SD inc;
          SD oldcolor;
          if(key=='C')
            inc=1;
          else
            inc=16;
          oldcolor=g3_polygon_colorget(ptrpolygon);
          g3_polygon_colorput(ptrpolygon,g3_polygon_colorget(ptrpolygon)+inc);
          ptrpoly=g3_object_envaddr(&g3_object[0]);
          if(ptrpoly!=NULL)
            {
            while(g3_polygon_typeget(ptrpoly)!=G3_ENDMARK)
              {
              if(g3_polygon_colorget(ptrpoly)==oldcolor)
                {
                g3_polygon_colorput(ptrpoly,g3_polygon_colorget(ptrpolygon));
                }
              ptrpoly++;
              }
            }
          win[WIN_MAIN].refresh=G3_TRUE;
          }
        break;
      case 't':
        if(ptrpolygon!=NULL)
          {
          g3_polygon_typeput(ptrpolygon,g3_polygon_typeget(ptrpolygon)+1);
          win[WIN_MAIN].refresh=G3_TRUE;
          }
        break;
      case 'T':
        if(ptrpolygon!=NULL)
          {
          G3_POLYGON *ptrpoly;
          g3_polygon_typeput(ptrpolygon,g3_polygon_typeget(ptrpolygon)+1);
          ptrpoly=g3_object_envaddr(&g3_object[0]);
          if(ptrpoly!=NULL)
            {
            while(g3_polygon_typeget(ptrpoly)!=G3_ENDMARK)
              {
              if(g3_polygon_colorget(ptrpoly)==g3_polygon_colorget(ptrpolygon))
                {
                g3_polygon_typeput(ptrpoly,g3_polygon_typeget(ptrpolygon));
                }
              ptrpoly++;
              }
            }
          win[WIN_MAIN].refresh=G3_TRUE;
          }
        break;
      case ',':
      case '?':
        if(ptrpolygon!=NULL)
          {
          currentcolor=g3_polygon_colorget(ptrpolygon);
          win[WIN_PAL].refresh=G3_TRUE;
          }
        break;
     case '2':
        g3_object[0].rotation.x=G3_FPANGLE(0);
        g3_object[0].rotation.y=G3_FPANGLE(0);
        g3_object[0].rotation.z=G3_FPANGLE(0);
        g3_object[0].translation.x=G3_FPVALUE(0);
        g3_object[0].translation.y=G3_FPVALUE(0);
        g3_object[0].translation.z=G3_FPVALUE(0);
        win[WIN_MAIN].refresh=G3_TRUE;
        break;
      case 's':
      case 'S':
        if(!g3_object_envsave(&g3_object[0],argv[1])) exit(-1);
        break;
      }

    }
  g3_restore();
  g3_vesa_setmode(oldmode);
  g3_free(addr);
  return(0);
  }


