
#include <windows.h>
#include <windowsx.h>
#include <mmsystem.h>
#include <math.h>
#include <string.h>

#include "dd.h"
#include "ddlib.h"
#include "cdlib.h"

#include "g3def.h"
#include "g3lib.h"
#include "g3obj.h"
#include "g3malloc.h"
#include "g3timer.h"
#include "g3land.h"
#include "g3snd.h"
#include "g3pad.h"
#include "g3error.h"
#include "stm.h"

#define SCREEN_W  320
#define SCREEN_H  240
#define SCREEN_BBP 16

/**************************************************************************
  Global Application Variables
 **************************************************************************/

static char app_name[]="AquaSub 1996 L.C.";
static HINSTANCE  app_hinst;
static BOOL       app_active;
static BOOL       app_paused;
static HWND       app_hwnd;
static HACCEL     app_haccel;
static POINT      app_move;
static POINT      app_last;
static int        app_currentmode=0;
static int        app_halfline=FALSE;
static RECT       app_rect;
static RECT       app_clientrect;
static BOOL       app_displaystat=FALSE;

PIXEL *font;

/**************************************************************************
   Internal function declarations
 **************************************************************************/

LONG  CALLBACK appwndproc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lParam);
BOOL  appidle(void);
int   render(void);
int   init_all(void);
int   release_all(void);

/**************************************************************************
  AppAbout
 **************************************************************************/

BOOL FAR PASCAL dialog_about(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lParam)
{
  switch(msg)
  {
    case WM_COMMAND:
      if(LOWORD(wparam)==IDOK)
        EndDialog(hwnd,TRUE);
      break;
    case WM_INITDIALOG:
      return TRUE;
  }
  return FALSE;
}

/**************************************************************************
  AppInit
 **************************************************************************/
BOOL AppInit(HINSTANCE hInst,HINSTANCE hPrev,int sw,LPSTR szCmdLine)
{
  WNDCLASS cls;
  app_hinst=hInst;
  if(!hPrev)
  {
    cls.style         =CS_DBLCLKS;
    cls.hCursor       =LoadCursor(0,IDC_ARROW);
    cls.hIcon         =LoadIcon(hInst,"APPICON");
    cls.lpszMenuName  =NULL; //"AppMenu";
    cls.lpszClassName =app_name;
    cls.hbrBackground =NULL;//(HBRUSH)GetStockObject(BLACK_BRUSH);
    cls.hInstance     =hInst;
    cls.lpfnWndProc   =(WNDPROC)appwndproc;
    cls.cbClsExtra    =0;
    cls.cbWndExtra    =0;
    if(!RegisterClass(&cls))
      return FALSE;
  }
  app_haccel=LoadAccelerators(hInst,"AppAccel");

  SystemParametersInfo(SPI_GETWORKAREA,NULL,&app_rect,NULL);
  
  app_hwnd=CreateWindowEx(
                  WS_EX_APPWINDOW,
                  app_name,       // Class name
                  app_name,       // Caption
                  WS_OVERLAPPEDWINDOW /*| WS_CAPTION | WS_THICKFRAME | WS_MINIMIZEBOX | WS_MAXIMIZEBOX | WS_QUIT*/,
                  ((app_rect.right-app_rect.left)-SCREEN_W)>>1,
                  ((app_rect.bottom-app_rect.top)-SCREEN_H)>>1,          // Position
                  SCREEN_W,
                  SCREEN_H,   // Size
                  0,          // Parent window (no parent)
                  0,          // use class menu
                  hInst,      // handle to window instance
                  0           // no params to pass on
                  );
  ShowWindow(app_hwnd,sw);
  UpdateWindow(app_hwnd);
  GetWindowRect(app_hwnd,&app_rect);
  GetClientRect(app_hwnd,&app_clientrect);

  // initialise directdraw system
  #if 0
  if(!tdfx_init())
  {
    g3_warning("3DFX Chip Not detected. Software rendering");
  }
  else
  {
    tdfx_release();
  }
  #endif

  DD_init(app_hwnd,app_clientrect.right-app_clientrect.left,app_clientrect.bottom-app_clientrect.top,16,FALSE);
  g3_window_set(0,0,DD_getw(),DD_geth()/(app_halfline ? 2 : 1));
  g3_init(
            DD_getw(),
            (app_halfline ? DD_geth()/2 : DD_geth()),
            (app_halfline ? DD_getlinesize()*2 : DD_getlinesize()),
            G3_ZOOMSHIFT,
            G3_FPVALUE(g3_window_getw())/320,
            G3_FPVALUE(g3_window_geth())/240,
            TRUE
            );
  DD_cls();  
  // initialize sound system
  //if(!g3_snd_init(app_hwnd))
  //{
  //  return FALSE;
  //}

  // store a .waw in sound system

  //if(!g3_snd_addwav("SND\\tir04.wav"))
  //{
  //  return(FALSE);
  //}

  //g3_snd_play(0,0,0,0);

  // initialise various systems
  if(!init_all())
  {
    return FALSE;
  }
   
  return TRUE;

}

/**************************************************************************
  WinMain
 **************************************************************************/
int PASCAL WinMain(HINSTANCE hInst,HINSTANCE hPrev,LPSTR szCmdLine,int sw)
{
  MSG     msg;


  //*** Call initialization procedure
  if(!AppInit(hInst,hPrev,sw,szCmdLine))
    return FALSE;
  //*** Polling messages from event queue until quit
  for (;;)
  {
    if(PeekMessage(&msg,0,0,0,PM_REMOVE))
    {
      if(msg.message==WM_QUIT)
        break;
      if(!app_hwnd || !TranslateAccelerator(app_hwnd,app_haccel,&msg))
      {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
      }
    }
    else
    {
      if(appidle())
        WaitMessage();
    }
  }
  DD_release();
  return msg.wParam;
}

/**************************************************************************
  AppPause
 **************************************************************************/
void AppPause(BOOL f)
{
    if(f)
    {
      app_paused=TRUE;
      RedrawWindow(app_hwnd,NULL,NULL,RDW_FRAME);
    }
    else
    {
      app_paused=FALSE;
    }
}

/**************************************************************************
  appidle
 **************************************************************************/
BOOL appidle()
{
  if(app_active && !app_paused)
  {
    /*if(GetAsyncKeyState(VK_LBUTTON)<0)
    {
        if(app_move.x || app_move.y)
        {
        app_move.x=0;
        app_move.y=0;
        }
    }
    */
    render();
    return FALSE;
  }
  else
  {
    return TRUE;
  }
}

/**************************************************************************
  init_all
 **************************************************************************/
int init_all(void)
{
  //CD_open(app_hwnd);
  //CD_playtrack(11);
  font=g3_font_load("DAT\\FNT\\FONTLOW.DAT");
  g3_land_load("DAT\\LND\\LAND");
  app_currentmode=0;
  g3_bank_load("DAT\\BNK\\BANK",DD_getrpos(),DD_getgpos(),DD_getbpos(),DD_getrbits(),DD_getgbits(),DD_getbbits());
  g3_nearplane_set(G3_FPVALUE(32));
  g3_camera_set(G3_FPANGLE(0),G3_FPANGLE(G3_NBANGLE/4),G3_FPANGLE(0),G3_FPVALUE(0),G3_FPVALUE(128),G3_FPVALUE(0));
  g3_particle_init();
  g3_pad_init();
  g3_timer_init();
  g3_allobject_init();
  g3_snd_init(app_hwnd);

  g3_snd_add("dat\\wav\\lazer1");
  
  {
    SD i;
    G3_VECTOR pos;
    G3_ROTATION rot;
    rot.x=rot.y=rot.z=0;
    pos.y=G3_FPVALUE(G3_LANDGRIDSTEP*10);
    pos.x=pos.z=G3_FPVALUE(G3_LANDGRIDSTEP*30);
    g3_object_add("dat\\obj\\dolph\\dolph1",stm_dolph,0,0,&pos,&rot,1000,1000);
    pos.y=G3_FPVALUE(G3_LANDGRIDSTEP*5);
    pos.x=pos.z=G3_FPVALUE(G3_LANDGRIDSTEP*32);
    g3_object_setplayer(g3_object_add("dat\\obj\\sub\\sub",stm_sub,0,0,&pos,&rot,1000,1000));
    pos.y=G3_FPVALUE(G3_LANDGRIDSTEP*3);
    pos.x=pos.z=G3_FPVALUE(G3_LANDGRIDSTEP*40);
    //i=g3_object_add("dat\\obj\\modul\\modul",stm_dummy,0,0,&pos,&rot,1000,1000);
    pos.y=G3_FPVALUE(G3_LANDGRIDSTEP*10);
    pos.x=pos.z=G3_FPVALUE(G3_LANDGRIDSTEP*35);
    i=g3_object_add("dat\\obj\\cube\\cube",stm_canon,0,0,&pos,&rot,1000,1000);
    for(i=1;i<30;i++)
    {
      pos.x=(rand() & 0xFF)*G3_FPVALUE(G3_LANDGRIDSTEP);
      pos.z=(rand() & 0xFF)*G3_FPVALUE(G3_LANDGRIDSTEP);
      pos.y=G3_FPVALUE(128);
      g3_object_add("dat\\obj\\cube\\cube",stm_dummy,0,0,&pos,&rot,1,1);
    }
  }
  return(TRUE);
  }
  
int release_all(void)
{
  g3_snd_release();
  g3_particle_restore();
  g3_allobject_restore();
  g3_pad_restore();
  g3_timer_restore();
  g3_restore();
  //CD_stop();
  //CD_close();
  return(TRUE);
}

/**************************************************************************
  Render
 **************************************************************************/



 
int render(void)
{
  static float framerate=0;
  static int   passcount=0;
  static int   oldtime=0;
  static char  string[256];
  static int   ottime=0;
  static float otpercent=0;
  static int   managetime=0;
  static float managepercent=0;
  static int   computetime=0;
  static float computepercent=0;
  static int   fliptime=0;
  static float flippercent=0;
  int newtime;
  unsigned short *ptr;

  ptr=(unsigned short *)DD_lock();
  if(ptr)
  {
    DD_unlock();

    g3_timer_sample();
    g3_screen_setaddr(ptr);
    g3_ot_clear();
    g3_pad_sample();

    {
      int startcomputetime;  
      startcomputetime=timeGetTime();  
      g3_particle_compute();
      g3_allobject_compute();
      g3_particle_compute();
      g3_land_compute();
      computetime+=timeGetTime()-startcomputetime;  
    }

    {
      int startottime;  
      startottime=timeGetTime();  
      g3_ot_display();
      ottime+=timeGetTime()-startottime;  
    }

    {
      int startmanagetime;  
      startmanagetime=timeGetTime();
      g3_particle_manage();
      g3_allobject_manage();
      camera_compute();
      managetime+=timeGetTime()-startmanagetime;  
    }

    if(app_displaystat)
    {
      passcount++;
      if(oldtime==0) oldtime=timeGetTime();
      newtime=timeGetTime();
      if(newtime-oldtime>=1000)
      {
        framerate=(float)(passcount*1000)/(newtime-oldtime);
        otpercent=((float)ottime*100)/(newtime-oldtime);
        managepercent=((float)managetime*100)/(newtime-oldtime);
        computepercent=((float)computetime*100)/(newtime-oldtime);
        flippercent=((float)fliptime*100)/(newtime-oldtime);
        oldtime=newtime;
        passcount=0;
        ottime=0;
        fliptime=0;
        managetime=0;
        computetime=0;    
      }
      sprintf(string,"Frame rate:%5.2f",framerate);
      g3_string_put(font,0, 0,string,(passcount==0 ? 15:7));
      sprintf(string,"otdisplay:%5.2f \%",otpercent);
      g3_string_put(font,0, 8,string,(passcount==0 ? 15:7));
      sprintf(string,"compute:%5.2f \%",computepercent);
      g3_string_put(font,0,16,string,(passcount==0 ? 15:7));
      sprintf(string,"manage:%6.2f \%",managepercent);
      g3_string_put(font,0,24,string,(passcount==0 ? 15:7));
      sprintf(string,"flipping:%6.2f \%",flippercent);
      g3_string_put(font,0,32,string,(passcount==0 ? 15:7));
      sprintf(string,"total:%6.2f \%",managepercent+computepercent+otpercent+flippercent);
      g3_string_put(font,0,40,string,(passcount==0 ? 15:7));
      sprintf(string,"objects:%d",g3_object_getcount());
      g3_string_put(font,0,48,string,(passcount==0 ? 15:7));
      sprintf(string,"objects active:%d",g3_object_getactivecount());
      g3_string_put(font,0,56,string,(passcount==0 ? 15:7));
      sprintf(string,"polygons:%d",g3_ot_getcount());
      g3_string_put(font,0,64,string,(passcount==0 ? 15:7));
    }
  }
  
  {
    int startfliptime;  
    startfliptime=timeGetTime();  
    DD_flip();
    fliptime+=timeGetTime()-startfliptime;  
  }
  return(TRUE);
}

/**************************************************************************
  appwndproc
 **************************************************************************/
LONG CALLBACK appwndproc(HWND hwnd,UINT msg,WPARAM wparam,LPARAM lParam)
{
  switch(msg)
  {
    case WM_CREATE:
      break;
    case WM_SYSKEYUP:
        switch( wparam )
        {
        case VK_RETURN:
          if(DD_getfullscreen())
          {
            //windowed configuration
            ShowCursor(TRUE);
            DD_init(app_hwnd,app_clientrect.right-app_clientrect.left,app_clientrect.bottom-app_clientrect.top,16,FALSE);
            SetWindowPos(app_hwnd,NULL,app_rect.left,app_rect.top,app_rect.right-app_rect.left,app_rect.bottom-app_rect.top,/*SWP_NOMOVE |*/ SWP_NOZORDER | SWP_NOACTIVATE);
            GetClientRect(app_hwnd,&app_clientrect);
          }
          else
          {
            //full screen configuration
            ShowCursor(FALSE);
            DD_init(app_hwnd,DD_getmodew(app_currentmode),DD_getmodeh(app_currentmode),DD_getmodebpp(app_currentmode),TRUE);
          }
          g3_window_set(0,0,DD_getw(),DD_geth()/(app_halfline ? 2 : 1));
          g3_init(
            DD_getw(),
            (app_halfline ? DD_geth()/2 : DD_geth()),
            (app_halfline ? DD_getlinesize()*2 : DD_getlinesize()),
            G3_ZOOMSHIFT,
            G3_FPVALUE(g3_window_getw())/320,
            G3_FPVALUE(g3_window_geth())/240,
            TRUE
            );
          if(DD_pixelformatchanged())
          {
            g3_bank_load("DAT\\BNK\\BANK",DD_getrpos(),DD_getgpos(),DD_getbpos(),DD_getrbits(),DD_getgbits(),DD_getbbits());
          }
        }
        break;
    case WM_PAINT:
      if(DD_active())
        DD_flip();
      break;
    case WM_ACTIVATEAPP:
      app_active=(BOOL)wparam;
      break;
    case WM_SETCURSOR:
      break;
    case WM_ENTERMENULOOP:
      break;
    case WM_EXITMENULOOP:
      break;
    case WM_INITMENUPOPUP:
      break;
    case WM_KEYDOWN:
      switch(wparam)
      {
        case VK_F1:
        if(!DD_getfullscreen())
        {
        FARPROC proc;
        proc=MakeProcInstance(dialog_about,app_hinst);
        DialogBox(app_hinst,"DIALOG_ABOUT",app_hwnd,proc);
        FreeProcInstance(proc);
        }
        break;
        case VK_F2:
        app_displaystat=!app_displaystat;
        break;        
        case VK_F3:
        if(g3_graphcall_getlevel()==G3_GRAPHCALL_LEVEL_FULL)
        {
          g3_graphcall_setall(G3_GRAPHCALL_LEVEL_ALLFLAT);
        }
        else
        {
          if(g3_graphcall_getlevel()==G3_GRAPHCALL_LEVEL_ALLFLAT)
          {
          g3_graphcall_setall(G3_GRAPHCALL_LEVEL_ALLFILLEDWIRE);
          }
          else
          {
          g3_graphcall_setall(G3_GRAPHCALL_LEVEL_FULL);
          }
        }
        break;
        case VK_F4:
        if(DD_getfullscreen())
        {
          app_currentmode=(app_currentmode+1)%DD_getmodecount();
          DD_init(app_hwnd,DD_getmodew(app_currentmode),DD_getmodeh(app_currentmode),DD_getmodebpp(app_currentmode),TRUE);
          g3_window_set(0,0,DD_getw(),DD_geth()/(app_halfline ? 2 : 1));
          g3_init(
            DD_getw(),
            (app_halfline ? DD_geth()/2 : DD_geth()),
            (app_halfline ? DD_getlinesize()*2 : DD_getlinesize()),
            G3_ZOOMSHIFT,
            G3_FPVALUE(g3_window_getw())/320,
            G3_FPVALUE(g3_window_geth())/240,
            TRUE
            );
          DD_cls();  
        } 
        break;
        case VK_F5:
        {
          app_halfline=!app_halfline;
          g3_window_set(0,0,DD_getw(),DD_geth()/(app_halfline ? 2 : 1));
          g3_init(
            DD_getw(),
            (app_halfline ? DD_geth()/2 : DD_geth()),
            (app_halfline ? DD_getlinesize()*2 : DD_getlinesize()),
            G3_ZOOMSHIFT,
            G3_FPVALUE(g3_window_getw())/320,
            G3_FPVALUE(g3_window_geth())/240,
            TRUE
            );
          DD_cls();  
        } 
        break;
        case VK_F6:
        {
          if(g3_object_getplayer()==g3_object_getaddr(0))
            g3_object_setplayer(g3_object_getaddr(1));
          else
            g3_object_setplayer(g3_object_getaddr(0));
        }
        break;
      }
      break;
    case WM_CHAR:
      switch(wparam)
      {
        case 'B':
        case 'b':
        {
          g3_bank_load(
            "DAT\\BNK\\BANK",
            DD_getrpos(),
            DD_getgpos(),
            DD_getbpos(),
            DD_getrbits(),
            DD_getgbits(),
            DD_getbbits()
            );
        }
        break;
      }
      break;        
    case WM_COMMAND:
      switch(LOWORD(wparam))
      {
      }
      return 0L;
    case WM_DESTROY:
      app_hwnd=NULL;
      release_all();
      PostQuitMessage(0);
      break;
    case WM_MOVE:
      if(!DD_getfullscreen())
      {
        GetWindowRect(app_hwnd,&app_rect);
        GetClientRect(app_hwnd,&app_clientrect);
      }
      break;
    case WM_SIZE:
      if(!DD_getfullscreen())
      {
        SD hmargin,vmargin;
        GetWindowRect(app_hwnd,&app_rect);
        GetClientRect(app_hwnd,&app_clientrect);
        hmargin=(app_rect.right-app_rect.left)-(app_clientrect.right-app_clientrect.left);
        vmargin=(app_rect.bottom-app_rect.top)-(app_clientrect.bottom-app_clientrect.top);
        app_rect.bottom=app_rect.top+(((app_clientrect.right-app_clientrect.left)*480)/640)+vmargin;
        SetWindowPos(app_hwnd,NULL,app_rect.left,app_rect.right,app_rect.right-app_rect.left,app_rect.bottom-app_rect.top,SWP_NOMOVE | SWP_NOZORDER | SWP_NOACTIVATE);
        GetClientRect(app_hwnd,&app_clientrect);
        DD_init(app_hwnd,app_clientrect.right-app_clientrect.left,app_clientrect.bottom-app_clientrect.top,16,FALSE);
        g3_window_set(0,0,DD_getw(),DD_geth()/(app_halfline ? 2 : 1));
        g3_init(
            DD_getw(),
            (app_halfline ? DD_geth()/2 : DD_geth()),
            (app_halfline ? DD_getlinesize()*2 : DD_getlinesize()),
            G3_ZOOMSHIFT,
            G3_FPVALUE(g3_window_getw())/320,
            G3_FPVALUE(g3_window_geth())/240,
            TRUE
            );
        DD_cls();  
      }
      break;
    case WM_DISPLAYCHANGE:
      break;
    case WM_LBUTTONDOWN:
      app_last.x=LOWORD(lParam);
      app_last.y=HIWORD(lParam);
      break;
    //case WM_NCMOUSEMOVE:        // get mouse position in non client area. 
    case WM_MOUSEMOVE:
      if(GetKeyState(VK_LBUTTON) < 0)
      {
        app_move.x=(int)LOWORD(lParam)-app_last.x;
        app_move.y=(int)HIWORD(lParam)-app_last.y;
        app_last.x=LOWORD(lParam);
        app_last.y=HIWORD(lParam);
      }
      break;
  }
  return DefWindowProc(hwnd,msg,wparam,lParam);
}




