
#include <windows.h>
#include <windowsx.h>
#include <mmsystem.h>
#include <math.h>
#include <string.h>

#include "ddraw.h"
#include "ddlib.h"


/**************************************************************************
  DirectDraw Globals
 **************************************************************************/

#define DD_MAXMODE  64

typedef struct
  {
    int w;
    int h;
    int bpp;
  }
DD_MODE;

DD_MODE               dd_mode[DD_MAXMODE];
int                   dd_modecount=0;
HWND                  dd_hwnd=NULL;         //handle of directdraw window
IDirectDraw           *dd=NULL;        
IDirectDrawSurface    *dd_primary=NULL;
IDirectDrawSurface    *dd_frontbuffer=NULL;
IDirectDrawSurface    *dd_backbuffer=NULL;
IDirectDrawClipper    *dd_clipper=NULL;
int                   dd_fullscreen=TRUE;
int                   dd_w;
int                   dd_h;
int                   dd_bpp;
int                   dd_rpos,dd_gpos,dd_bpos;
int                   dd_rbits,dd_gbits,dd_bbits;
int                   dd_linesize;
int                   dd_pixelformatchanged=TRUE;


HRESULT CALLBACK DD_modecallback(LPDDSURFACEDESC pdds,LPVOID lParam);


/**************************************************************************
  ModeCallback
 **************************************************************************/

int DD_getposfrommask(unsigned int mask)
{
  int i;
  i=0;
  while(!(mask & 0x1) && i<32)
  {
    mask>>=1;
    i++;
  }
  return(i);
}

int DD_getbitsfrommask(unsigned int mask)
{
  int i,j;
  i=0;
  while(!(mask & 0x1) && i<32)
  {
    mask>>=1;
    i++;
  }
  j=0;
  while((mask & 0x1) && i<32)
  {
    mask>>=1;
    i++;
    j++;
  }
  return(j);
}

HRESULT CALLBACK DD_modecallback(LPDDSURFACEDESC pdds,LPVOID lParam)
{
  if(pdds->ddpfPixelFormat.dwRGBBitCount==16||pdds->ddpfPixelFormat.dwRGBBitCount==15)
  {
    int i;

    dd_mode[dd_modecount].w=pdds->dwWidth;
    dd_mode[dd_modecount].h=pdds->dwHeight;
    dd_mode[dd_modecount].bpp=pdds->ddpfPixelFormat.dwRGBBitCount;

    i=0;
    while(i<dd_modecount)
    {
      if(memcmp(&dd_mode[dd_modecount],&dd_mode[i],sizeof(DD_MODE))==0)
      {
        goto duplicate;
      }
      i++;
    }
    dd_modecount++;
    duplicate:;
  }
    return S_FALSE;
}


/**************************************************************************
  DD_release
 **************************************************************************/
void DD_release()
{
    if(dd_backbuffer)
    {
      IDirectDrawSurface_Release(dd_backbuffer);
      dd_backbuffer=NULL;
    }
    if(dd_frontbuffer)
    {
      IDirectDrawSurface_Release(dd_frontbuffer);
      dd_frontbuffer=NULL;
    }
    if(dd_primary)
    {
      IDirectDrawSurface_Release(dd_primary);
      dd_primary=NULL;
    }
    if(dd_clipper)
    {
      IDirectDrawClipper_Release(dd_clipper);
      dd_clipper=NULL;
    }
    if(dd)
    {
      IDirectDraw_Release(dd);
      dd=NULL;
    }
}

int DD_active(void)
{
  return(dd!=NULL);
}

/**************************************************************************
  DD_getfullscreen
 **************************************************************************/
int DD_getfullscreen(void)
{
  return(dd_fullscreen);
}

/**************************************************************************
  DD_setfullscreen
 **************************************************************************/
void DD_setfullscreen(int n)
{
  dd_fullscreen=n;
}

/**************************************************************************
  DD_getw
 **************************************************************************/
int DD_getw(void)
{
  return(dd_w);
}

/**************************************************************************
  DD_geth
 **************************************************************************/
int DD_geth(void)
{
  return(dd_h);
}


/**************************************************************************
  DD_init
 **************************************************************************/

BOOL DD_init(HWND hwnd,int width,int height,int bpp,int fullscreen)
{
  HRESULT err;
  DDSURFACEDESC ddsd;
  DDSCAPS caps;
  int rpos,gpos,bpos,rbits,gbits,bbits;
  
  dd_fullscreen=fullscreen;
  dd_hwnd=hwnd;
  dd_w=width;
  dd_h=height;
  DD_release();
  err=DirectDrawCreate(NULL,&dd,NULL);
  if(err != DD_OK)
  {
        g3_error("DD_init : Reinstall DirectX(tm) please");
        return FALSE;
  }
  dd_modecount=0;
  IDirectDraw_EnumDisplayModes(dd,0,NULL,NULL,DD_modecallback);

  if(dd_modecount==0)
  {
        g3_error("DD_init : No 32K or 65K video mode. Reinstall DirectX(tm) or drivers please");
        return FALSE;
  }
  // sort modes by increasing width and height
  {
    int i,j;
    DD_MODE tmp;
    for(i=0;i<dd_modecount-1;i++)
    {
      for(j=i+1;j<dd_modecount;j++)
      {
        if(dd_mode[j].w*65536+dd_mode[j].h<dd_mode[i].w*65536+dd_mode[i].h)
        {
          tmp=dd_mode[i];
          dd_mode[i]=dd_mode[j];
          dd_mode[j]=tmp;
        }
      }
    }
  }

  if(dd_fullscreen)
  {
    err=IDirectDraw_SetCooperativeLevel(dd,dd_hwnd,DDSCL_EXCLUSIVE|DDSCL_FULLSCREEN/*|DDSCL_ALLOWMODEX*/);
    if(err != DD_OK)
      return FALSE;
    err=IDirectDraw_SetDisplayMode(dd,width,height,bpp);
    if(err != DD_OK)
      return FALSE;
    memset(&ddsd,0,sizeof(ddsd));
    ddsd.dwSize=sizeof(ddsd);
    ddsd.dwFlags=DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
    ddsd.dwBackBufferCount=2;
    ddsd.ddsCaps.dwCaps=DDSCAPS_PRIMARYSURFACE|DDSCAPS_FLIP|DDSCAPS_COMPLEX|DDSCAPS_VIDEOMEMORY;
    // try to get a triple buffered video memory surface.
    err=IDirectDraw_CreateSurface(dd,&ddsd,&dd_frontbuffer,NULL);
    if(err != DD_OK)
    {
      // try to get a double buffered video memory surface.
      ddsd.dwBackBufferCount=1;
      err=IDirectDraw_CreateSurface(dd,&ddsd,&dd_frontbuffer,NULL);
    }
    if(err != DD_OK)
    {
      // settle for a main memory surface.
      ddsd.ddsCaps.dwCaps &= ~DDSCAPS_VIDEOMEMORY;
      err=IDirectDraw_CreateSurface(dd,&ddsd,&dd_frontbuffer,NULL);
    }
    if(err != DD_OK)
      return FALSE;
    // get a pointer to the back buffer
    caps.dwCaps=DDSCAPS_BACKBUFFER;
    err=IDirectDrawSurface_GetAttachedSurface(dd_frontbuffer,&caps,&dd_backbuffer);
    if(err != DD_OK)
      return FALSE;
  }
  else
  {
        //RECT rcWork;
        //RECT rc;
        //HDC hdc;
        DWORD dwStyle;
        err=IDirectDraw_SetCooperativeLevel(dd,dd_hwnd,DDSCL_NORMAL);
        if(err != DD_OK)
        {
          return FALSE;
        }
 
        SetWindowPos(dd_hwnd,HWND_NOTOPMOST,0,0,0,0,
            SWP_NOSIZE | SWP_NOMOVE | SWP_NOACTIVATE);

        memset(&ddsd,0,sizeof(DDSURFACEDESC));
        ddsd.dwSize=sizeof(DDSURFACEDESC);
        ddsd.dwFlags=DDSD_CAPS;
        ddsd.ddsCaps.dwCaps=DDSCAPS_PRIMARYSURFACE;
        err=IDirectDraw_CreateSurface(dd,&ddsd,&dd_primary,NULL);
        if(err!=DD_OK)
        {
          return FALSE;
        }
        ddsd.dwFlags=DDSD_WIDTH | DDSD_HEIGHT | DDSD_CAPS;
        ddsd.dwWidth=width;
        ddsd.dwHeight=height;
        ddsd.ddsCaps.dwCaps= DDSCAPS_OFFSCREENPLAIN|DDSCAPS_VIDEOMEMORY;
        err=IDirectDraw_CreateSurface(dd,&ddsd,&dd_backbuffer,NULL);
        if(err!=DD_OK)
        {
          // settle for a main memory surface.
          ddsd.ddsCaps.dwCaps &= ~DDSCAPS_VIDEOMEMORY;
          err=IDirectDraw_CreateSurface(dd,&ddsd,&dd_backbuffer,NULL);
        }
        if(err!=DD_OK)
        {
          return FALSE;
        }
        err=IDirectDraw_CreateClipper(dd,0,&dd_clipper,NULL);
        if(err!=DD_OK)
        {
          return FALSE;
        }
        err=IDirectDrawClipper_SetHWnd(dd_clipper,0,dd_hwnd);
        if(err!=DD_OK)
        {
          return FALSE;
        }
        err=IDirectDrawSurface_SetClipper(dd_primary,dd_clipper);
        if(err!=DD_OK)
        {
          return FALSE;
        }
  }

  memset(&ddsd,0,sizeof(ddsd));
  ddsd.dwSize=sizeof(ddsd);
  err=IDirectDrawSurface_GetSurfaceDesc(dd_backbuffer,&ddsd);
  if(err!=DD_OK)
  {
    return FALSE;
  }

  dd_linesize=ddsd.lPitch;
  rpos=DD_getposfrommask(ddsd.ddpfPixelFormat.dwRBitMask);
  gpos=DD_getposfrommask(ddsd.ddpfPixelFormat.dwGBitMask);
  bpos=DD_getposfrommask(ddsd.ddpfPixelFormat.dwBBitMask);
  rbits=DD_getbitsfrommask(ddsd.ddpfPixelFormat.dwRBitMask);
  gbits=DD_getbitsfrommask(ddsd.ddpfPixelFormat.dwGBitMask);
  bbits=DD_getbitsfrommask(ddsd.ddpfPixelFormat.dwBBitMask);
  dd_pixelformatchanged=FALSE;
  if(rpos!=dd_rpos || gpos!=dd_gpos || bpos!=dd_bpos ||
     rbits!=dd_rbits || gbits!=dd_gbits || bbits!=dd_bbits)
     {
       dd_pixelformatchanged=TRUE;
     }
  dd_rpos=rpos;
  dd_gpos=gpos;
  dd_bpos=bpos;
  dd_rbits=rbits;
  dd_gbits=gbits;
  dd_bbits=bbits;

  return TRUE;
}

/**************************************************************************
  DD_flip
 **************************************************************************/
int DD_flip(void)
{
  HRESULT err;
  if(dd_fullscreen)
  {
      err=IDirectDrawSurface_Flip(dd_frontbuffer,NULL,DDFLIP_WAIT);
      if(err!=DD_OK)
      {
      return FALSE;
      }
  }
  else
  {
    RECT rc;
    GetClientRect(dd_hwnd,&rc);
    ClientToScreen(dd_hwnd,(LPPOINT)&rc);
    ClientToScreen(dd_hwnd,(LPPOINT)&rc+1);
    err=IDirectDrawSurface_Blt(dd_primary,&rc,dd_backbuffer,NULL,DDBLT_WAIT,NULL);
    if(err!=DD_OK)
    {
      return FALSE;
    }
  }
  return(TRUE);
}

/**************************************************************************
  DD_lock
 **************************************************************************/
void *DD_lock()
{
  DDSURFACEDESC ddsd;
  ddsd.dwSize=sizeof(ddsd);
  if(IDirectDrawSurface_Lock(dd_backbuffer,NULL,&ddsd,DDLOCK_WAIT,NULL)!=DD_OK)
  {
    return NULL;
  }
  return((void *)ddsd.lpSurface);
}


/**************************************************************************
  DD_unlock
 **************************************************************************/
int DD_unlock(void)
{
  return IDirectDrawSurface_Unlock(dd_backbuffer,NULL);
}

int DD_getmodecount(void)
{
  return(dd_modecount);
}

int DD_getmodew(int mode)
{
  return(dd_mode[mode].w);
}

int DD_getmodeh(int mode)
{
  return(dd_mode[mode].h);
}

int DD_getmodebpp(int mode)
{
  return(dd_mode[mode].bpp);
}

int DD_getbpp(void)
{
  return(dd_bpp);
}

int DD_getrpos(void)
{
  return(dd_rpos);
}

int DD_getgpos(void)
{
  return(dd_gpos);
}

int DD_getbpos(void)
{
  return(dd_bpos);
}

int DD_getrbits(void)
{
  return(dd_rbits);
}

int DD_getgbits(void)
{
  return(dd_gbits);
}

int DD_getbbits(void)
{
  return(dd_bbits);
}

int DD_getlinesize(void)
{
  return(dd_linesize);
}




void DD_cls(void)
{
  int i;
  unsigned short *ptr;
  for(i=0;i<3;i++)
  {
    ptr=(unsigned short *)DD_lock();
    memset(ptr,0,dd_h*dd_linesize);
    DD_unlock();
    DD_flip();
  }
}

int DD_pixelformatchanged(void)
{
  return(dd_pixelformatchanged);
}

