
#include "g3def.h"
#include "g3file.h"
#include "g3error.h"
#include "g3malloc.h"
#include "g3env.h"

/**************************************************************************/
/*              MANAGEMENT OF .ENV                                        */
/**************************************************************************/

SD g3_env_count(G3_POLYGON *ptrpoly)
  {
  SD count;
  count=0;
  while(ptrpoly->type!=G3_ENDMARK)
      {
      count++;
      ptrpoly++;
      }
  return(count);
  }


G3_POLYGON *g3_env_load(char *name)
  {
  SD version;
  G3_FILE *handle;
  char filename[_MAX_PATH];
  SD polycount,i;
  G3_POLYGON *ptrpoly,*pptrpoly;
  /*----- count number of vertex in this object -----*/
  ptrpoly=NULL;
  g3_fextput(name,".ENV",filename);
  handle=g3_fopen(filename,"rb");
  if(handle!=NULL)
    {
    g3_fread(&version,sizeof(version),1,handle);
    g3_fread(&polycount,sizeof(polycount),1,handle);
    pptrpoly=ptrpoly=(G3_POLYGON *)(g3_malloc(sizeof(G3_POLYGON)*polycount+sizeof(SD)));
      for(i=0;i<polycount;i++)
        {
        switch(version)
          {
          case 1:
            g3_fread(&ptrpoly->type  ,sizeof(ptrpoly->type    ),1,handle);
            g3_fread(&ptrpoly->color ,sizeof(ptrpoly->color   ),1,handle);
            g3_fread(&ptrpoly->bank,sizeof(ptrpoly->bank  ),1,handle);
            g3_fread(&ptrpoly->pts1  ,sizeof(ptrpoly->pts1    ),1,handle);
            g3_fread(&ptrpoly->pts2  ,sizeof(ptrpoly->pts2    ),1,handle);
            g3_fread(&ptrpoly->pts3  ,sizeof(ptrpoly->pts3    ),1,handle);
            ptrpoly->pts1*=sizeof(G3_DVERTEX);
            ptrpoly->pts2*=sizeof(G3_DVERTEX);
            ptrpoly->pts3*=sizeof(G3_DVERTEX);
            g3_fread(&ptrpoly->xa  ,sizeof(ptrpoly->xa    ),1,handle);
            g3_fread(&ptrpoly->ya  ,sizeof(ptrpoly->ya    ),1,handle);
            g3_fread(&ptrpoly->xb  ,sizeof(ptrpoly->xb    ),1,handle);
            g3_fread(&ptrpoly->yb  ,sizeof(ptrpoly->yb    ),1,handle);
            g3_fread(&ptrpoly->xc  ,sizeof(ptrpoly->xc    ),1,handle);
            g3_fread(&ptrpoly->yc  ,sizeof(ptrpoly->yc    ),1,handle);
            break;
          default:
            g3_error("g3_env_load : Error with version");
          }
        ptrpoly++;
        }
      g3_fread(&ptrpoly->type ,sizeof(ptrpoly->type),1,handle);
    g3_fclose(handle);
    }
  else
    {
    g3_error("g3_env_load : Error while openning file");
    }
  return(pptrpoly);
  }

SD g3_env_save(G3_POLYGON *ptrpoly,char *name)
  {
  SD version;
  G3_FILE *handle;
  char filename[_MAX_PATH];
  SD polycount,i;
  VERTEXINDEX pts;
  /*----- count number of vertex in this object -----*/
  polycount=g3_env_count(ptrpoly);
  if(polycount>0)
    {
    g3_fextput(name,".ENV",filename);
    handle=g3_fopen(filename,"wb");
    if(handle!=NULL)
      {
      version=1;
      g3_fwrite(&version,sizeof(version),1,handle);
      g3_fwrite(&polycount,sizeof(polycount),1,handle);
      for(i=0;i<polycount;i++)
        {
        g3_fwrite(&ptrpoly->type                     ,sizeof(ptrpoly->type    ),1,handle);
        g3_fwrite(&ptrpoly->color                    ,sizeof(ptrpoly->color   ),1,handle);
        g3_fwrite(&ptrpoly->bank                   ,sizeof(ptrpoly->bank  ),1,handle);


        pts=ptrpoly->pts1/sizeof(G3_DVERTEX);
        g3_fwrite(&pts,sizeof(pts),1,handle);
        pts=ptrpoly->pts2/sizeof(G3_DVERTEX);
        g3_fwrite(&pts,sizeof(pts),1,handle);
        pts=ptrpoly->pts3/sizeof(G3_DVERTEX);
        g3_fwrite(&pts,sizeof(pts),1,handle);

        g3_fwrite(&ptrpoly->xa  ,sizeof(ptrpoly->xa      ),1,handle);
        g3_fwrite(&ptrpoly->ya  ,sizeof(ptrpoly->ya      ),1,handle);
        g3_fwrite(&ptrpoly->xb  ,sizeof(ptrpoly->xb      ),1,handle);
        g3_fwrite(&ptrpoly->yb  ,sizeof(ptrpoly->yb      ),1,handle);
        g3_fwrite(&ptrpoly->xc  ,sizeof(ptrpoly->xc      ),1,handle);
        g3_fwrite(&ptrpoly->yc  ,sizeof(ptrpoly->yc      ),1,handle);
        ptrpoly++;
        }
      g3_fwrite(&ptrpoly->type ,sizeof(ptrpoly->type),1,handle);
      g3_fclose(handle);
      return(G3_TRUE);
      }
    else
      {
      g3_error("g3_env_save : open error");
      }
    }
  else
    {
    g3_error("g3_env_save : polycount<=0");
    }
  return(G3_FALSE);
  }

