

#include "g3def.h"
#include "g3graph.h"
#include "g3error.h"
#include "g3land.h"

  /*---------- land storage -------------*/

  typedef struct
  {
    SD  x1;
    SD  y1;
    SD  x2;
    SD  y2;
    SD  x3;
    SD  y3;
    SD  x4;
    SD  y4;
  }
  G3_LANDTEX;

  G3_LANDTEX  g3_land_texture[(256/G3_LANDTEXW)*(256/G3_LANDTEXW)*8];
  G3_VECTOR   g3_land_tyvertex[256];
  G3_VECTOR   g3_land_wtyvertex[256];
  G3_OBJECT   g3_llandobject;
  G3_OBJECT   g3_hlandobject;
  G3_LANDITEM *g3_ptrlanditem;
  SD          g3_landtime=0;



/**************************************************************************/
/*              LAND FUNCTIONS                                            */
/**************************************************************************/

void g3_land_texinit(void)
  {
    SD x,y;
    SD i,j,tmp;
    i=0;
    for(i=0;i<=(G3_LAND_MTEXNUM | G3_LAND_MTEXROT | G3_LAND_MTEXFLIP);i++)
    {
      //
      //    2+---------+3
      //     |         |
      //     |         |
      //     |         |
      //    1+---------+4
      //
      x=((i & G3_LAND_MTEXNUM)%(256/G3_LANDTEXW))*G3_LANDTEXW;
      y=((i & G3_LAND_MTEXNUM)/(256/G3_LANDTEXW))*G3_LANDTEXW;
      g3_land_texture[i].x1=((x)<<G3_GFPSHIFT)+(G3_GFPUNIT>>1);
      g3_land_texture[i].y1=((y+G3_LANDTEXW)<<G3_GFPSHIFT)-(G3_GFPUNIT>>1);
      g3_land_texture[i].x2=((x)<<G3_GFPSHIFT)+(G3_GFPUNIT>>1);
      g3_land_texture[i].y2=((y)<<G3_GFPSHIFT)+(G3_GFPUNIT>>1);
      g3_land_texture[i].x3=((x+G3_LANDTEXW)<<G3_GFPSHIFT)-(G3_GFPUNIT>>1);
      g3_land_texture[i].y3=((y)<<G3_GFPSHIFT)+(G3_GFPUNIT>>1);
      g3_land_texture[i].x4=((x+G3_LANDTEXW)<<G3_GFPSHIFT)-(G3_GFPUNIT>>1);
      g3_land_texture[i].y4=((y+G3_LANDTEXW)<<G3_GFPSHIFT)-(G3_GFPUNIT>>1);
      for(j=0;j<((i & G3_LAND_MTEXROT)>>G3_LAND_STEXROT);j++)
      {
        tmp=g3_land_texture[i].x4;
        g3_land_texture[i].x4=g3_land_texture[i].x3;
        g3_land_texture[i].x3=g3_land_texture[i].x2;
        g3_land_texture[i].x2=g3_land_texture[i].x1;
        g3_land_texture[i].x1=tmp;
        tmp=g3_land_texture[i].y4;
        g3_land_texture[i].y4=g3_land_texture[i].y3;
        g3_land_texture[i].y3=g3_land_texture[i].y2;
        g3_land_texture[i].y2=g3_land_texture[i].y1;
        g3_land_texture[i].y1=tmp;
      }
      if(i & G3_LAND_MTEXFLIP)
      {
        tmp=g3_land_texture[i].x1;
        g3_land_texture[i].x1=g3_land_texture[i].x4;
        g3_land_texture[i].x4=tmp;
        tmp=g3_land_texture[i].x2;
        g3_land_texture[i].x2=g3_land_texture[i].x3;
        g3_land_texture[i].x3=tmp;

        tmp=g3_land_texture[i].y1;
        g3_land_texture[i].y1=g3_land_texture[i].y4;
        g3_land_texture[i].y4=tmp;
        tmp=g3_land_texture[i].y2;
        g3_land_texture[i].y2=g3_land_texture[i].y3;
        g3_land_texture[i].y3=tmp;
      }
    }
  }



void g3_land_init(void)
  {
  g3_ptrlanditem=NULL;
  memset(&g3_llandobject,0,sizeof(G3_OBJECT));
  memset(&g3_hlandobject,0,sizeof(G3_OBJECT));
  g3_land_texinit();
  }


void g3_land_free(void)
  {
  if(g3_ptrlanditem!=NULL)          g3_free(g3_ptrlanditem);
  if(g3_llandobject.ptrvertex!=NULL) g3_free(g3_llandobject.ptrvertex);
  if(g3_llandobject.ptrpoly!=NULL)   g3_free(g3_llandobject.ptrpoly);
  if(g3_llandobject.ptrskl!=NULL)    g3_free(g3_llandobject.ptrskl);
  if(g3_hlandobject.ptrvertex!=NULL) g3_free(g3_hlandobject.ptrvertex);
  if(g3_hlandobject.ptrpoly!=NULL)   g3_free(g3_hlandobject.ptrpoly);
  if(g3_hlandobject.ptrskl!=NULL)    g3_free(g3_hlandobject.ptrskl);
  }

#if 0
void g3_land_adjust(void)
  {
  SD x,z;
  G3_LANDITEM *ptrland;
  G3_LANDITEM *nextptrland;
  G3_LANDITEM *prevptrland;
  for(x=0;x<G3_LANDSIZE+1;x++)
  {
    for(z=0;z<G3_LANDSIZE+1;z++)
    {
      ptrland=g3_land_getlanditemptr(x,z);
      if(ptrland->y<128)
      {
        prevptrland=g3_land_getlanditemptr(x+1,z);
        nextptrland=g3_land_getlanditemptr(x-1,z);
        if(nextptrland->y>128)
          ptrland->y=128;
        if(prevptrland->y>128)
          ptrland->y=128;


        prevptrland=g3_land_getlanditemptr(x,z-1);
        nextptrland=g3_land_getlanditemptr(x,z+1);
        if(nextptrland->y>128)
          ptrland->y=128;
        if(prevptrland->y>128)
          ptrland->y=128;

      }

    }
  }





  for(x=0;x<G3_LANDSIZE+1;x++)
  {
    for(z=0;z<G3_LANDSIZE+1;z++)
    {
      ptrland=g3_land_getlanditemptr(x,z);
      if(ptrland->y==128)
      {
        ptrland->y=129;
      }
      //ptrland->y=0;
    }
  }

  }
#endif



SD g3_land_load(char *filename)
  {
  static SD callcount=0;
  G3_FILE *handle;
  char string[_MAX_PATH];

  if(callcount==0)
    {
    g3_land_init();
    callcount++;
    }

  g3_land_free();
  sprintf(string,"%s.LND",filename);
  handle=g3_fopen(string,"rb");
  if(handle>NULL)
    {
    g3_ptrlanditem=(G3_LANDITEM *)(g3_malloc(sizeof(G3_LANDITEM)*G3_LANDSIZE*G3_LANDSIZE));
    if(g3_ptrlanditem!=NULL)
      {
      SD x,z;
      g3_fread(g3_ptrlanditem,sizeof(G3_LANDITEM),G3_LANDSIZE*G3_LANDSIZE,handle);
      for(x=0;x<G3_LANDSIZE;x++)
        for(z=0;z<G3_LANDSIZE;z++)
          {
          G3_LANDITEM *ptrland;  
          g3_land_computelight(x,z);
          ptrland=g3_land_getlanditemptr(x,z);
          //ptrland->y=0;
          
          }
      }
    g3_fclose(handle);
    }



  sprintf(string,"%sl",filename);
  g3_object_vtxload(string,&g3_llandobject);
  g3_object_envload(string,&g3_llandobject);
  g3_object_sklload(string,&g3_llandobject);
  sprintf(string,"%sh",filename);
  g3_object_vtxload(string,&g3_hlandobject);
  g3_object_envload(string,&g3_hlandobject);
  g3_object_sklload(string,&g3_hlandobject);
  if  (
      g3_ptrlanditem==NULL ||
      g3_llandobject.ptrvertex==NULL || g3_llandobject.ptrpoly==NULL || g3_llandobject.ptrskl==NULL ||
      g3_hlandobject.ptrvertex==NULL || g3_hlandobject.ptrpoly==NULL || g3_hlandobject.ptrskl==NULL
      )
    {
    g3_land_free();
    return(G3_FALSE);
    }
//  g3_land_adjust();
  return(G3_TRUE);
  }

SD g3_land_save(char *filename)
  {
  G3_FILE *handle;
  SD x,z;
  char string[_MAX_PATH];
  sprintf(string,"%s.LND",filename);
  handle=g3_fopen(string,"wb");
  if(handle==NULL) return(G3_FALSE);
  for(x=0;x<G3_LANDSIZE;x++)
    for(z=0;z<G3_LANDSIZE;z++)
      g3_land_computelight(x,z);
  g3_fwrite(g3_ptrlanditem,sizeof(G3_LANDITEM),G3_LANDSIZE*G3_LANDSIZE,handle);
  g3_fclose(handle);
  return(G3_TRUE);
  }


void g3_land_setbank(SD x,SD z,SD bank)
  {
  G3_LANDITEM *ptrland;
  ptrland=g3_ptrlanditem+(((z)&(G3_LANDSIZE-1))<<(G3_LANDSIZESHIFT)) + ((x)&(G3_LANDSIZE-1));
  ptrland->texture=(ptrland->texture & (0xFFFF-G3_LAND_MTEXBNK)) | ((bank<<G3_LAND_STEXBNK) & G3_LAND_MTEXBNK);
  }

SD g3_land_getbank(SD x,SD z)
  {
  G3_LANDITEM *ptrland;
  ptrland=g3_ptrlanditem+(((z)&(G3_LANDSIZE-1))<<(G3_LANDSIZESHIFT)) + ((x)&(G3_LANDSIZE-1));
  return((ptrland->texture & G3_LAND_MTEXBNK)>>G3_LAND_STEXBNK);
  }

void g3_land_settexture(SD x,SD z,SD texture)
  {
  G3_LANDITEM *ptrland;
  ptrland=g3_ptrlanditem+(((z)&(G3_LANDSIZE-1))<<(G3_LANDSIZESHIFT)) + ((x)&(G3_LANDSIZE-1));
  ptrland->texture=(ptrland->texture & (0xFFFF-G3_LAND_MTEXNUM)) | ((texture<<G3_LAND_STEXNUM) & G3_LAND_MTEXNUM);
  }

SD g3_land_gettexture(SD x,SD z)
  {
  G3_LANDITEM *ptrland;
  ptrland=g3_ptrlanditem+(((z)&(G3_LANDSIZE-1))<<(G3_LANDSIZESHIFT)) + ((x)&(G3_LANDSIZE-1));
  return((ptrland->texture & G3_LAND_MTEXNUM)>>G3_LAND_STEXNUM);
  }

void g3_land_setflip(SD x,SD z,SD flip)
  {
  G3_LANDITEM *ptrland;
  ptrland=g3_ptrlanditem+(((z)&(G3_LANDSIZE-1))<<(G3_LANDSIZESHIFT)) + ((x)&(G3_LANDSIZE-1));
  ptrland->texture=(ptrland->texture & (0xFFFF-G3_LAND_MTEXFLIP)) | ((flip<<G3_LAND_STEXFLIP) & G3_LAND_MTEXFLIP);
  }

SD g3_land_getflip(SD x,SD z)
  {
  G3_LANDITEM *ptrland;
  ptrland=g3_ptrlanditem+(((z)&(G3_LANDSIZE-1))<<(G3_LANDSIZESHIFT)) + ((x)&(G3_LANDSIZE-1));
  return((ptrland->texture & G3_LAND_MTEXFLIP)>>G3_LAND_STEXFLIP);
  }

void g3_land_setrot(SD x,SD z,SD rot)
  {
  G3_LANDITEM *ptrland;
  ptrland=g3_ptrlanditem+(((z)&(G3_LANDSIZE-1))<<(G3_LANDSIZESHIFT)) + ((x)&(G3_LANDSIZE-1));
  ptrland->texture=(ptrland->texture & (0xFFFF-G3_LAND_MTEXROT)) | ((rot<<G3_LAND_STEXROT) & G3_LAND_MTEXROT);
  }

SD g3_land_getrot(SD x,SD z)
  {
  G3_LANDITEM *ptrland;
  ptrland=g3_ptrlanditem+(((z)&(G3_LANDSIZE-1))<<(G3_LANDSIZESHIFT)) + ((x)&(G3_LANDSIZE-1));
  return((ptrland->texture & G3_LAND_MTEXROT)>>G3_LAND_STEXROT);
  }

void g3_land_sethbank(SD x,SD z,SD bank)
  {
  G3_LANDITEM *ptrland;
  ptrland=g3_ptrlanditem+(((z)&(G3_LANDSIZE-1))<<(G3_LANDSIZESHIFT)) + ((x)&(G3_LANDSIZE-1));
  ptrland->htexture=(ptrland->htexture & (0xFFFF-G3_LAND_MTEXBNK)) | ((bank<<G3_LAND_STEXBNK) & G3_LAND_MTEXBNK);
  }

SD g3_land_gethbank(SD x,SD z)
  {
  G3_LANDITEM *ptrland;
  ptrland=g3_ptrlanditem+(((z)&(G3_LANDSIZE-1))<<(G3_LANDSIZESHIFT)) + ((x)&(G3_LANDSIZE-1));
  return((ptrland->htexture & G3_LAND_MTEXBNK)>>G3_LAND_STEXBNK);
  }

void g3_land_sethtexture(SD x,SD z,SD texture)
  {
  G3_LANDITEM *ptrland;
  ptrland=g3_ptrlanditem+(((z)&(G3_LANDSIZE-1))<<(G3_LANDSIZESHIFT)) + ((x)&(G3_LANDSIZE-1));
  ptrland->htexture=(ptrland->htexture & (0xFFFF-G3_LAND_MTEXNUM)) | ((texture<<G3_LAND_STEXNUM) & G3_LAND_MTEXNUM);
  }

SD g3_land_gethtexture(SD x,SD z)
  {
  G3_LANDITEM *ptrland;
  ptrland=g3_ptrlanditem+(((z)&(G3_LANDSIZE-1))<<(G3_LANDSIZESHIFT)) + ((x)&(G3_LANDSIZE-1));
  return((ptrland->htexture & G3_LAND_MTEXNUM)>>G3_LAND_STEXNUM);
  }

void g3_land_sethflip(SD x,SD z,SD flip)
  {
  G3_LANDITEM *ptrland;
  ptrland=g3_ptrlanditem+(((z)&(G3_LANDSIZE-1))<<(G3_LANDSIZESHIFT)) + ((x)&(G3_LANDSIZE-1));
  ptrland->htexture=(ptrland->htexture & (0xFFFF-G3_LAND_MTEXFLIP)) | ((flip<<G3_LAND_STEXFLIP) & G3_LAND_MTEXFLIP);
  }

SD g3_land_gethflip(SD x,SD z)
  {
  G3_LANDITEM *ptrland;
  ptrland=g3_ptrlanditem+(((z)&(G3_LANDSIZE-1))<<(G3_LANDSIZESHIFT)) + ((x)&(G3_LANDSIZE-1));
  return((ptrland->htexture & G3_LAND_MTEXFLIP)>>G3_LAND_STEXFLIP);
  }

void g3_land_sethrot(SD x,SD z,SD rot)
  {
  G3_LANDITEM *ptrland;
  ptrland=g3_ptrlanditem+(((z)&(G3_LANDSIZE-1))<<(G3_LANDSIZESHIFT)) + ((x)&(G3_LANDSIZE-1));
  ptrland->htexture=(ptrland->htexture & (0xFFFF-G3_LAND_MTEXROT)) | ((rot<<G3_LAND_STEXROT) & G3_LAND_MTEXROT);
  }

SD g3_land_gethrot(SD x,SD z)
  {
  G3_LANDITEM *ptrland;
  ptrland=g3_ptrlanditem+(((z)&(G3_LANDSIZE-1))<<(G3_LANDSIZESHIFT)) + ((x)&(G3_LANDSIZE-1));
  return((ptrland->htexture & G3_LAND_MTEXROT)>>G3_LAND_STEXROT);
  }



void g3_land_compute(void)
  {
  G3_LANDITEM *ptrland;
  G3_LANDTEX *ptrtexture;
  SD xx,zz,x,z,bank;
  SD dist;
  SD wave;
  G3_POLYGON *ptrpoly;
  G3_MATRIX  tmpmatrix,matrix;
  G3_DVERTEX *ptrdvertex,*ptrdv;
  G3_DVERTEX *ptra,*ptrb,*ptrc,*ptrd;
  G3_VERTEX  *ptrvertex;
  G3_VECTOR  xvertex;
  G3_VECTOR  yvertex;
  G3_VECTOR  dyvertex;
  G3_VECTOR  wyvertex;
  G3_VECTOR  wdyvertex;
  G3_VECTOR  zvertex;
  G3_VECTOR  dvertex;
  G3_VECTOR  ddvertex;
  g3_landtime+=g3_timer_sampledelay();
  /* first compute vertex of the object */
  g3_matrix_copy(&g3_camera.matrix,&tmpmatrix);
  tmpmatrix.m41=(FP)((SD)(tmpmatrix.m41) & ((G3_LANDGRIDSTEP<<G3_FPSHIFT)-1) );
  tmpmatrix.m43=(FP)((SD)(tmpmatrix.m43) & ((G3_LANDGRIDSTEP<<G3_FPSHIFT)-1) );
  g3_matrix_invert(&tmpmatrix,&matrix);
  g3_matrix_xyzoom(&matrix);
  xx=((SD)(g3_camera.translation.x)>>(G3_FPSHIFT+G3_LANDGRIDSTEPSHIFT));
  zz=((SD)(g3_camera.translation.z)>>(G3_FPSHIFT+G3_LANDGRIDSTEPSHIFT));
  /* compute transform of vertex (0,1,0)  */
  dyvertex.x=(matrix.m21*(1<<G3_LANDYMULTIPLYSHIFT));
  dyvertex.y=(matrix.m22*(1<<G3_LANDYMULTIPLYSHIFT));
  dyvertex.z=(matrix.m23*(1<<G3_LANDYMULTIPLYSHIFT));
  wdyvertex.x=(matrix.m21/(1<<(8-G3_LANDWAVEAMPSHIFT)));
  wdyvertex.y=(matrix.m22/(1<<(8-G3_LANDWAVEAMPSHIFT)));
  wdyvertex.z=(matrix.m23/(1<<(8-G3_LANDWAVEAMPSHIFT)));
  /* compute transform of vertex (0,n,0)  */
  yvertex.x=0;
  yvertex.y=0;
  yvertex.z=0;
  wyvertex.x=0;
  wyvertex.y=0;
  wyvertex.z=0;
  for(x=0;x<256;x++)
    {
    g3_land_tyvertex[x].x=yvertex.x;
    g3_land_tyvertex[x].y=yvertex.y;
    g3_land_tyvertex[x].z=yvertex.z;
    yvertex.x+=dyvertex.x;
    yvertex.y+=dyvertex.y;
    yvertex.z+=dyvertex.z;
    g3_land_wtyvertex[x].x=wyvertex.x;
    g3_land_wtyvertex[x].y=wyvertex.y;
    g3_land_wtyvertex[x].z=wyvertex.z;
    wyvertex.x+=wdyvertex.x;
    wyvertex.y+=wdyvertex.y;
    wyvertex.z+=wdyvertex.z;
    }
  /* compute transform of vertex (GRIDSTEP,0,0)*/
  xvertex.x=(matrix.m11*(1<<G3_LANDGRIDSTEPSHIFT));
  xvertex.y=(matrix.m12*(1<<G3_LANDGRIDSTEPSHIFT));
  xvertex.z=(matrix.m13*(1<<G3_LANDGRIDSTEPSHIFT));
  /* compute transform of vertex (0,0,GRIDSTEP)*/
  zvertex.x=(matrix.m31*(1<<G3_LANDGRIDSTEPSHIFT));
  zvertex.y=(matrix.m32*(1<<G3_LANDGRIDSTEPSHIFT));
  zvertex.z=(matrix.m33*(1<<G3_LANDGRIDSTEPSHIFT));
  /* compute the first vertex of ground */
  ddvertex.x  =     (matrix.m11*(-G3_GRIDSIZE*G3_LANDGRIDSTEP/2)) +
                    (matrix.m31*(-G3_GRIDSIZE*G3_LANDGRIDSTEP/2)) +
                    (matrix.m41);
  ddvertex.y  =     (matrix.m12*(-G3_GRIDSIZE*G3_LANDGRIDSTEP/2)) +
                    (matrix.m32*(-G3_GRIDSIZE*G3_LANDGRIDSTEP/2)) +
                    (matrix.m42);
  ddvertex.z  =     (matrix.m13*(-G3_GRIDSIZE*G3_LANDGRIDSTEP/2)) +
                    (matrix.m33*(-G3_GRIDSIZE*G3_LANDGRIDSTEP/2)) +
                    (matrix.m43);
  //******************************************************************
  //                     UNDER THE SEA
  //******************************************************************
  if(g3_camera.translation.y<=G3_FPVALUE(G3_LANDWAVEHEIGHT))
    {
    //------------------------------------------------------------------
    //                     UNDER THE SEA LEVEL VERTEX
    //------------------------------------------------------------------
    /* compute the first vertex of ground */
    ddvertex.x  =     (matrix.m11*(-G3_GRIDSIZE*G3_LANDGRIDSTEP/2)) +
                      (matrix.m31*(-G3_GRIDSIZE*G3_LANDGRIDSTEP/2)) +
                      (matrix.m41);
    ddvertex.y  =     (matrix.m12*(-G3_GRIDSIZE*G3_LANDGRIDSTEP/2)) +
                      (matrix.m32*(-G3_GRIDSIZE*G3_LANDGRIDSTEP/2)) +
                      (matrix.m42);
    ddvertex.z  =     (matrix.m13*(-G3_GRIDSIZE*G3_LANDGRIDSTEP/2)) +
                      (matrix.m33*(-G3_GRIDSIZE*G3_LANDGRIDSTEP/2)) +
                      (matrix.m43);
    ptrdvertex=g3_objectdvertex;

    for(z=-G3_GRIDSIZE/2;z<=G3_GRIDSIZE/2;z++)
      {
      dvertex=ddvertex;
      for(x=-G3_GRIDSIZE/2;x<=G3_GRIDSIZE/2;x++)
        {
        ptrland=g3_ptrlanditem + (((zz+z)&(G3_LANDSIZE-1))<<(G3_LANDSIZESHIFT)) + ((xx+x)&(G3_LANDSIZE-1));
        //-------------------------------------
        // Vertex of ground under the sea
        //-------------------------------------

        // wave=0 a 255
        wave=abs(g3_sincos[((g3_landtime<<3)+((zz+z)*(G3_NBANGLE/4)))&(G3_NBANGLE-1)].sin+g3_sincos[ ((g3_landtime<<2)+((xx+x)*(G3_NBANGLE/8))) & (G3_NBANGLE-1)].cos)>>(G3_FPSHIFT+1-8);

        // compute vertex
        ptrdvertex->height=(SD)((UB)ptrland->y);
        if(ptrdvertex->height>128) ptrdvertex->height=128;
        ptrdvertex->x=dvertex.x+g3_land_tyvertex[ptrdvertex->height].x;
        ptrdvertex->y=dvertex.y+g3_land_tyvertex[ptrdvertex->height].y;
        ptrdvertex->z=dvertex.z+g3_land_tyvertex[ptrdvertex->height].z;
        G3_PTRDVERTEX_PROJECT(ptrdvertex);
        if(ptrland->hlight)
        {
        ptrdvertex->light=(SD)(ptrland->hlight);
        }
        else
        if(ptrdvertex->z>g3_fogplane)
        {
          ptrdvertex->light=(G3_MAXLIGHTING-(G3_MAXLIGHTING/4)-1)-((ptrdvertex->z-g3_fogplane)/(g3_fogdepth/(G3_MAXLIGHTING/2)))-(wave>>4);
          if(ptrdvertex->light<=(G3_MAXLIGHTING/4)) ptrdvertex->light=(G3_MAXLIGHTING/4)+1;
        }
        else
        {
          ptrdvertex->light=(G3_MAXLIGHTING-(G3_MAXLIGHTING/4)-1)-(wave>>4)+(ptrland->light);
        }

        //-------------------------------------
        // vertex of surface under the sea
        //-------------------------------------
        ptrdvertex+=(G3_GRIDSIZE+1)*(G3_GRIDSIZE+1);
        ptrdvertex->x=dvertex.x+g3_land_tyvertex[128].x-(g3_land_wtyvertex[wave].x);
        ptrdvertex->y=dvertex.y+g3_land_tyvertex[128].y-(g3_land_wtyvertex[wave].y);
        ptrdvertex->z=dvertex.z+g3_land_tyvertex[128].z-(g3_land_wtyvertex[wave].z);
        G3_PTRDVERTEX_PROJECT(ptrdvertex);
        if(ptrland->hlight)
        {
        ptrdvertex->light=(SD)(ptrland->hlight);
        }
        else
        if(ptrdvertex->z>g3_fogplane)
        {
          ptrdvertex->light=(G3_MAXLIGHTING-(G3_MAXLIGHTING/4)-1)-((ptrdvertex->z-g3_fogplane)/(g3_fogdepth/(G3_MAXLIGHTING/2)))-(wave>>3);
          if(ptrdvertex->light<=(G3_MAXLIGHTING/4)) ptrdvertex->light=(G3_MAXLIGHTING/4)+1;
        }
        else
        {
          ptrdvertex->light=(G3_MAXLIGHTING-(G3_MAXLIGHTING/4)-1)-(wave>>3);
        }
        // next vertex please...
        ptrdvertex-=(G3_GRIDSIZE+1)*(G3_GRIDSIZE+1);
        ptrdvertex++;
        dvertex.x+=xvertex.x;
        dvertex.y+=xvertex.y;
        dvertex.z+=xvertex.z;
        }
      ddvertex.x+=zvertex.x;
      ddvertex.y+=zvertex.y;
      ddvertex.z+=zvertex.z;
      }
    //------------------------------------------------------------------
    //                     UNDER THE SEA LEVEL POLYGONS
    //------------------------------------------------------------------
    ptrdv=g3_objectdvertex;
    for(z=-G3_GRIDSIZE/2;z<G3_GRIDSIZE/2;z++)
      {
      for(x=-G3_GRIDSIZE/2;x<G3_GRIDSIZE/2;x++)
        {
        //--------------------------------
        // POLYGONS OF SEA GROUND
        //--------------------------------
        ptrland=g3_ptrlanditem + (((zz+z)&(G3_LANDSIZE-1))<<(G3_LANDSIZESHIFT)) + ((xx+x)&(G3_LANDSIZE-1));
        ptrtexture=&g3_land_texture[ptrland->texture & (G3_LAND_MTEXNUM|G3_LAND_MTEXROT|G3_LAND_MTEXFLIP)];
        bank=(ptrland->texture & G3_LAND_MTEXBNK)>>G3_LAND_STEXBNK;
        ptra=ptrdv;
        ptrb=ptrdv+(G3_GRIDSIZE+1);
        ptrc=ptrdv+((G3_GRIDSIZE+1)+1);
        ptrd=ptrdv+1;
        if(ptrb->height==ptrd->height)
            {
            //           B+----+C
            //            |   /|
            //            |  / |
            //            | /  |
            //            |/   |
            //           A+----+D
            dist=g3_pvmaxdistance(ptra,ptrb,ptrc);
            if(dist>0)
              {
              dist>>=G3_FPSHIFT;
        dist+=G3_LANDGRIDSTEP;
              g3_ptrsubpoly->type=G3_GRAPHCALL_LPOLYTG;
              g3_ptrsubpoly->pts1=(UD)((UB *)ptra-(UB *)g3_objectdvertex);
              g3_ptrsubpoly->pts2=(UD)((UB *)ptrb-(UB *)g3_objectdvertex);
              g3_ptrsubpoly->pts3=(UD)((UB *)ptrc-(UB *)g3_objectdvertex);
              #ifdef G3_EDITOR
              g3_ptrsubpoly->color=(((zz+z)&(G3_LANDSIZE-1))*G3_LANDSIZE) + ((xx+x)&(G3_LANDSIZE-1));
              #endif
              g3_ptrsubpoly->bank=bank;
              g3_ptrsubpoly->xa=ptrtexture->x1;
              g3_ptrsubpoly->ya=ptrtexture->y1;
              g3_ptrsubpoly->xb=ptrtexture->x2;
              g3_ptrsubpoly->yb=ptrtexture->y2;
              g3_ptrsubpoly->xc=ptrtexture->x3;
              g3_ptrsubpoly->yc=ptrtexture->y3;
              G3_OT_ADD(dist,g3_ptrsubpoly);
              g3_ptrsubpoly++;
              }
            dist=g3_pvmaxdistance(ptra,ptrc,ptrd);
            if(dist>0)
              {
              dist>>=G3_FPSHIFT;
        dist+=G3_LANDGRIDSTEP;
              g3_ptrsubpoly->type=G3_GRAPHCALL_LPOLYTG;
              g3_ptrsubpoly->pts1=(UD)((UB *)ptra-(UB *)g3_objectdvertex);
              g3_ptrsubpoly->pts2=(UD)((UB *)ptrc-(UB *)g3_objectdvertex);
              g3_ptrsubpoly->pts3=(UD)((UB *)ptrd-(UB *)g3_objectdvertex);

              #ifdef G3_EDITOR
              g3_ptrsubpoly->color=(((zz+z)&(G3_LANDSIZE-1))*G3_LANDSIZE) + ((xx+x)&(G3_LANDSIZE-1));
              #endif
              g3_ptrsubpoly->bank=bank;
              g3_ptrsubpoly->xa=ptrtexture->x1;
              g3_ptrsubpoly->ya=ptrtexture->y1;
              g3_ptrsubpoly->xb=ptrtexture->x3;
              g3_ptrsubpoly->yb=ptrtexture->y3;
              g3_ptrsubpoly->xc=ptrtexture->x4;
              g3_ptrsubpoly->yc=ptrtexture->y4;
              G3_OT_ADD(dist,g3_ptrsubpoly);
              g3_ptrsubpoly++;
              }
            }
          else
            {
            //           B+----+C
            //            |\   |
            //            | \  |
            //            |  \ |
            //            |   \|
            //           A+----+D
            dist=g3_pvmaxdistance(ptra,ptrb,ptrd);
            if(dist>0)
              {
              dist>>=G3_FPSHIFT;
        dist+=G3_LANDGRIDSTEP;
              g3_ptrsubpoly->type=G3_GRAPHCALL_LPOLYTG;
              g3_ptrsubpoly->pts1=(UD)((UB *)ptra-(UB *)g3_objectdvertex);
              g3_ptrsubpoly->pts2=(UD)((UB *)ptrb-(UB *)g3_objectdvertex);
              g3_ptrsubpoly->pts3=(UD)((UB *)ptrd-(UB *)g3_objectdvertex);

              #ifdef G3_EDITOR
              g3_ptrsubpoly->color=(((zz+z)&(G3_LANDSIZE-1))*G3_LANDSIZE) + ((xx+x)&(G3_LANDSIZE-1));
              #endif
              g3_ptrsubpoly->bank=bank;
              g3_ptrsubpoly->xa=ptrtexture->x1;
              g3_ptrsubpoly->ya=ptrtexture->y1;
              g3_ptrsubpoly->xb=ptrtexture->x2;
              g3_ptrsubpoly->yb=ptrtexture->y2;
              g3_ptrsubpoly->xc=ptrtexture->x4;
              g3_ptrsubpoly->yc=ptrtexture->y4;
              G3_OT_ADD(dist,g3_ptrsubpoly);
              g3_ptrsubpoly++;
              }
            dist=g3_pvmaxdistance(ptrd,ptrb,ptrc);
            if(dist>0)
              {
              dist>>=G3_FPSHIFT;
        dist+=G3_LANDGRIDSTEP;
              g3_ptrsubpoly->type=G3_GRAPHCALL_LPOLYTG;
              g3_ptrsubpoly->pts1=(UD)((UB *)ptrd-(UB *)g3_objectdvertex);
              g3_ptrsubpoly->pts2=(UD)((UB *)ptrb-(UB *)g3_objectdvertex);
              g3_ptrsubpoly->pts3=(UD)((UB *)ptrc-(UB *)g3_objectdvertex);

              #ifdef G3_EDITOR
              g3_ptrsubpoly->color=(((zz+z)&(G3_LANDSIZE-1))*G3_LANDSIZE) + ((xx+x)&(G3_LANDSIZE-1));
              #endif
              g3_ptrsubpoly->bank=bank;
              g3_ptrsubpoly->xa=ptrtexture->x4;
              g3_ptrsubpoly->ya=ptrtexture->y4;
              g3_ptrsubpoly->xb=ptrtexture->x2;
              g3_ptrsubpoly->yb=ptrtexture->y2;
              g3_ptrsubpoly->xc=ptrtexture->x3;
              g3_ptrsubpoly->yc=ptrtexture->y3;
              G3_OT_ADD(dist,g3_ptrsubpoly);
              g3_ptrsubpoly++;
              }
            }
        //------------------------------
        // POLYGONS OF SURFACE
        //------------------------------
        ptrtexture=&g3_land_texture[ptrland->htexture & (G3_LAND_MTEXNUM|G3_LAND_MTEXROT|G3_LAND_MTEXFLIP)];
        bank=(ptrland->htexture & G3_LAND_MTEXBNK)>>G3_LAND_STEXBNK;
        ptrdv+=(G3_GRIDSIZE+1)*(G3_GRIDSIZE+1);
        ptra=ptrdv;
        ptrb=ptrdv+(G3_GRIDSIZE+1);
        ptrc=ptrdv+((G3_GRIDSIZE+1)+1);
        ptrd=ptrdv+1;
        //           B+----+C
        //            |   /|
        //            |  / |
        //            | /  |
        //            |/   |
        //           A+----+D
        dist=g3_pvmaxdistance(ptra,ptrc,ptrb);
        if(dist>0)
          {
          dist>>=G3_FPSHIFT;
      dist+=G3_LANDGRIDSTEP;
          g3_ptrsubpoly->type=G3_GRAPHCALL_LPOLYTG;
          g3_ptrsubpoly->pts1=(UD)((UB *)ptra-(UB *)g3_objectdvertex);
          g3_ptrsubpoly->pts2=(UD)((UB *)ptrc-(UB *)g3_objectdvertex);
          g3_ptrsubpoly->pts3=(UD)((UB *)ptrb-(UB *)g3_objectdvertex);

          #ifdef G3_EDITOR
          g3_ptrsubpoly->color=(((zz+z)&(G3_LANDSIZE-1))*G3_LANDSIZE) + ((xx+x)&(G3_LANDSIZE-1));
          g3_ptrsubpoly->color|=0x80000000;
          #endif
          g3_ptrsubpoly->bank=bank;
          g3_ptrsubpoly->xa=ptrtexture->x1;
          g3_ptrsubpoly->ya=ptrtexture->y1;
          g3_ptrsubpoly->xb=ptrtexture->x3;
          g3_ptrsubpoly->yb=ptrtexture->y3;
          g3_ptrsubpoly->xc=ptrtexture->x2;
          g3_ptrsubpoly->yc=ptrtexture->y2;
          G3_OT_ADD(dist,g3_ptrsubpoly);
          g3_ptrsubpoly++;
          }
        dist=g3_pvmaxdistance(ptra,ptrd,ptrc);
        if(dist>0)
          {
          dist>>=G3_FPSHIFT;
      dist+=G3_LANDGRIDSTEP;
          g3_ptrsubpoly->type=G3_GRAPHCALL_LPOLYTG;
          g3_ptrsubpoly->pts1=(UD)((UB *)ptra-(UB *)g3_objectdvertex);
          g3_ptrsubpoly->pts2=(UD)((UB *)ptrd-(UB *)g3_objectdvertex);
          g3_ptrsubpoly->pts3=(UD)((UB *)ptrc-(UB *)g3_objectdvertex);

          #ifdef G3_EDITOR
          g3_ptrsubpoly->color=(((zz+z)&(G3_LANDSIZE-1))*G3_LANDSIZE) + ((xx+x)&(G3_LANDSIZE-1));
          g3_ptrsubpoly->color|=0x80000000;
          #endif
          g3_ptrsubpoly->bank=bank;
          g3_ptrsubpoly->xa=ptrtexture->x1;
          g3_ptrsubpoly->ya=ptrtexture->y1;
          g3_ptrsubpoly->xb=ptrtexture->x4;
          g3_ptrsubpoly->yb=ptrtexture->y4;
          g3_ptrsubpoly->xc=ptrtexture->x3;
          g3_ptrsubpoly->yc=ptrtexture->y3;
          G3_OT_ADD(dist,g3_ptrsubpoly);
          g3_ptrsubpoly++;
          }
        ptrdv-=(G3_GRIDSIZE+1)*(G3_GRIDSIZE+1);
        ptrdv++;
        }
      ptrdv++;
      }
    g3_objectdvertex=ptrdvertex+((G3_GRIDSIZE+1)*(G3_GRIDSIZE+1));
    }


  //******************************************************************
  //                     OVER THE SEA
  //******************************************************************
  if(g3_camera.translation.y>=G3_FPVALUE(G3_LANDWAVEHEIGHT-G3_LANDWAVEAMP))
    {
    //------------------------------------------------------------------
    //                     OVER THE SEA VERTEX
    //------------------------------------------------------------------
    /* compute the first vertex of ground */
    ddvertex.x  =     (matrix.m11*(-G3_GRIDSIZE*G3_LANDGRIDSTEP/2)) +
                      (matrix.m31*(-G3_GRIDSIZE*G3_LANDGRIDSTEP/2)) +
                      (matrix.m41);
    ddvertex.y  =     (matrix.m12*(-G3_GRIDSIZE*G3_LANDGRIDSTEP/2)) +
                      (matrix.m32*(-G3_GRIDSIZE*G3_LANDGRIDSTEP/2)) +
                      (matrix.m42);
    ddvertex.z  =     (matrix.m13*(-G3_GRIDSIZE*G3_LANDGRIDSTEP/2)) +
                      (matrix.m33*(-G3_GRIDSIZE*G3_LANDGRIDSTEP/2)) +
                      (matrix.m43);
    ptrdvertex=g3_objectdvertex;
    for(z=-G3_GRIDSIZE/2;z<=G3_GRIDSIZE/2;z++)
      {
      dvertex=ddvertex;
      for(x=-G3_GRIDSIZE/2;x<=G3_GRIDSIZE/2;x++)
        {
        ptrland=g3_ptrlanditem + (((zz+z)&(G3_LANDSIZE-1))<<(G3_LANDSIZESHIFT)) + ((xx+x)&(G3_LANDSIZE-1));
        // Vertex of ground
        ptrdvertex->height=(SD)((UB)ptrland->y);
        if(ptrdvertex->height<128) ptrdvertex->height=128;
        if(ptrdvertex->height==128)
          {
          wave=abs(g3_sincos[((g3_landtime<<3)+((zz+z)*G3_NBANGLE/4))&(G3_NBANGLE-1)].sin+g3_sincos[ ((g3_landtime<<2)+((xx+x)*G3_NBANGLE/8)) & (G3_NBANGLE-1)].cos)>>(G3_FPSHIFT+1-8);
          if(wave<0 || wave>255)
          {
        g3_error("G3_LAND_COMPUTE : wave out of range");
          }

          ptrdvertex->x=dvertex.x+g3_land_tyvertex[128].x-(g3_land_wtyvertex[wave].x);
          ptrdvertex->y=dvertex.y+g3_land_tyvertex[128].y-(g3_land_wtyvertex[wave].y);
          ptrdvertex->z=dvertex.z+g3_land_tyvertex[128].z-(g3_land_wtyvertex[wave].z);
          }
        else
          {
          wave=0;
          ptrdvertex->x=dvertex.x+g3_land_tyvertex[ptrdvertex->height].x;
          ptrdvertex->y=dvertex.y+g3_land_tyvertex[ptrdvertex->height].y;
          ptrdvertex->z=dvertex.z+g3_land_tyvertex[ptrdvertex->height].z;
          }
        G3_PTRDVERTEX_PROJECT(ptrdvertex);
        if(ptrland->hlight)
        {
        ptrdvertex->light=G3_MAXLIGHTING+(SD)(ptrland->hlight);
        }
        else
        if(ptrdvertex->z>g3_fogplane)
        {
          ptrdvertex->light=G3_MAXLIGHTING+(G3_MAXLIGHTING-(G3_MAXLIGHTING/4)-1)-((ptrdvertex->z-g3_fogplane)/(g3_fogdepth/(G3_MAXLIGHTING/2)))-(wave>>3);
          if(ptrdvertex->light<=G3_MAXLIGHTING+(G3_MAXLIGHTING/4)) ptrdvertex->light=G3_MAXLIGHTING+(G3_MAXLIGHTING/4)+1;
        }
        else
        {
          ptrdvertex->light=G3_MAXLIGHTING+(G3_MAXLIGHTING-(G3_MAXLIGHTING/4)-1)-(wave>>3);
        }
        ptrdvertex++;
        dvertex.x+=xvertex.x;
        dvertex.y+=xvertex.y;
        dvertex.z+=xvertex.z;
        }
      ddvertex.x+=zvertex.x;
      ddvertex.y+=zvertex.y;
      ddvertex.z+=zvertex.z;
      }
    ptrdv=g3_objectdvertex;
    for(z=-G3_GRIDSIZE/2;z<G3_GRIDSIZE/2;z++)
      {
      for(x=-G3_GRIDSIZE/2;x<G3_GRIDSIZE/2;x++)
        {
        //--------------------------------
        // POLYGONS OF GROUND
        //--------------------------------
        ptrland=g3_ptrlanditem + (((zz+z)&(G3_LANDSIZE-1))<<(G3_LANDSIZESHIFT)) + ((xx+x)&(G3_LANDSIZE-1));
        ptrtexture=&g3_land_texture[ptrland->htexture & (G3_LAND_MTEXNUM|G3_LAND_MTEXROT|G3_LAND_MTEXFLIP)];
        bank=(ptrland->htexture & G3_LAND_MTEXBNK)>>G3_LAND_STEXBNK;
        ptra=ptrdv;
        ptrb=ptrdv+(G3_GRIDSIZE+1);
        ptrc=ptrdv+((G3_GRIDSIZE+1)+1);
        ptrd=ptrdv+1;
          if(ptrb->height==ptrd->height)
            {
            //           B+----+C
            //            |   /|
            //            |  / |
            //            | /  |
            //            |/   |
            //           A+----+D
            dist=g3_pvmaxdistance(ptra,ptrb,ptrc);
            if(dist>0)
              {
              dist>>=G3_FPSHIFT;
        dist+=G3_LANDGRIDSTEP;
              g3_ptrsubpoly->type=G3_GRAPHCALL_LPOLYTG;
              g3_ptrsubpoly->pts1=(UD)((UB *)ptra-(UB *)g3_objectdvertex);
              g3_ptrsubpoly->pts2=(UD)((UB *)ptrb-(UB *)g3_objectdvertex);
              g3_ptrsubpoly->pts3=(UD)((UB *)ptrc-(UB *)g3_objectdvertex);

              #ifdef G3_EDITOR
              g3_ptrsubpoly->color=(((zz+z)&(G3_LANDSIZE-1))*G3_LANDSIZE) + ((xx+x)&(G3_LANDSIZE-1));
              g3_ptrsubpoly->color|=0x80000000;
              #endif
              g3_ptrsubpoly->bank=bank;
              g3_ptrsubpoly->xa=ptrtexture->x1;
              g3_ptrsubpoly->ya=ptrtexture->y1;
              g3_ptrsubpoly->xb=ptrtexture->x2;
              g3_ptrsubpoly->yb=ptrtexture->y2;
              g3_ptrsubpoly->xc=ptrtexture->x3;
              g3_ptrsubpoly->yc=ptrtexture->y3;
              G3_OT_ADD(dist,g3_ptrsubpoly);
              g3_ptrsubpoly++;
              }
            dist=g3_pvmaxdistance(ptra,ptrc,ptrd);
            if(dist>0)
              {
              dist>>=G3_FPSHIFT;
        dist+=G3_LANDGRIDSTEP;
              g3_ptrsubpoly->type=G3_GRAPHCALL_LPOLYTG;
              g3_ptrsubpoly->pts1=(UD)((UB *)ptra-(UB *)g3_objectdvertex);
              g3_ptrsubpoly->pts2=(UD)((UB *)ptrc-(UB *)g3_objectdvertex);
              g3_ptrsubpoly->pts3=(UD)((UB *)ptrd-(UB *)g3_objectdvertex);

              #ifdef G3_EDITOR
              g3_ptrsubpoly->color=(((zz+z)&(G3_LANDSIZE-1))*G3_LANDSIZE) + ((xx+x)&(G3_LANDSIZE-1));
              g3_ptrsubpoly->color|=0x80000000;
              #endif
              g3_ptrsubpoly->bank=bank;
              g3_ptrsubpoly->xa=ptrtexture->x1;
              g3_ptrsubpoly->ya=ptrtexture->y1;
              g3_ptrsubpoly->xb=ptrtexture->x3;
              g3_ptrsubpoly->yb=ptrtexture->y3;
              g3_ptrsubpoly->xc=ptrtexture->x4;
              g3_ptrsubpoly->yc=ptrtexture->y4;
              G3_OT_ADD(dist,g3_ptrsubpoly);
              g3_ptrsubpoly++;
              }
            }
          else
            {
            //           B+----+C
            //            |\   |
            //            | \  |
            //            |  \ |
            //            |   \|
            //           A+----+D
            dist=g3_pvmaxdistance(ptra,ptrb,ptrd);
            if(dist>0)
              {
              dist>>=G3_FPSHIFT;
        dist+=G3_LANDGRIDSTEP;
              g3_ptrsubpoly->type=G3_GRAPHCALL_LPOLYTG;
              g3_ptrsubpoly->pts1=(UD)((UB *)ptra-(UB *)g3_objectdvertex);
              g3_ptrsubpoly->pts2=(UD)((UB *)ptrb-(UB *)g3_objectdvertex);
              g3_ptrsubpoly->pts3=(UD)((UB *)ptrd-(UB *)g3_objectdvertex);

              #ifdef G3_EDITOR
              g3_ptrsubpoly->color=(((zz+z)&(G3_LANDSIZE-1))*G3_LANDSIZE) + ((xx+x)&(G3_LANDSIZE-1));
              g3_ptrsubpoly->color|=0x80000000;
              #endif
              g3_ptrsubpoly->bank=bank;
              g3_ptrsubpoly->xa=ptrtexture->x1;
              g3_ptrsubpoly->ya=ptrtexture->y1;
              g3_ptrsubpoly->xb=ptrtexture->x2;
              g3_ptrsubpoly->yb=ptrtexture->y2;
              g3_ptrsubpoly->xc=ptrtexture->x4;
              g3_ptrsubpoly->yc=ptrtexture->y4;
              G3_OT_ADD(dist,g3_ptrsubpoly);
              g3_ptrsubpoly++;
              }
            dist=g3_pvmaxdistance(ptrd,ptrb,ptrc);
            if(dist>0)
              {
              dist>>=G3_FPSHIFT;
        dist+=G3_LANDGRIDSTEP;
              g3_ptrsubpoly->type=G3_GRAPHCALL_LPOLYTG;
              g3_ptrsubpoly->pts1=(UD)((UB *)ptrd-(UB *)g3_objectdvertex);
              g3_ptrsubpoly->pts2=(UD)((UB *)ptrb-(UB *)g3_objectdvertex);
              g3_ptrsubpoly->pts3=(UD)((UB *)ptrc-(UB *)g3_objectdvertex);

              #ifdef G3_EDITOR
              g3_ptrsubpoly->color=(((zz+z)&(G3_LANDSIZE-1))*G3_LANDSIZE) + ((xx+x)&(G3_LANDSIZE-1));
              g3_ptrsubpoly->color|=0x80000000;
              #endif
              g3_ptrsubpoly->bank=bank;
              g3_ptrsubpoly->xa=ptrtexture->x4;
              g3_ptrsubpoly->ya=ptrtexture->y4;
              g3_ptrsubpoly->xb=ptrtexture->x2;
              g3_ptrsubpoly->yb=ptrtexture->y2;
              g3_ptrsubpoly->xc=ptrtexture->x3;
              g3_ptrsubpoly->yc=ptrtexture->y3;
              G3_OT_ADD(dist,g3_ptrsubpoly);
              g3_ptrsubpoly++;
              }
            }
        ptrdv++;
        }
      ptrdv++;
      }
      g3_objectdvertex=ptrdvertex;
    }

  g3_matrix_copy(&g3_camera.matrix,&tmpmatrix);

  tmpmatrix.m41=(FP)((SD)(tmpmatrix.m41) & ((G3_LANDGRIDSTEP<<G3_FPSHIFT)-1) );
  tmpmatrix.m43=(FP)((SD)(tmpmatrix.m43) & ((G3_LANDGRIDSTEP<<G3_FPSHIFT)-1) );
  tmpmatrix.m41 &=G3_FPVALUE(G3_LANDGRIDSTEP)-1;
  tmpmatrix.m43 &=G3_FPVALUE(G3_LANDGRIDSTEP)-1;

  if(g3_camera.translation.y>=G3_FPVALUE(G3_LANDWAVEHEIGHT))
  {
  tmpmatrix.m42 += (G3_LANDGRIDSTEP<<G3_FPSHIFT);
  }
  else
  {
  tmpmatrix.m42 += (tmpmatrix.m42 - (128<<(G3_LANDYMULTIPLYSHIFT+G3_FPSHIFT))) >> (G3_LANDYMULTIPLYSHIFT+8-G3_LANDGRIDSTEPSHIFT);
  }
  g3_matrix_invert(&tmpmatrix,&matrix);
  g3_matrix_xyzoom(&matrix);

  if(g3_camera.translation.y<=G3_FPVALUE(G3_LANDWAVEHEIGHT))
    {
    ptrdvertex=g3_objectdvertex;
    ptrvertex=g3_llandobject.ptrvertex;
    ptrpoly=g3_llandobject.ptrpoly;
    while(ptrvertex->flag!=G3_ENDMARK)
      {
      ptrdvertex->x  =  (matrix.m11*ptrvertex->x) +
                        (matrix.m21*ptrvertex->y) +
                        (matrix.m31*ptrvertex->z) +
                        (matrix.m41);
      ptrdvertex->y  =  (matrix.m12*ptrvertex->x) +
                        (matrix.m22*ptrvertex->y) +
                        (matrix.m32*ptrvertex->z) +
                        (matrix.m42);
      ptrdvertex->z  =  (matrix.m13*ptrvertex->x) +
                        (matrix.m23*ptrvertex->y) +
                        (matrix.m33*ptrvertex->z) +
                        (matrix.m43);
      G3_PTRDVERTEX_PROJECT(ptrdvertex);
      ptrvertex++;
      ptrdvertex++;
      }
    while(ptrpoly->type!=G3_ENDMARK)
      {
      ptra=G3_OBJECT_DVERTEXADR(g3_objectdvertex,ptrpoly->pts1);     /* read a point       */
      ptrb=G3_OBJECT_DVERTEXADR(g3_objectdvertex,ptrpoly->pts2);     /* read b point       */
      ptrc=G3_OBJECT_DVERTEXADR(g3_objectdvertex,ptrpoly->pts3);     /* read c point       */

      dist=g3_pvmaxdistance(ptra,ptrb,ptrc);
      if(dist)
        {
        G3_OT_ADDEND(ptrpoly);
        }
      ptrpoly++;
      }
    g3_objectdvertex=ptrdvertex;
    }

  if(g3_camera.translation.y>=G3_FPVALUE(G3_LANDWAVEHEIGHT-G3_LANDWAVEAMP))
    {
    ptrdvertex=g3_objectdvertex;
    ptrvertex=g3_hlandobject.ptrvertex;
    ptrpoly=g3_hlandobject.ptrpoly;
    while(ptrvertex->flag!=G3_ENDMARK)
      {
      ptrdvertex->x  =  (matrix.m11*ptrvertex->x) +
                        (matrix.m21*ptrvertex->y) +
                        (matrix.m31*ptrvertex->z) +
                        (matrix.m41);
      ptrdvertex->y  =  (matrix.m12*ptrvertex->x) +
                        (matrix.m22*ptrvertex->y) +
                        (matrix.m32*ptrvertex->z) +
                        (matrix.m42);
      ptrdvertex->z  =  (matrix.m13*ptrvertex->x) +
                        (matrix.m23*ptrvertex->y) +
                        (matrix.m33*ptrvertex->z) +
                        (matrix.m43);
      G3_PTRDVERTEX_PROJECT(ptrdvertex);
      ptrvertex++;
      ptrdvertex++;
      }
    while(ptrpoly->type!=G3_ENDMARK)
      {
      ptra=G3_OBJECT_DVERTEXADR(g3_objectdvertex,ptrpoly->pts1);     /* read a point       */
      ptrb=G3_OBJECT_DVERTEXADR(g3_objectdvertex,ptrpoly->pts2);     /* read b point       */
      ptrc=G3_OBJECT_DVERTEXADR(g3_objectdvertex,ptrpoly->pts3);     /* read c point       */

      dist=g3_pvmaxdistance(ptra,ptrb,ptrc);
      if(dist)
        {
        dist=G3_FPMAX;
        G3_OT_ADD(dist,ptrpoly);
        }
      ptrpoly++;
      }
    g3_objectdvertex=ptrdvertex;
    }
  }

SD g3_land_gety(SD x,SD z)
  {
  G3_LANDITEM *ptrland;
  SD xx,zz,rx,rz;
  SD ya,yb,yc,yd;
  SD vyonx,vyonz;
  //           B+----+C
  //            |    |
  //            |    |
  //            |    |
  //            |    |
  //           A+----+D
  xx=(x>>(G3_FPSHIFT+G3_LANDGRIDSTEPSHIFT)) & (G3_LANDSIZE-1);
  zz=(z>>(G3_FPSHIFT+G3_LANDGRIDSTEPSHIFT)) & (G3_LANDSIZE-1);
  rx=(x & ((1<<(G3_FPSHIFT+G3_LANDGRIDSTEPSHIFT))-1));
  rz=(z & ((1<<(G3_FPSHIFT+G3_LANDGRIDSTEPSHIFT))-1));
  ptrland=g3_ptrlanditem + (((zz)&(G3_LANDSIZE-1))<<(G3_LANDSIZESHIFT)) + ((xx)&(G3_LANDSIZE-1));
  ya=(SD)((UB)(ptrland->y));
  ptrland=g3_ptrlanditem + (((zz+1)&(G3_LANDSIZE-1))<<(G3_LANDSIZESHIFT)) + ((xx)&(G3_LANDSIZE-1));
  yb=(SD)((UB)(ptrland->y));
  ptrland=g3_ptrlanditem + (((zz+1)&(G3_LANDSIZE-1))<<(G3_LANDSIZESHIFT)) + ((xx+1)&(G3_LANDSIZE-1));
  yc=(SD)((UB)(ptrland->y));
  ptrland=g3_ptrlanditem + (((zz)&(G3_LANDSIZE-1))<<(G3_LANDSIZESHIFT)) + ((xx+1)&(G3_LANDSIZE-1));
  yd=(SD)((UB)(ptrland->y));
  ya<<=(G3_LANDYMULTIPLYSHIFT+G3_FPSHIFT);
  yb<<=(G3_LANDYMULTIPLYSHIFT+G3_FPSHIFT);
  yc<<=(G3_LANDYMULTIPLYSHIFT+G3_FPSHIFT);
  yd<<=(G3_LANDYMULTIPLYSHIFT+G3_FPSHIFT);
  if(yb==yd)
    {
    //           B+----+C
    //            |   /|
    //            |  / |
    //            | /  |
    //            |/   |
    //           A+----+D
    if(rx>=rz)
      {
      vyonx=(yd-ya)/(G3_LANDGRIDSTEP);
      vyonz=(yc-yd)/(G3_LANDGRIDSTEP);
      return(yd+G3_FPMULS(vyonx,rx-(1<<(G3_FPSHIFT+G3_LANDGRIDSTEPSHIFT)))+G3_FPMULS(vyonz,rz));
      }
    else
      {
      vyonx=(yc-yb)/(G3_LANDGRIDSTEP);
      vyonz=(yb-ya)/(G3_LANDGRIDSTEP);
      return(ya+G3_FPMULS(vyonx,rx)+G3_FPMULS(vyonz,rz-(1<<(G3_FPSHIFT+G3_LANDGRIDSTEPSHIFT))));
      }
    }
  else
    {
    //           B+----+C
    //            |\   |
    //            | \  |
    //            |  \ |
    //            |   \|
    //           A+----+D
    if(rx+rz<=(G3_LANDGRIDSTEP<<G3_FPSHIFT))
      {
      vyonx=(yd-ya)/(G3_LANDGRIDSTEP);
      vyonz=(yb-ya)/(G3_LANDGRIDSTEP);
      return(ya+G3_FPMULS(vyonx,rx)+G3_FPMULS(vyonz,rz));
      }
    else
      {
      vyonx=(yc-yb)/(G3_LANDGRIDSTEP);
      vyonz=(yc-yd)/(G3_LANDGRIDSTEP);
      return(yc+G3_FPMULS(vyonx,rx-(1<<(G3_FPSHIFT+G3_LANDGRIDSTEPSHIFT)))+G3_FPMULS(vyonz,rz-(1<<(G3_FPSHIFT+G3_LANDGRIDSTEPSHIFT))));
      }
    }
  }


G3_LANDITEM *g3_land_getlanditemptr(SD x,SD z)
  {
  return(g3_ptrlanditem + ((z&(G3_LANDSIZE-1))*G3_LANDSIZE) + (x&(G3_LANDSIZE-1)) );
  }

G3_LANDITEM *g3_land_fplandptr(SD x,SD z)
  {
  SD rx,rz;
  rx=x>>(G3_FPSHIFT+G3_LANDGRIDSTEPSHIFT);
  rz=z>>(G3_FPSHIFT+G3_LANDGRIDSTEPSHIFT);
  return(g3_ptrlanditem + ((rz&(G3_LANDSIZE-1))*G3_LANDSIZE) + (rx&(G3_LANDSIZE-1)) );
  }

void g3_land_computedvertex(SD x,SD z,G3_DVERTEX *ptrdvertex)
  {
  G3_VERTEX src;
  G3_LANDITEM *ptrland;
  SD m41,m42,m43;
  x&=(G3_LANDSIZE-1);
  z&=(G3_LANDSIZE-1);
  m41=g3_camera.invmatrix.m41;
  m42=g3_camera.invmatrix.m42;
  m43=g3_camera.invmatrix.m43;
  g3_camera.invmatrix.m41=0;
  g3_camera.invmatrix.m42=0;
  g3_camera.invmatrix.m43=0;
  ptrland=g3_land_getlanditemptr(x,z);
  src.x=(x<<(G3_LANDGRIDSTEPSHIFT))-(g3_camera.translation.x>>G3_FPSHIFT);
  src.y=((SD)((UB)ptrland->y)<<G3_LANDYMULTIPLYSHIFT)-(g3_camera.translation.y>>G3_FPSHIFT);
  src.z=(z<<(G3_LANDGRIDSTEPSHIFT))-(g3_camera.translation.z>>G3_FPSHIFT);
  g3_vertex_mult_matrix(&src,&g3_camera.invmatrix,ptrdvertex);
  g3_dvertex_project(ptrdvertex);
  g3_camera.invmatrix.m41=m41;
  g3_camera.invmatrix.m42=m42;
  g3_camera.invmatrix.m43=m43;
  }

void g3_land_computelight(SD x,SD z)
  {
   G3_VECTOR  normal,snormal;
   FP lenght;
   G3_LANDITEM *ptrland;
   SD xx,zz;
   xx=x<<(G3_LANDGRIDSTEPSHIFT+G3_FPSHIFT);
   zz=z<<(G3_LANDGRIDSTEPSHIFT+G3_FPSHIFT);

   g3_land_getnormal(xx+G3_FPVALUE(2),zz+G3_FPVALUE(1),&normal);

   g3_land_getnormal(xx+G3_FPVALUE(-2),zz+G3_FPVALUE(1),&snormal);
   normal.x+=snormal.x;normal.y+=snormal.y;normal.z+=snormal.z;
   
   g3_land_getnormal(xx+G3_FPVALUE(2),zz+G3_FPVALUE(-1),&snormal);
   normal.x+=snormal.x;normal.y+=snormal.y;normal.z+=snormal.z;

   g3_land_getnormal(xx+G3_FPVALUE(-2),zz+G3_FPVALUE(-1),&snormal);
   normal.x+=snormal.x;normal.y+=snormal.y;normal.z+=snormal.z;

   g3_land_getnormal(xx+G3_FPVALUE(1),zz+G3_FPVALUE(2),&snormal);
   normal.x+=snormal.x;normal.y+=snormal.y;normal.z+=snormal.z;

   g3_land_getnormal(xx+G3_FPVALUE(-1),zz+G3_FPVALUE(2),&snormal);
   normal.x+=snormal.x;normal.y+=snormal.y;normal.z+=snormal.z;

   g3_land_getnormal(xx+G3_FPVALUE(1),zz+G3_FPVALUE(-2),&snormal);
   normal.x+=snormal.x;normal.y+=snormal.y;normal.z+=snormal.z;

   g3_land_getnormal(xx+G3_FPVALUE(-1),zz+G3_FPVALUE(-2),&snormal);
   normal.x+=snormal.x;normal.y+=snormal.y;normal.z+=snormal.z;

   lenght=normal.x*normal.x+normal.y*normal.y+normal.z*normal.z;
   lenght=G3_FPSQRT(lenght);
   normal.x=G3_FPDIVS(normal.x,lenght);
   normal.y=G3_FPDIVS(normal.y,lenght);
   normal.z=G3_FPDIVS(normal.z,lenght);

   ptrland=g3_ptrlanditem+((z&(G3_LANDSIZE-1))*G3_LANDSIZE)+(x&(G3_LANDSIZE-1));
   ptrland->light=(SB)((normal.x-G3_FPVALUE(0.3))>>(G3_FPSHIFT-4));
  }

#undef G3_GRAVITY
#define G3_GRAVITY G3_FPVALUE(-0.005)

FP g3_land_getgravityfromy(FP y)
{
  if(y<G3_FPVALUE(G3_LANDWAVEHEIGHT))
  {
    return(G3_FPVALUE(0));
  }
  else /*if(y>=G3_FPVALUE(G3_LANDWAVEHEIGHT))*/
  {
    return(G3_GRAVITY);
  }
/*  else
  {
    FP dy;
    FP pente;
    dy=y-G3_FPVALUE(G3_LANDWAVEHEIGHT-G3_LANDWAVEAMP);
    pente=G3_FPDIVS(G3_GRAVITY,G3_FPVALUE(G3_LANDWAVEAMP));
    return(G3_FPMULS(dy,pente));
  }
  */
}

FP g3_land_getsurfacey(void)
{
  return(G3_FPVALUE(G3_LANDWAVEHEIGHT));
}

void g3_land_getnormal(SD x,SD z,G3_VECTOR *normal)
{
  G3_LANDITEM *ptrland;
  SD xx,zz,rx,rz;
  SD ya,yb,yc,yd;
  FP nx,ny,nz,lenght;
  //           B+----+C
  //            |    |
  //            |    |
  //            |    |
  //            |    |
  //           A+----+D
  
  xx=(x>>(G3_FPSHIFT+G3_LANDGRIDSTEPSHIFT)) & (G3_LANDSIZE-1);
  zz=(z>>(G3_FPSHIFT+G3_LANDGRIDSTEPSHIFT)) & (G3_LANDSIZE-1);

  rx=(x & ((1<<(G3_FPSHIFT+G3_LANDGRIDSTEPSHIFT))-1));
  rz=(z & ((1<<(G3_FPSHIFT+G3_LANDGRIDSTEPSHIFT))-1));

  ptrland=g3_ptrlanditem + (((zz)&(G3_LANDSIZE-1))<<(G3_LANDSIZESHIFT)) + ((xx)&(G3_LANDSIZE-1));
  ya=(SD)((UB)(ptrland->y));
  ya<<=(G3_LANDYMULTIPLYSHIFT+G3_FPSHIFT);

  ptrland=g3_ptrlanditem + (((zz+1)&(G3_LANDSIZE-1))<<(G3_LANDSIZESHIFT)) + ((xx)&(G3_LANDSIZE-1));
  yb=(SD)((UB)(ptrland->y));
  yb<<=(G3_LANDYMULTIPLYSHIFT+G3_FPSHIFT);

  ptrland=g3_ptrlanditem + (((zz+1)&(G3_LANDSIZE-1))<<(G3_LANDSIZESHIFT)) + ((xx+1)&(G3_LANDSIZE-1));
  yc=(SD)((UB)(ptrland->y));
  yc<<=(G3_LANDYMULTIPLYSHIFT+G3_FPSHIFT);

  ptrland=g3_ptrlanditem + (((zz)&(G3_LANDSIZE-1))<<(G3_LANDSIZESHIFT)) + ((xx+1)&(G3_LANDSIZE-1));
  yd=(SD)((UB)(ptrland->y));
  yd<<=(G3_LANDYMULTIPLYSHIFT+G3_FPSHIFT);


  if(yb==yd)
    {
    //           B+----+C
    //            |   /|
    //            |  / |
    //            | /  |
    //            |/   |
    //           A+----+D
    if(rx>=rz)
    {
    G3_VECTOR v1,v2;
    v1.x=-(G3_LANDGRIDSTEP<<G3_FPSHIFT);
    v1.y=(ya-yd);
    v1.z=0;
    v2.x=0;
    v2.y=(yc-yd);
    v2.z=(G3_LANDGRIDSTEP<<G3_FPSHIFT);
    nx=G3_FPMULS(v1.y,v2.z)-G3_FPMULS(v2.y,v1.z);//20:10
    ny=G3_FPMULS(v2.x,v1.z)-G3_FPMULS(v1.x,v2.z);
    nz=G3_FPMULS(v1.x,v2.y)-G3_FPMULS(v2.x,v1.y);
    }
    else
    {
    G3_VECTOR v1,v2;
    v1.x=(G3_LANDGRIDSTEP<<G3_FPSHIFT);
    v1.y=(yc-yb);
    v1.z=0;
    v2.x=0;
    v2.y=(ya-yb);
    v2.z=-(G3_LANDGRIDSTEP<<G3_FPSHIFT);
    nx=G3_FPMULS(v1.y,v2.z)-G3_FPMULS(v2.y,v1.z);//20:10
    ny=G3_FPMULS(v2.x,v1.z)-G3_FPMULS(v1.x,v2.z);
    nz=G3_FPMULS(v1.x,v2.y)-G3_FPMULS(v2.x,v1.y);
    }
    }
    else
    {
    //           B+----+C
    //            |\   |
    //            | \  |
    //            |  \ |
    //            |   \|
    //           A+----+D
    if(rx+rz<=(G3_LANDGRIDSTEP<<G3_FPSHIFT))
    {
    G3_VECTOR v1,v2;
    v1.x=(G3_LANDGRIDSTEP<<G3_FPSHIFT);
    v1.y=(yd-ya);
    v1.z=0;
    v2.x=0;
    v2.y=(yb-ya);
    v2.z=(G3_LANDGRIDSTEP<<G3_FPSHIFT);
    nx=G3_FPMULS(v1.y,v2.z)-G3_FPMULS(v2.y,v1.z);//20:10
    ny=G3_FPMULS(v2.x,v1.z)-G3_FPMULS(v1.x,v2.z);
    nz=G3_FPMULS(v1.x,v2.y)-G3_FPMULS(v2.x,v1.y);
    }
    else
    {
    G3_VECTOR v1,v2;
    v1.x=-(G3_LANDGRIDSTEP<<G3_FPSHIFT);
    v1.y=(yb-yc);
    v1.z=0;
    v2.x=0;
    v2.y=(yd-yc);
    v2.z=-(G3_LANDGRIDSTEP<<G3_FPSHIFT);
    nx=G3_FPMULS(v1.y,v2.z)-G3_FPMULS(v2.y,v1.z);//20:10
    ny=G3_FPMULS(v2.x,v1.z)-G3_FPMULS(v1.x,v2.z);
    nz=G3_FPMULS(v1.x,v2.y)-G3_FPMULS(v2.x,v1.y);
    }
    }
    lenght=G3_FPMULSPREC(nx,nx,30)+G3_FPMULSPREC(ny,ny,30)+G3_FPMULSPREC(nz,nz,30);
    lenght=G3_FPSQRT(lenght);
    normal->x=G3_FPSHIFTR(nx/lenght,5);
    normal->y=G3_FPSHIFTR(ny/lenght,5);
    normal->z=G3_FPSHIFTR(nz/lenght,5);
}

