/************************************************************************/
/*  G3_LIB  : 3D GRAPHIC LIBRARY                                        */
/************************************************************************/

  #include <stdio.h>
  #include <stdlib.h>
  #include <math.h>
  #include <malloc.h>
  #include <conio.h>
  #include <string.h>

  #define _G3LIB_C_

/************************************************************************/
/*            STANDARD INCLUDES                                         */
/************************************************************************/

  #include "g3def.h"
  #include "g3graph.h"
  #include "g3malloc.h"
  #include "g3raw.h"
  #include "g3pcx.h"
  #include "g3lib.h"
  #include "g3land.h"
  #include "g3timer.h"
  #include "g3math.h"
  #include "g3objgrd.h"
  #include "g3file.h"
  #include "g3error.h"

  UD asmprotect;

  /*---------- fast access graphic functions ---------------*/

  G3_FUNCTION g3_graphcall[G3_GRAPHCALL_COUNT];
  char *g3_polygontypename[G3_GRAPHCALL_COUNT]=
    {
    "flat",
    "textured",
    "wireframe",
    "light flat",
    "light gourauld",
    "light texture gourauld",
    "light texture flat",
    "texture transparent",
    "light texture gourauld R by y",
    "light texture gourauld R by z",
    "light translucide gourault",
    "Filled Wireframe"
    };


  char *g3_polygonbadtypename="UNKNOWN";

  typedef struct
  {
    SD r;
    SD g;
    SD b;
  }
  G3_FPCOLORREGISTER;

  G3_FPCOLORREGISTER g3_fpcolor[G3_MAXCOLOR];


/************************************************************************/
/*     VARIOUS  MATH FUNCTIONS                                          */
/************************************************************************/

SD g3_arcsin(FP psin)
  {
  SD min,max,pivot,oldpivot,sinpivot;
  min=-G3_NBANGLE/4;
  max=G3_NBANGLE/4;
  pivot=0;
  do
    {
    sinpivot=g3_sincos[pivot & (G3_NBANGLE-1)].sin;
    if(sinpivot>=psin)
      max=pivot;
    else
      min=pivot;
    oldpivot=pivot;
    pivot=(min+max)>>1;
    }
  while(oldpivot!=pivot);
  return(pivot);
  }

SD g3_arctan2(FP psin,FP pcos)
  {
  SD angle;
  angle=g3_arcsin(psin);
  if(pcos<0)
  {
    angle=(G3_NBANGLE/2)-angle;
  }
  return(angle);
  }

FP g3_clipping3d(FP xinf,FP zinf,FP xsup,FP zsup)
  {
  //             +(xsup,zsup)
  //            /
  //----------------------------nearplane
  //          /
  //         +(xinf,zinf)
  //
  return(xinf+G3_FPMULS(G3_FPDIVS(xsup-xinf,zsup-zinf),g3_nearplane-zinf));
  }

SD g3_angle2angle(SD a,SD b)
  {
  b-=a;
  b=G3_ANGLE(b);
  if(b>G3_FPANGLE(G3_NBANGLE/2) && b<G3_FPANGLE(G3_NBANGLE)) b-=G3_FPANGLE(G3_NBANGLE);
  return(b);
  }

SD g3_pvmaxdistance(G3_DVERTEX *ptra,G3_DVERTEX *ptrb,G3_DVERTEX *ptrc)
  {
  G3_DVERTEX *ptrdvertexmaxdist;
  ptrdvertexmaxdist=ptra;
  if(ptrb->z>ptrdvertexmaxdist->z) ptrdvertexmaxdist=ptrb;
  if(ptrc->z>ptrdvertexmaxdist->z) ptrdvertexmaxdist=ptrc;
  if(ptrdvertexmaxdist->z<=g3_nearplane) return(0);
  if((ptra->z>=g3_nearplane) && (ptrb->z>=g3_nearplane) && (ptrc->z>=g3_nearplane))
    {
    SD cohen;
    cohen=2;
    G3_COHEN(cohen,ptra->px,g3_xmax);
    G3_COHEN(cohen,ptrb->px,g3_xmax);
    G3_COHEN(cohen,ptrc->px,g3_xmax);
    if((UD)(cohen)<(3+2))
    {
      cohen=2;
      G3_COHEN(cohen,ptra->py,g3_ymax);
      G3_COHEN(cohen,ptrb->py,g3_ymax);
      G3_COHEN(cohen,ptrc->py,g3_ymax);
      if((UD)(cohen)<(3+2))
        {
        if( ((ptrb->px-ptra->px)*(ptrc->py-ptrb->py)-(ptrb->py-ptra->py)*(ptrc->px-ptrb->px))>0) return(ptrdvertexmaxdist->z);
        }
    }
    return(0);
    }
  else
    {
    SD bx,by,bz;
    SD ax,ay,az;
    SD nx,ny,nz;
    ax=(ptrb->x-ptra->x);
    ay=(ptrb->y-ptra->y);
    az=(ptrb->z-ptra->z);
    PSEUDONORMALIZE(ax,ay,az);
    bx=(ptrc->x-ptrb->x);
    by=(ptrc->y-ptrb->y);
    bz=(ptrc->z-ptrb->z);
    PSEUDONORMALIZE(bx,by,bz);
    nx=(az*by)-(ay*bz);
    ny=(ax*bz)-(az*bx);
    nz=(ay*bx)-(ax*by);
    PSEUDONORMALIZE(nx,ny,nz);
    ax=ptrdvertexmaxdist->x;
    ay=ptrdvertexmaxdist->y;
    az=ptrdvertexmaxdist->z;
    PSEUDONORMALIZE(ax,ay,az);
    if(((ax*nx)+(ay*ny)+(az*nz))>0) return(ptrdvertexmaxdist->z);
    return(0);
    }
  }

SD g3_pvmindistance(G3_DVERTEX *ptra,G3_DVERTEX *ptrb,G3_DVERTEX *ptrc)
  {
  G3_DVERTEX *ptrdvertexmindist;
  G3_DVERTEX *ptrdvertexmaxdist;
  ptrdvertexmindist=ptra;
  ptrdvertexmaxdist=ptra;
  if(ptrb->z>ptrdvertexmaxdist->z)
    ptrdvertexmaxdist=ptrb;
  else
    if(ptrb->z<ptrdvertexmindist->z) ptrdvertexmindist=ptrb;
  if(ptrc->z>ptrdvertexmaxdist->z)
    ptrdvertexmaxdist=ptrc;
  else
    if(ptrc->z<ptrdvertexmindist->z) ptrdvertexmindist=ptrc;
  if(ptrdvertexmaxdist->z<=g3_nearplane) return(0);
  if((ptra->z>=g3_nearplane) && (ptrb->z>=g3_nearplane) && (ptrc->z>=g3_nearplane))
    {
    SD cohen;
    cohen=2;
    G3_COHEN(cohen,ptra->px,g3_xmax);
    G3_COHEN(cohen,ptrb->px,g3_xmax);
    G3_COHEN(cohen,ptrc->px,g3_xmax);
    if((UD)(cohen)<(3+2))
    {
      cohen=2;
      G3_COHEN(cohen,ptra->py,g3_ymax);
      G3_COHEN(cohen,ptrb->py,g3_ymax);
      G3_COHEN(cohen,ptrc->py,g3_ymax);
      if((UD)(cohen)<(3+2))
        {
        if( ((ptrb->px-ptra->px)*(ptrc->py-ptrb->py)-(ptrb->py-ptra->py)*(ptrc->px-ptrb->px))>0) return(ptrdvertexmaxdist->z);
        }
    }
    return(0);
    }
  else
    {
    SD bx,by,bz;
    SD ax,ay,az;
    SD nx,ny,nz;
    ax=(ptrb->x-ptra->x);
    ay=(ptrb->y-ptra->y);
    az=(ptrb->z-ptra->z);
    PSEUDONORMALIZE(ax,ay,az);
    bx=(ptrc->x-ptrb->x);
    by=(ptrc->y-ptrb->y);
    bz=(ptrc->z-ptrb->z);
    PSEUDONORMALIZE(bx,by,bz);
    nx=(az*by)-(ay*bz);
    ny=(ax*bz)-(az*bx);
    nz=(ay*bx)-(ax*by);
    PSEUDONORMALIZE(nx,ny,nz);
    ax=ptrdvertexmaxdist->x;
    ay=ptrdvertexmaxdist->y;
    az=ptrdvertexmaxdist->z;
    PSEUDONORMALIZE(ax,ay,az);
    if(((ax*nx)+(ay*ny)+(az*nz))>0) return(ptrdvertexmindist->z);
    return(0);
    }
  }

void g3_setpolarcamera(G3_ROTATION *rotation,G3_VECTOR *translation)
  {
  g3_matrix_rotzrotxrotytranslate(rotation,translation,&g3_camera.matrix);
  g3_matrix_invert(&g3_camera.matrix,&g3_camera.invmatrix);
  }

SD g3_pv2d(SD x1,SD y1,SD x2,SD y2,SD x3,SD y3)
  {
  return( (x2-x1)*(y3-y2) - (y2-y1)*(x3-x2) );
  }

SD g3_isintriangle(SD x1,SD y1,SD x2,SD y2,SD x3,SD y3,SD x,SD y)
  {
  return( g3_pv2d(x1,y1,x2,y2,x,y)>0 &&
          g3_pv2d(x2,y2,x3,y3,x,y)>0 &&
          g3_pv2d(x3,y3,x1,y1,x,y)>0 );
  }

SD g3_isinquad(SD x1,SD y1,SD x2,SD y2,SD x3,SD y3,SD x4,SD y4,SD x,SD y)
  {
  return( g3_pv2d(x1,y1,x2,y2,x,y)>0 &&
          g3_pv2d(x2,y2,x3,y3,x,y)>0 &&
          g3_pv2d(x3,y3,x4,y4,x,y)>0 &&
          g3_pv2d(x4,y4,x1,y1,x,y)>0 );
  }


/************************************************************************/
/*  CAMERA MANAGEMENT                                                   */
/************************************************************************/

void g3_camera_lockonobject(G3_OBJECT *ptrobject)
  {
  g3_camera.rotation.x=G3_ANGLE(ptrobject->rotation.x+G3_FPANGLE(G3_NBANGLE/16));
  g3_camera.rotation.y=G3_ANGLE(ptrobject->rotation.y+G3_FPANGLE(G3_NBANGLE/2)+G3_FPANGLE(G3_NBANGLE/8));
  g3_camera.rotation.z=G3_ANGLE(ptrobject->rotation.z);
  g3_camera.translation.x=ptrobject->translation.x+G3_FPVALUE(512);
  g3_camera.translation.y=ptrobject->translation.y+G3_FPVALUE(256);
  g3_camera.translation.z=ptrobject->translation.z+G3_FPVALUE(512);
  g3_matrix_rotzrotxrotytranslate(&g3_camera.rotation,&g3_camera.translation,&g3_camera.matrix);
  g3_matrix_invert(&g3_camera.matrix,&g3_camera.invmatrix);
  //g3_matrix_xyzoom(&g3_camera.invmatrix);
  }

void g3_camera_set(SD rx,SD ry,SD rz,FP tx,FP ty,FP tz)
  {
  g3_camera.rotation.x=G3_ANGLE(rx);
  g3_camera.rotation.y=G3_ANGLE(ry);
  g3_camera.rotation.z=G3_ANGLE(rz);
  g3_camera.translation.x=tx;
  g3_camera.translation.y=ty;
  g3_camera.translation.z=tz;
  g3_matrix_rotzrotxrotytranslate(&g3_camera.rotation,&g3_camera.translation,&g3_camera.matrix);
  g3_matrix_invert(&g3_camera.matrix,&g3_camera.invmatrix);
  //g3_matrix_xyzoom(&g3_camera.invmatrix);
  }

FP g3_camera_getx(void)
  {
  return(g3_camera.translation.x);
  }

FP g3_camera_gety(void)
  {
  return(g3_camera.translation.y);
  }

FP g3_camera_getz(void)
  {
  return(g3_camera.translation.z);
  }

void g3_camera_setx(FP x)
  {
  g3_camera.translation.x=x;
  }

void g3_camera_sety(FP y)
  {
  g3_camera.translation.y=y;
  }

void g3_camera_setz(FP z)
  {
  g3_camera.translation.z=z;
  }

SD g3_camera_getrx(void)
  {
  return(g3_camera.rotation.x);
  }

SD g3_camera_getry(void)
  {
  return(g3_camera.rotation.y);
  }

SD g3_camera_getrz(void)
  {
  return(g3_camera.rotation.z);
  }

void g3_camera_setrx(SD x)
  {
  g3_camera.rotation.x=x;
  }

void g3_camera_setry(SD y)
  {
  g3_camera.rotation.y=y;
  }

void g3_camera_setrz(SD z)
  {
  g3_camera.rotation.z=z;
  }


void g3_camera_polar()
  {
  G3_MATRIX mry;
  G3_MATRIX mrx;
  g3_matrix_rotx(g3_camera.rotation.x,&mrx);
  g3_matrix_roty(g3_camera.rotation.y,&mry);
  g3_matrix_mult(&mrx,&mry,&g3_camera.matrix);
  g3_camera.matrix.m41=g3_camera.translation.x;
  g3_camera.matrix.m42=g3_camera.translation.y;
  g3_camera.matrix.m43=g3_camera.translation.z;
  g3_matrix_invert(&g3_camera.matrix,&g3_camera.invmatrix);
  //g3_matrix_xyzoom(&g3_camera.invmatrix);
  }

void g3_camera_progress(SD drx,SD dry,SD drz,FP dtx,FP dty,FP dtz)
  {
  G3_ROTATION rotation;
  G3_VECTOR translation;
  G3_MATRIX updatingmatrix,finalmatrix,originalmatrix;
  /* use sampling values */
  drx*=g3_timer_sampledelay();
  dry*=g3_timer_sampledelay();
  drz*=g3_timer_sampledelay();
  dtx*=g3_timer_sampledelay();
  dty*=g3_timer_sampledelay();
  dtz*=g3_timer_sampledelay();

  /* compute progression of the camera */

  g3_camera.translation.x+=G3_FPMULS(g3_camera.matrix.m11,dtx)+G3_FPMULS(g3_camera.matrix.m21,dty)+G3_FPMULS(g3_camera.matrix.m31,dtz);
  g3_camera.translation.y+=G3_FPMULS(g3_camera.matrix.m12,dtx)+G3_FPMULS(g3_camera.matrix.m22,dty)+G3_FPMULS(g3_camera.matrix.m32,dtz);
  g3_camera.translation.z+=G3_FPMULS(g3_camera.matrix.m13,dtx)+G3_FPMULS(g3_camera.matrix.m23,dty)+G3_FPMULS(g3_camera.matrix.m33,dtz);

  g3_matrix_copy(&g3_camera.matrix,&originalmatrix);
  originalmatrix.m41=0;
  originalmatrix.m42=0;
  originalmatrix.m43=0;

  /* compute the updating matrix */
  rotation.x=G3_ANGLE(drx);
  rotation.y=G3_ANGLE(dry);
  rotation.z=G3_ANGLE(drz);
  translation.x=0;
  translation.y=0;
  translation.z=0;
  g3_matrix_rotzrotxrotytranslate(&rotation,&translation,&updatingmatrix);

  /* compute the final matrix */

  g3_matrix_mult(&updatingmatrix,&originalmatrix,&finalmatrix);
  finalmatrix.m41=g3_camera.translation.x;
  finalmatrix.m42=g3_camera.translation.y;
  finalmatrix.m43=g3_camera.translation.z;
  g3_matrix_copy(&finalmatrix,&g3_camera.matrix);

  /* adjust camera matrix */
  if(g3_matrix_getrotation(&g3_camera.matrix,&g3_camera.rotation))
  {
      g3_matrix_rotzrotxrotytranslate(&g3_camera.rotation,&g3_camera.translation,&g3_camera.matrix);
  }

  /* compute inverse matrix of camera matrix */
  g3_matrix_invert(&g3_camera.matrix,&g3_camera.invmatrix);
  //g3_matrix_xyzoom(&g3_camera.invmatrix);
  }

void g3_camera_lookat(G3_VECTOR *viewpoint)
  {
  G3_VECTOR v;
  G3_MATRIX m1,m2;

  v.x=viewpoint->x-g3_camera.translation.x;
  v.y=viewpoint->y-g3_camera.translation.y;
  v.z=viewpoint->z-g3_camera.translation.z;

  g3_matrix_phi(&v,&m1);      // rotation en x
  //g3_matrix_identity(&m1);      // rotation en x

  g3_matrix_theta(&v,&m2);    // rotation en y
  g3_matrix_mult(&m1,&m2,&g3_camera.matrix);
  g3_camera.matrix.m41=g3_camera.translation.x;
  g3_camera.matrix.m42=g3_camera.translation.y;
  g3_camera.matrix.m43=g3_camera.translation.z;
  g3_matrix_invert(&g3_camera.matrix,&g3_camera.invmatrix);
  //g3_matrix_xyzoom(&g3_camera.invmatrix);
  }

/************************************************************************/
/*  LIGHTENED COLOR MANAGEMENT                                          */
/************************************************************************/


void g3_colorlight_compute(SD start,SD shadowindex,SD fogindex,SD flashindex,DPIXEL *tab,SD rshift,SD gshift,SD bshift,SD rcount,SD gcount,SD bcount)
  {
  SD penter,penteg,penteb;
  SD searchr,searchg,searchb;
  SD color,light;

    g3_rmask=(0xFFFF>>(16-rcount))<<rshift;
    g3_gmask=(0xFFFF>>(16-gcount))<<gshift;
    g3_bmask=(0xFFFF>>(16-bcount))<<bshift;
    g3_halfmask=((g3_rmask>>1) & g3_rmask) | ((g3_gmask>>1) & g3_gmask) | ((g3_bmask>>1) & g3_bmask);

    for(color=0;color<256;color++)
    {
      g3_fpcolor[color].r=(SD)(((UD)g3_colorregister[color].red  )<<G3_FPSHIFT);
      g3_fpcolor[color].g=(SD)(((UD)g3_colorregister[color].green)<<G3_FPSHIFT);
      g3_fpcolor[color].b=(SD)(((UD)g3_colorregister[color].blue )<<G3_FPSHIFT);
    }
    
    // for each existing color....
    for(color=0;color<G3_MAXCOLOR;color++)
    {
      // compute displacement from shadow color to fog color
      penter=(g3_fpcolor[fogindex].r-g3_fpcolor[shadowindex].r)/(G3_MAXLIGHTING/4);
      penteg=(g3_fpcolor[fogindex].g-g3_fpcolor[shadowindex].g)/(G3_MAXLIGHTING/4);
      penteb=(g3_fpcolor[fogindex].b-g3_fpcolor[shadowindex].b)/(G3_MAXLIGHTING/4);
      searchr=g3_fpcolor[shadowindex].r;
      searchg=g3_fpcolor[shadowindex].g;
      searchb=g3_fpcolor[shadowindex].b;
      for(light=start;light<start+G3_MAXLIGHTING/4;light++)
      {
        G3_SYMCOLORLIGHT(tab,color,light)=((searchr>>(G3_FPSHIFT+1))<<(rshift+(rcount-5))) | ((searchg>>(G3_FPSHIFT+1))<<(gshift+(gcount-5))) | ((searchb>>(G3_FPSHIFT+1))<<(bshift+(bcount-5)));
        searchr+=penter;
        searchg+=penteg;
        searchb+=penteb;
      }
      // compute displacement from fog color to our color
      penter=(g3_fpcolor[color].r-g3_fpcolor[fogindex].r)/(G3_MAXLIGHTING/2);
      penteg=(g3_fpcolor[color].g-g3_fpcolor[fogindex].g)/(G3_MAXLIGHTING/2);
      penteb=(g3_fpcolor[color].b-g3_fpcolor[fogindex].b)/(G3_MAXLIGHTING/2);
      searchr=g3_fpcolor[fogindex].r;
      searchg=g3_fpcolor[fogindex].g;
      searchb=g3_fpcolor[fogindex].b;
      for(light=start+G3_MAXLIGHTING/4;light<start+G3_MAXLIGHTING-(G3_MAXLIGHTING/4);light++)
      {
        G3_SYMCOLORLIGHT(tab,color,light)=((searchr>>(G3_FPSHIFT+1))<<(rshift+(rcount-5))) | ((searchg>>(G3_FPSHIFT+1))<<(gshift+(gcount-5))) | ((searchb>>(G3_FPSHIFT+1))<<(bshift+(bcount-5)));
        searchr+=penter;
        searchg+=penteg;
        searchb+=penteb;
      }
      // compute displacement from our color to flash color
      penter=(g3_fpcolor[flashindex].r-g3_fpcolor[color].r)/(G3_MAXLIGHTING/4);
      penteg=(g3_fpcolor[flashindex].g-g3_fpcolor[color].g)/(G3_MAXLIGHTING/4);
      penteb=(g3_fpcolor[flashindex].b-g3_fpcolor[color].b)/(G3_MAXLIGHTING/4);
      searchr=g3_fpcolor[color].r;
      searchg=g3_fpcolor[color].g;
      searchb=g3_fpcolor[color].b;
      for(light=start+G3_MAXLIGHTING-(G3_MAXLIGHTING/4);light<start+G3_MAXLIGHTING;light++)
      {
        G3_SYMCOLORLIGHT(tab,color,light)=((searchr>>(G3_FPSHIFT+1))<<(rshift+(rcount-5))) | ((searchg>>(G3_FPSHIFT+1))<<(gshift+(gcount-5))) | ((searchb>>(G3_FPSHIFT+1))<<(bshift+(bcount-5)));
        searchr+=penter;
        searchg+=penteg;
        searchb+=penteb;
      }
    }
  }


void g3_colorlight_load(char *filename,SD rshift,SD gshift,SD bshift,SD rcount,SD gcount,SD bcount)
  {
      // fog 1 table
      g3_colorlight_compute(             0,5,1,3,g3_colorlight,rshift,gshift,bshift,rcount,gcount,bcount);
      // fog 2 table
      g3_colorlight_compute(G3_MAXLIGHTING,6,2,4,g3_colorlight,rshift,gshift,bshift,rcount,gcount,bcount);
  }


/************************************************************************/
/*     VERTEX FUNCTIONS                                                 */
/************************************************************************/

void g3_vertex_mult_matrix(G3_VERTEX *src,G3_MATRIX *m,G3_DVERTEX *dst)
  {
  dst->x  = (src->x*m->m11)
          + (src->y*m->m21)
          + (src->z*m->m31)
          + (m->m41);
  dst->y  = (src->x*m->m12)
          + (src->y*m->m22)
          + (src->z*m->m32)
          + (m->m42);
  dst->z  = (src->x*m->m13)
          + (src->y*m->m23)
          + (src->z*m->m33)
          + (m->m43);
  }

void g3_vector_mult_matrix(G3_VECTOR *src,G3_MATRIX *m,G3_VECTOR *dst)
  {
  dst->x  = G3_FPMULS(src->x,m->m11)
          + G3_FPMULS(src->y,m->m21)
          + G3_FPMULS(src->z,m->m31)
          + (m->m41);
  dst->y  = G3_FPMULS(src->x,m->m12)
          + G3_FPMULS(src->y,m->m22)
          + G3_FPMULS(src->z,m->m32)
          + (m->m42);
  dst->z  = G3_FPMULS(src->x,m->m13)
          + G3_FPMULS(src->y,m->m23)
          + G3_FPMULS(src->z,m->m33)
          + (m->m43);
  }

void g3_vertex_set(FP x,FP y,FP z,G3_VERTEX *rv)
  {
  rv->x=x;
  rv->y=y;
  rv->z=z;
  }

void g3_vertex_sub(G3_VERTEX *va,G3_VERTEX *vb,G3_VERTEX *rv)
  {
  rv->x=va->x-vb->x;
  rv->y=va->y-vb->y;
  rv->z=va->z-vb->z;
  }

void g3_vertex_add(G3_VERTEX *va,G3_VERTEX *vb,G3_VERTEX *rv)
  {
  rv->x=va->x+vb->x;
  rv->y=va->y+vb->y;
  rv->z=va->z+vb->z;
  }

void g3_vertex_mul(G3_VERTEX *va,G3_VERTEX *vb,G3_VERTEX *rv)
  {
  rv->x=va->x*vb->x;
  rv->y=va->y*vb->y;
  rv->z=va->z*vb->z;
  }

/************************************************************************/
/*     DVERTEX FUNCTIONS                                                 */
/************************************************************************/

void g3_dvertex_project(G3_DVERTEX *ptrdvertex)
  {
  G3_PTRDVERTEX_PROJECT(ptrdvertex);
  }


/************************************************************************/
/*     VECTOR FUNCTIONS                                                 */
/************************************************************************/

FP g3_vector_norm(G3_VECTOR *v)
  {
    return(G3_FPSQRT(G3_FPMULS(v->x,v->x)+G3_FPMULS(v->y,v->y)+G3_FPMULS(v->z,v->z)));
  }

void g3_z_transform(FP z,G3_MATRIX *m,G3_VECTOR *dst)
  {
  dst->x  = (z*m->m31);
  dst->y  = (z*m->m32);
  dst->z  = (z*m->m33);
  }

void g3_x_transform(FP x,G3_MATRIX *m,G3_VECTOR *dst)
  {
  dst->x  = (x*m->m11);
  dst->y  = (x*m->m12);
  dst->z  = (x*m->m13);
  }

void g3_y_transform(FP y,G3_MATRIX *m,G3_VECTOR *dst)
  {
  dst->x  = (y*m->m21);
  dst->y  = (y*m->m22);
  dst->z  = (y*m->m23);
  }




void g3_vector_normalize(G3_VECTOR *v)
  {
    FP norm;
    norm=g3_vector_norm(v);
    if(norm!=0)
      {
      v->x/=norm;
      v->y/=norm;
      v->z/=norm;
      }
  }

void g3_vector_normal(G3_VECTOR *a,G3_VECTOR *b,G3_VECTOR *normal)
  {
    normal->x=(a->z*b->y)-(a->y*b->z);
    normal->y=(a->x*b->z)-(a->z*b->x);
    normal->z=(a->y*b->x)-(a->x*b->y);
    g3_vector_normalize(normal);
  }

/************************************************************************/
/*     MATRIX FUNCTIONS                                                 */
/*                                                                      */
/*      y                                                               */
/*      |                                                               */
/*      |      z                                                        */
/*      |     /                                                         */
/*      |    /                                                          */
/*      |   /                                                           */
/*      |  /                                                            */
/*      | /                                                             */
/*      |/                                                              */
/*     O+-------------------->x                                         */
/*                                                                      */
/************************************************************************/
void g3_matrix_identity(G3_MATRIX *m)
  {
  m->m11=G3_FPUNIT; m->m12=0;     m->m13=0;
  m->m21=0;     m->m22=G3_FPUNIT; m->m23=0;
  m->m31=0;     m->m32=0;     m->m33=G3_FPUNIT;
  m->m41=0;     m->m42=0;     m->m43=0;
  }

void g3_matrix_xyzoom(G3_MATRIX *m)
  {
  m->m11=G3_FPMULS(m->m11,g3_xzoom);m->m12=G3_FPMULS(m->m12,g3_yzoom);
  m->m21=G3_FPMULS(m->m21,g3_xzoom);m->m22=G3_FPMULS(m->m22,g3_yzoom);
  m->m31=G3_FPMULS(m->m31,g3_xzoom);m->m32=G3_FPMULS(m->m32,g3_yzoom);
  m->m41=G3_FPMULS(m->m41,g3_xzoom);m->m42=G3_FPMULS(m->m42,g3_yzoom);
  }

SD g3_matrix_getrotation(G3_MATRIX *m,G3_ROTATION *r)
  {
  FP sinx,cosx;
  FP siny,cosy;
  FP sinz,cosz;
  sinx=-m->m32;
  cosx=G3_FPSQRT(G3_FPVALUE(G3_FPVALUE(1))-(sinx*sinx));
  if(cosx<=G3_FPEPSILON)
    {
    r->x=G3_NBANGLE/4;
    if(sinx<0) r->x=-G3_NBANGLE/4;
    r->y=0;
    sinz=-m->m21;
    cosz=-m->m11;
    r->z=g3_arctan2(sinz,cosz);
    r->x<<=G3_FPSHIFT;
    r->y<<=G3_FPSHIFT;
    r->z<<=G3_FPSHIFT;
    return(G3_FALSE);
    }
  siny=(m->m31<<G3_FPSHIFT)/cosx;
  cosy=(m->m33<<G3_FPSHIFT)/cosx;
  sinz=(m->m12<<G3_FPSHIFT)/cosx;
  cosz=(m->m22<<G3_FPSHIFT)/cosx;
  r->x=g3_arctan2(sinx,cosx);
  r->y=g3_arctan2(siny,cosy);
  r->z=g3_arctan2(sinz,cosz);
  r->x<<=G3_FPSHIFT;
  r->y<<=G3_FPSHIFT;
  r->z<<=G3_FPSHIFT;
  return(G3_TRUE);
  }

void g3_matrix_null(G3_MATRIX *m)
  {
  m->m11=0;     m->m12=0;     m->m13=0;
  m->m21=0;     m->m22=0;     m->m23=0;
  m->m31=0;     m->m32=0;     m->m33=0;
  m->m41=0;     m->m42=0;     m->m43=0;
  }

void g3_matrix_copy(G3_MATRIX *s,G3_MATRIX *d)
  {
  memcpy(d,s,sizeof(G3_MATRIX));
  }

void g3_matrix_rotx(SD angle,G3_MATRIX *m)
  {
  //  X=x;
  //  Y=y*cosx-z*sinx;
  //  Z=y*sinx+z*cosx;
  G3_SINCOS *ptr;
  ptr=&g3_sincos[(angle>>G3_FPSHIFT) & (G3_NBANGLE-1)];
  m->m11=G3_FPUNIT; m->m12=0;     m->m13=0;
  m->m21=0;     m->m22=ptr->cos;  m->m23=ptr->sin;
  m->m31=0;     m->m32=-ptr->sin; m->m33=ptr->cos;
  m->m41=0;     m->m42=0;     m->m43=0;
  }

void g3_matrix_roty(SD angle,G3_MATRIX *m)
  {
  //  X=x*cosy+z*siny
  //  Y=y;
  //  Z=-x*siny+z*cosy
  G3_SINCOS *ptr;
  ptr=&g3_sincos[(angle>>G3_FPSHIFT) & (G3_NBANGLE-1)];
  m->m11=ptr->cos;  m->m12=0;     m->m13=-ptr->sin;
  m->m21=0;     m->m22=G3_FPUNIT; m->m23=0;
  m->m31=ptr->sin;  m->m32=0;     m->m33=ptr->cos;
  m->m41=0;     m->m42=0;     m->m43=0;
  }

void g3_matrix_rotz(SD angle,G3_MATRIX *m)
  {
  //  X=x*cosz-y*sinz;
  //  Y=x*sinz+y*cosz;
  //  Z=z;
  G3_SINCOS *ptr;
  ptr=&g3_sincos[(angle>>G3_FPSHIFT) & (G3_NBANGLE-1)];
  m->m11=ptr->cos;  m->m12=ptr->sin;  m->m13=0;
  m->m21=-ptr->sin; m->m22=ptr->cos;  m->m23=0;
  m->m31=0;     m->m32=0;     m->m33=G3_FPUNIT;
  m->m41=0;     m->m42=0;     m->m43=0;
  }

void g3_matrix_translate(G3_VECTOR *v,G3_MATRIX *m)
  {
  m->m11=G3_FPUNIT; m->m12=0;         m->m13=0;
  m->m21=0;         m->m22=G3_FPUNIT; m->m23=0;
  m->m31=0;         m->m32=0;         m->m33=G3_FPUNIT;
  m->m41=v->x;      m->m42=v->y;      m->m43=v->z;
  }

void g3_matrix_display(G3_MATRIX *m)
  {
  }

void g3_matrix_invtranslate(G3_VECTOR *v,G3_MATRIX *m)
  {
  m->m11=G3_FPUNIT; m->m12=0;           m->m13=0;
  m->m21=0;         m->m22=G3_FPUNIT;   m->m23=0;
  m->m31=0;         m->m32=0;           m->m33=G3_FPUNIT;
  m->m41=-v->x;     m->m42=-v->y;       m->m43=-v->z;
  }


void g3_matrix_scale(G3_VERTEX *v,G3_MATRIX *rm)
  {
  rm->m11=v->x;   rm->m12=0;      rm->m13=0;
  rm->m21=0;      rm->m22=v->y;   rm->m23=0;
  rm->m31=0;      rm->m32=0;      rm->m33=v->z;
  rm->m41=0;      rm->m42=0;      rm->m43=0;
  }

void g3_matrix_scalexyz(G3_MATRIX *rm,SD scale)
  {
  rm->m11=G3_FPMULS(rm->m11,scale);
  rm->m12=G3_FPMULS(rm->m12,scale);
  rm->m13=G3_FPMULS(rm->m13,scale);
  rm->m21=G3_FPMULS(rm->m21,scale);
  rm->m22=G3_FPMULS(rm->m22,scale);
  rm->m23=G3_FPMULS(rm->m23,scale);
  rm->m31=G3_FPMULS(rm->m31,scale);
  rm->m32=G3_FPMULS(rm->m32,scale);
  rm->m33=G3_FPMULS(rm->m33,scale);
  }

void g3_matrix_invert(G3_MATRIX *sm,G3_MATRIX *rm)
  {
  memcpy(rm,sm,sizeof(G3_MATRIX));
  G3_FPSWAP(rm->m21,rm->m12);
  G3_FPSWAP(rm->m31,rm->m13);
  G3_FPSWAP(rm->m32,rm->m23);
  rm->m41=G3_FPMULS(-rm->m11,sm->m41) + G3_FPMULS(-rm->m21,sm->m42) + G3_FPMULS(-rm->m31,sm->m43);
  rm->m42=G3_FPMULS(-rm->m12,sm->m41) + G3_FPMULS(-rm->m22,sm->m42) + G3_FPMULS(-rm->m32,sm->m43);
  rm->m43=G3_FPMULS(-rm->m13,sm->m41) + G3_FPMULS(-rm->m23,sm->m42) + G3_FPMULS(-rm->m33,sm->m43);
  }

void g3_matrix_mult(G3_MATRIX *a,G3_MATRIX *b,G3_MATRIX *rm)
  {
  rm->m11=   ((a->m11*b->m11)
            +(a->m12*b->m21)
            +(a->m13*b->m31))>>G3_FPSHIFT;
  rm->m21=   ((a->m21*b->m11)
            +(a->m22*b->m21)
            +(a->m23*b->m31))>>G3_FPSHIFT;
  rm->m31=   ((a->m31*b->m11)
            +(a->m32*b->m21)
            +(a->m33*b->m31))>>G3_FPSHIFT;
  rm->m41=   G3_FPMULS(a->m41,b->m11)
            +G3_FPMULS(a->m42,b->m21)
            +G3_FPMULS(a->m43,b->m31)
            +(b->m41);

  rm->m12=   ((a->m11*b->m12)
            +(a->m12*b->m22)
            +(a->m13*b->m32))>>G3_FPSHIFT;
  rm->m22=   ((a->m21*b->m12)
            +(a->m22*b->m22)
            +(a->m23*b->m32))>>G3_FPSHIFT;
  rm->m32=   ((a->m31*b->m12)
            +(a->m32*b->m22)
            +(a->m33*b->m32))>>G3_FPSHIFT;
  rm->m42=   G3_FPMULS(a->m41,b->m12)
            +G3_FPMULS(a->m42,b->m22)
            +G3_FPMULS(a->m43,b->m32)
            +(b->m42);

  rm->m13=   ((a->m11*b->m13)
            +(a->m12*b->m23)
            +(a->m13*b->m33))>>G3_FPSHIFT;
  rm->m23=   ((a->m21*b->m13)
            +(a->m22*b->m23)
            +(a->m23*b->m33))>>G3_FPSHIFT;
  rm->m33=   ((a->m31*b->m13)
            +(a->m32*b->m23)
            +(a->m33*b->m33))>>G3_FPSHIFT;
  rm->m43=   G3_FPMULS(a->m41,b->m13)
            +G3_FPMULS(a->m42,b->m23)
            +G3_FPMULS(a->m43,b->m33)
            +(b->m43);
  }

void g3_matrix_multwithouttranslation(G3_MATRIX *a,G3_MATRIX *b,G3_MATRIX *rm)
  {
  rm->m11=   ((a->m11*b->m11)
            +(a->m12*b->m21)
            +(a->m13*b->m31))>>G3_FPSHIFT;
  rm->m21=   ((a->m21*b->m11)
            +(a->m22*b->m21)
            +(a->m23*b->m31))>>G3_FPSHIFT;
  rm->m31=   ((a->m31*b->m11)
            +(a->m32*b->m21)
            +(a->m33*b->m31))>>G3_FPSHIFT;
  rm->m41=   0;//a->m41;
  
  rm->m12=   ((a->m11*b->m12)
            +(a->m12*b->m22)
            +(a->m13*b->m32))>>G3_FPSHIFT;
  rm->m22=   ((a->m21*b->m12)
            +(a->m22*b->m22)
            +(a->m23*b->m32))>>G3_FPSHIFT;
  rm->m32=   ((a->m31*b->m12)
            +(a->m32*b->m22)
            +(a->m33*b->m32))>>G3_FPSHIFT;
  rm->m42=  0;//a->m42;

  rm->m13=   ((a->m11*b->m13)
            +(a->m12*b->m23)
            +(a->m13*b->m33))>>G3_FPSHIFT;
  rm->m23=   ((a->m21*b->m13)
            +(a->m22*b->m23)
            +(a->m23*b->m33))>>G3_FPSHIFT;
  rm->m33=   ((a->m31*b->m13)
            +(a->m32*b->m23)
            +(a->m33*b->m33))>>G3_FPSHIFT;
  rm->m43=  0;//a->m43;
  }


void g3_matrix_theta(G3_VECTOR *v,G3_MATRIX *m)
  {
  //  X=x*cosy+z*siny
  //  Y=y;
  //  Z=-x*siny+z*cosy
  FP norm,costheta,sintheta;
  norm=G3_FPSQRT(((v->x>>G3_FPSHIFT)*(v->x>>G3_FPSHIFT))+(v->z>>G3_FPSHIFT)*(v->z>>G3_FPSHIFT));
  if(norm==G3_FPVALUE(0))
    {
    g3_matrix_identity(m);
    return;
    }
  costheta=(v->z/norm);
  sintheta=(v->x/norm);
  m->m11=costheta;  m->m12=0;           m->m13=-sintheta;
  m->m21=0;         m->m22=G3_FPUNIT;   m->m23=0;
  m->m31=sintheta;  m->m32=0;           m->m33=costheta;
  m->m41=0;         m->m42=0;           m->m43=0;
  }

void g3_matrix_phi(G3_VECTOR *v,G3_MATRIX *m)
  {
  //  X=x;
  //  Y=y*cosx-z*sinx;
  //  Z=y*sinx+z*cosx;
  FP norm,norm2,cosphi,sinphi;
  norm=G3_FPSQRT( ((v->x>>G3_FPSHIFT)*(v->x>>G3_FPSHIFT)) + ((v->y>>G3_FPSHIFT)*(v->y>>G3_FPSHIFT)) + ((v->z>>G3_FPSHIFT)*(v->z>>G3_FPSHIFT)) );
  if(norm==0)
    {
    g3_matrix_identity(m);
    return;
    }
  norm2=g3_dwordsqrt( ((v->x>>G3_FPSHIFT)*(v->x>>G3_FPSHIFT)) + ((v->z>>G3_FPSHIFT)*(v->z>>G3_FPSHIFT)) );

  cosphi=((norm2<<G3_FPSHIFT)/norm);
  sinphi=-(v->y/norm);
  m->m11=G3_FPUNIT;   m->m12=0;         m->m13=0;
  m->m21=0;           m->m22=cosphi;    m->m23=sinphi;
  m->m31=0;           m->m32=-sinphi;   m->m33=cosphi;
  m->m41=0;           m->m42=0;         m->m43=0;
  }

void g3_matrix_rotzrotxroty(G3_ROTATION *rot,G3_MATRIX *m)
  {
//  G3_SINCOS *ptr;
  FP sinx,siny,sinz;
  FP cosx,cosy,cosz;
  FP tmp;
  sinx=g3_sincos[G3_ANGLE(rot->x)>>G3_FPSHIFT].sin;
  cosx=g3_sincos[G3_ANGLE(rot->x)>>G3_FPSHIFT].cos;
  siny=g3_sincos[G3_ANGLE(rot->y)>>G3_FPSHIFT].sin;
  cosy=g3_sincos[G3_ANGLE(rot->y)>>G3_FPSHIFT].cos;
  sinz=g3_sincos[G3_ANGLE(rot->z)>>G3_FPSHIFT].sin;
  cosz=g3_sincos[G3_ANGLE(rot->z)>>G3_FPSHIFT].cos;
  tmp=((sinx*siny)>>G3_FPSHIFT);
  m->m11=((cosz*cosy)>>G3_FPSHIFT) + ((sinz*tmp)>>G3_FPSHIFT);
  m->m21=((cosz * tmp)>>G3_FPSHIFT) - ((sinz*cosy)>>G3_FPSHIFT);
  m->m31=(cosx*siny)>>G3_FPSHIFT;
  m->m41=0;
  m->m12=(sinz*cosx)>>G3_FPSHIFT;
  m->m22=(cosz*cosx)>>G3_FPSHIFT;
  m->m32=-sinx;
  m->m42=0;
  tmp=((sinx*cosy)>>G3_FPSHIFT);
  m->m13=((tmp*sinz)>>G3_FPSHIFT) - ((cosz*siny)>>G3_FPSHIFT);
  m->m23=((sinz*siny)>>G3_FPSHIFT) + ((tmp*cosz)>>G3_FPSHIFT);
  m->m33=(cosx*cosy)>>G3_FPSHIFT;
  m->m43=0;
  }

void g3_matrix_rotzrotxrotytranslate(G3_ROTATION *rot,G3_VECTOR *trs,G3_MATRIX *m)
  {
#if 0
  FP sinx,siny,sinz;
  FP cosx,cosy,cosz;
  FP tmp;
  sinx=g3_sincos[G3_ANGLE(rot->x)>>G3_FPSHIFT].sin;
  cosx=g3_sincos[G3_ANGLE(rot->x)>>G3_FPSHIFT].cos;
  siny=g3_sincos[G3_ANGLE(rot->y)>>G3_FPSHIFT].sin;
  cosy=g3_sincos[G3_ANGLE(rot->y)>>G3_FPSHIFT].cos;
  sinz=g3_sincos[G3_ANGLE(rot->z)>>G3_FPSHIFT].sin;
  cosz=g3_sincos[G3_ANGLE(rot->z)>>G3_FPSHIFT].cos;
  tmp=((sinx*siny)>>G3_FPSHIFT);
  m->m11=((cosz*cosy)>>G3_FPSHIFT) + ((sinz*tmp)>>G3_FPSHIFT);
  m->m21=((cosz * tmp)>>G3_FPSHIFT) - ((sinz*cosy)>>G3_FPSHIFT);
  m->m31=(cosx*siny)>>G3_FPSHIFT;
  m->m41=trs->x;
  m->m12=(sinz*cosx)>>G3_FPSHIFT;
  m->m22=(cosz*cosx)>>G3_FPSHIFT;
  m->m32=-sinx;
  m->m42=trs->y;
  tmp=((sinx*cosy)>>G3_FPSHIFT);
  m->m13=((tmp*sinz)>>G3_FPSHIFT) - ((cosz*siny)>>G3_FPSHIFT);
  m->m23=((sinz*siny)>>G3_FPSHIFT) + ((tmp*cosz)>>G3_FPSHIFT);
  m->m33=(cosx*cosy)>>G3_FPSHIFT;
  m->m43=trs->z;
#else
  FP coszpy;
  FP coszmy;
  FP coszpx;
  FP coszmx;
  FP cosxpy;
  FP cosxmy;
  FP sinxpz;
  FP sinxmz;
  FP sinxpy;
  FP sinxmy;
  FP sinypz;
  FP sinymz;
  FP cosxpzmy;           //cos(+x-y+z)
  FP cosxmzpy;           //cos(+x+y-z)
  FP cosxmzmymcosxpypz;  //cos(+x-y-z)-cos(+x+y+z)
  FP sinxpypz;           //sin(+x+y+z)
  FP sinxpymzpsinxmypz;  //sin(+x+y-z)+sin(+x-y+z)
  FP sinxmymz;           //sin(+x-y-z)

  coszpy    =g3_sincos[G3_ANGLE(rot->z+rot->y)>>G3_FPSHIFT].cos;
  sinypz    =g3_sincos[G3_ANGLE(rot->y+rot->z)>>G3_FPSHIFT].sin;

  coszpx    =g3_sincos[G3_ANGLE(rot->z+rot->x)>>G3_FPSHIFT].cos;
  sinxpz    =g3_sincos[G3_ANGLE(rot->x+rot->z)>>G3_FPSHIFT].sin;

  cosxpy    =g3_sincos[G3_ANGLE(rot->x+rot->y)>>G3_FPSHIFT].cos;
  sinxpy    =g3_sincos[G3_ANGLE(rot->x+rot->y)>>G3_FPSHIFT].sin;

  cosxmy    =g3_sincos[G3_ANGLE(rot->x-rot->y)>>G3_FPSHIFT].cos;
  sinxmy    =g3_sincos[G3_ANGLE(rot->x-rot->y)>>G3_FPSHIFT].sin;

  coszmx    =g3_sincos[G3_ANGLE(rot->z-rot->x)>>G3_FPSHIFT].cos;
  sinxmz    =g3_sincos[G3_ANGLE(rot->x-rot->z)>>G3_FPSHIFT].sin;

  sinymz    =g3_sincos[G3_ANGLE(rot->y-rot->z)>>G3_FPSHIFT].sin;
  coszmy    =g3_sincos[G3_ANGLE(rot->z-rot->y)>>G3_FPSHIFT].cos;

  sinxpymzpsinxmypz = g3_sincos[G3_ANGLE(rot->x+rot->y-rot->z)>>G3_FPSHIFT].sin+
                      g3_sincos[G3_ANGLE(rot->x-rot->y+rot->z)>>G3_FPSHIFT].sin;
  sinxpypz  =g3_sincos[G3_ANGLE(rot->x+rot->y+rot->z)>>G3_FPSHIFT].sin;
  sinxmymz  =g3_sincos[G3_ANGLE(rot->x-rot->y-rot->z)>>G3_FPSHIFT].sin;
  cosxmzmymcosxpypz = g3_sincos[G3_ANGLE(rot->x-rot->z-rot->y)>>G3_FPSHIFT].cos-
                      g3_sincos[G3_ANGLE(rot->x+rot->y+rot->z)>>G3_FPSHIFT].cos;
  cosxpzmy=g3_sincos[G3_ANGLE(rot->x+rot->z-rot->y)>>G3_FPSHIFT].cos;
  cosxmzpy=g3_sincos[G3_ANGLE(rot->x-rot->z+rot->y)>>G3_FPSHIFT].cos;

  m->m32=-g3_sincos[G3_ANGLE(rot->x)>>G3_FPSHIFT].sin;

  m->m11=((sinxpymzpsinxmypz-sinxmymz-sinxpypz)>>2)+((coszpy+coszmy)>>1);
  m->m12=(sinxpz-sinxmz)>>1;
  m->m13=((cosxmzmymcosxpypz+cosxmzpy-cosxpzmy)>>2)-((sinypz+sinymz)>>1);
  m->m21=((cosxmzmymcosxpypz-cosxmzpy+cosxpzmy)>>2)-((sinypz-sinymz)>>1);
  m->m22=(coszpx+coszmx)>>1;
  m->m23=((sinxpymzpsinxmypz+sinxpypz+sinxmymz)>>2)+((coszmy-coszpy)>>1);
  m->m31=(sinxpy-sinxmy)>>1;
  m->m33=(cosxpy+cosxmy)>>1;
  m->m41=trs->x;
  m->m42=trs->y;
  m->m43=trs->z;

  //  angle = G3_ANGLE(rot->x)>>G3_FPSHIFT;
  //  ptr = &g3_sincos[angle];
  //  sinx=ptr->sin;
  //  cosx=ptr->cos;
  //  angle = G3_ANGLE(rot->y)>>G3_FPSHIFT;
  //  ptr = &g3_sincos[angle];
  //  siny=ptr->sin;
  //  cosy=ptr->cos;
  //  angle = G3_ANGLE(rot->z)>>G3_FPSHIFT;
  //  ptr = &g3_sincos[angle];
  //  sinz=ptr->sin;
  //  cosz=ptr->cos;
  //  tmp=((sinx*siny)>>G3_FPSHIFT);
  //  m->m11=((cosz*cosy)>>G3_FPSHIFT) + ((sinz*tmp)>>G3_FPSHIFT);
  //  m->m21=((cosz * tmp)>>G3_FPSHIFT) - ((sinz*cosy)>>G3_FPSHIFT);
  //  m->m31=(cosx*siny)>>G3_FPSHIFT;
  //  m->m41=trs->x;
  //  m->m12=(sinz*cosx)>>G3_FPSHIFT;
  //  m->m22=(cosz*cosx)>>G3_FPSHIFT;
  //  m->m32=-sinx;
  //  m->m42=trs->y;
  //  tmp=((sinx*cosy)>>G3_FPSHIFT);
  //  m->m13=((tmp*sinz)>>G3_FPSHIFT) - ((cosz*siny)>>G3_FPSHIFT);
  //  m->m23=((sinz*siny)>>G3_FPSHIFT) + ((tmp*cosz)>>G3_FPSHIFT);
  //  m->m33=(cosx*cosy)>>G3_FPSHIFT;
  //  m->m43=trs->z;
#endif

  }

#if 0
void oldg3_matrix_rotzrotxrotytranslate(G3_ROTATION *rot,G3_VECTOR *trs,G3_MATRIX *m)
  {
  FP sinx,siny,sinz;
  FP cosx,cosy,cosz;
  FP tmp;
  sinx=g3_sincos[G3_ANGLE(rot->x)>>G3_FPSHIFT].sin;
  cosx=g3_sincos[G3_ANGLE(rot->x)>>G3_FPSHIFT].cos;
  siny=g3_sincos[G3_ANGLE(rot->y)>>G3_FPSHIFT].sin;
  cosy=g3_sincos[G3_ANGLE(rot->y)>>G3_FPSHIFT].cos;
  sinz=g3_sincos[G3_ANGLE(rot->z)>>G3_FPSHIFT].sin;
  cosz=g3_sincos[G3_ANGLE(rot->z)>>G3_FPSHIFT].cos;
  tmp=((sinx*siny)>>G3_FPSHIFT);
  m->m11=((cosz*cosy)>>G3_FPSHIFT) + ((sinz*tmp)>>G3_FPSHIFT);
  m->m21=((cosz * tmp)>>G3_FPSHIFT) - ((sinz*cosy)>>G3_FPSHIFT);
  m->m31=(cosx*siny)>>G3_FPSHIFT;
  m->m41=trs->x;
  m->m12=(sinz*cosx)>>G3_FPSHIFT;
  m->m22=(cosz*cosx)>>G3_FPSHIFT;
  m->m32=-sinx;
  m->m42=trs->y;
  tmp=((sinx*cosy)>>G3_FPSHIFT);
  m->m13=((tmp*sinz)>>G3_FPSHIFT) - ((cosz*siny)>>G3_FPSHIFT);
  m->m23=((sinz*siny)>>G3_FPSHIFT) + ((tmp*cosz)>>G3_FPSHIFT);
  m->m33=(cosx*cosy)>>G3_FPSHIFT;
  m->m43=trs->z;
  }
#endif


/**************************************************************************/
/*  fog functions                                                         */
/**************************************************************************/
void g3_fogplane_set(FP fogplane)
  {
  g3_fogplane=fogplane;
  }

FP g3_fogplane_get(void)
  {
  return(g3_fogplane);
  }

void g3_fogdepth_set(FP fogdepth)
  {
  g3_fogdepth=fogdepth;
  }

FP g3_fogdepth_get(void)
  {
  return(g3_fogdepth);
  }

/**************************************************************************/
/*  nearplane functions                                                   */
/**************************************************************************/

void g3_nearplane_set(FP nearplane)
  {
  if(nearplane<G3_FPVALUE(16))
  {
    g3_error("g3_nearplane_set : sorry nearplane MUST be > 16)");
  }
  g3_nearplane=nearplane;
  }

FP g3_nearplane_get( void )
  {
  return(g3_nearplane);
  }

/**************************************************************************/
/*  farplane  functions                                                   */
/**************************************************************************/

void g3_farplane_set(FP farplane)
  {
  if(farplane<=g3_nearplane)
    g3_error("g3_farplane_set : sorry farplane must be > nearplane");
  g3_farplane=farplane;
  }

FP g3_farplane_get( void )
  {
  return(g3_farplane);
  }

/**************************************************************************/
/*  graphcall functions                                                   */
/**************************************************************************/

G3_FUNCTION g3_graphcall_get(SD nofunction)
  {
  if(nofunction<0 || nofunction>=G3_GRAPHCALL_COUNT)
    {
    g3_error("g3_graphcall_get : nofunction not in range");
    }
  return(g3_graphcall[nofunction]);
  }

void g3_graphcall_set(SD nofunction,G3_FUNCTION ptrfunction)
  {
  if(nofunction<0 || nofunction>=G3_GRAPHCALL_COUNT)
    {
    g3_error("G3_graphcall_set : nofunction not in range");
    }
  g3_graphcall[nofunction]=ptrfunction;
  }

void g3_graphcall_setall(SD level)
  {
  level=level % G3_GRAPHCALL_LEVEL_MAX;
  switch(level)
    {
    case G3_GRAPHCALL_LEVEL_FULL:
/*  "flat",
    "textured",
    "wireframe",
    "L translucide gouraud",
    "L flat",
    "L gouraud",
    "L texture gouraud",
    "L texture",
    "L texture gouraud R by x",
    "L texture gouraud R by y",
    "L texture gouraud R by z",
    "L translucide gourault",
    "L Wireframe"
    */
      g3_graphcall_set(G3_GRAPHCALL_POLYF     ,(G3_FUNCTION)(&G3_POLYF));
      g3_graphcall_set(G3_GRAPHCALL_POLYT     ,(G3_FUNCTION)(&G3_POLYT));
      g3_graphcall_set(G3_GRAPHCALL_POLYW     ,(G3_FUNCTION)(&G3_POLYW));
      g3_graphcall_set(G3_GRAPHCALL_LTPOLYG   ,(G3_FUNCTION)(&G3_LTPOLYG));
      g3_graphcall_set(G3_GRAPHCALL_LPOLYF    ,(G3_FUNCTION)(&G3_LPOLYF));
      g3_graphcall_set(G3_GRAPHCALL_LPOLYG    ,(G3_FUNCTION)(&G3_LPOLYG));
      g3_graphcall_set(G3_GRAPHCALL_LPOLYTG   ,(G3_FUNCTION)(&G3_LPOLYTG));
      g3_graphcall_set(G3_GRAPHCALL_LPOLYT    ,(G3_FUNCTION)(&G3_LPOLYT));
      g3_graphcall_set(G3_GRAPHCALL_POLYTT    ,(G3_FUNCTION)(&G3_POLYTT));
      g3_graphcall_set(G3_GRAPHCALL_LPOLYTGRY ,(G3_FUNCTION)(&G3_LPOLYTGRY));
      g3_graphcall_set(G3_GRAPHCALL_LPOLYTGRZ ,(G3_FUNCTION)(&G3_LPOLYTGRZ));
      g3_graphcall_set(G3_GRAPHCALL_POLYFW    ,(G3_FUNCTION)(&G3_POLYFW));
      break;
    case G3_GRAPHCALL_LEVEL_ALLWIRE:
      break;
    case G3_GRAPHCALL_LEVEL_ALLFILLEDWIRE:
      g3_graphcall_set(G3_GRAPHCALL_POLYF     ,(G3_FUNCTION)(&G3_POLYF));
      g3_graphcall_set(G3_GRAPHCALL_POLYT     ,(G3_FUNCTION)(&G3_POLYFW));
      g3_graphcall_set(G3_GRAPHCALL_POLYW     ,(G3_FUNCTION)(&G3_POLYFW));
      g3_graphcall_set(G3_GRAPHCALL_LTPOLYG   ,(G3_FUNCTION)(&G3_POLYFW));
      g3_graphcall_set(G3_GRAPHCALL_LPOLYF    ,(G3_FUNCTION)(&G3_POLYFW));
      g3_graphcall_set(G3_GRAPHCALL_LPOLYG    ,(G3_FUNCTION)(&G3_POLYFW));
      g3_graphcall_set(G3_GRAPHCALL_LPOLYTG   ,(G3_FUNCTION)(&G3_POLYFW));
      g3_graphcall_set(G3_GRAPHCALL_LPOLYT    ,(G3_FUNCTION)(&G3_POLYFW));
      g3_graphcall_set(G3_GRAPHCALL_POLYTT    ,(G3_FUNCTION)(&G3_POLYFW));
      g3_graphcall_set(G3_GRAPHCALL_LPOLYTGRY ,(G3_FUNCTION)(&G3_POLYFW));
      g3_graphcall_set(G3_GRAPHCALL_LPOLYTGRZ ,(G3_FUNCTION)(&G3_POLYFW));
      g3_graphcall_set(G3_GRAPHCALL_POLYFW    ,(G3_FUNCTION)(&G3_POLYFW));
      break;
    case G3_GRAPHCALL_LEVEL_ALLFLAT:
      g3_graphcall_set(G3_GRAPHCALL_POLYF     ,(G3_FUNCTION)(&G3_POLYF));
      g3_graphcall_set(G3_GRAPHCALL_POLYT     ,(G3_FUNCTION)(&G3_POLYF));
      g3_graphcall_set(G3_GRAPHCALL_POLYW     ,(G3_FUNCTION)(&G3_POLYF));
      g3_graphcall_set(G3_GRAPHCALL_LTPOLYG   ,(G3_FUNCTION)(&G3_POLYF));
      g3_graphcall_set(G3_GRAPHCALL_LPOLYF    ,(G3_FUNCTION)(&G3_POLYF));
      g3_graphcall_set(G3_GRAPHCALL_LPOLYG    ,(G3_FUNCTION)(&G3_POLYF));
      g3_graphcall_set(G3_GRAPHCALL_LPOLYTG   ,(G3_FUNCTION)(&G3_POLYF));
      g3_graphcall_set(G3_GRAPHCALL_LPOLYT    ,(G3_FUNCTION)(&G3_POLYF));
      g3_graphcall_set(G3_GRAPHCALL_POLYTT    ,(G3_FUNCTION)(&G3_POLYF));
      g3_graphcall_set(G3_GRAPHCALL_LPOLYTGRY ,(G3_FUNCTION)(&G3_POLYF));
      g3_graphcall_set(G3_GRAPHCALL_LPOLYTGRZ ,(G3_FUNCTION)(&G3_POLYF));
      g3_graphcall_set(G3_GRAPHCALL_POLYFW    ,(G3_FUNCTION)(&G3_POLYF));
      break;
    default:
      g3_error("g3_graphcall_setall : level not in range");
      break;
    }
  g3_graphcall_level=level;
  }


void g3_graphcall_setlevel(SD level)
  {
  g3_graphcall_setall(level);
  }

SD g3_graphcall_getlevel(void)
  {
  return(g3_graphcall_level);
  }

void g3_graphcall_reset(void)
  {
  g3_graphcall_setall(G3_GRAPHCALL_LEVEL_FULL);
  }

/**************************************************************************/
/*  init / restore  functions                                             */
/**************************************************************************/

void g3_xzoom_set(FP zoom)
  {
  g3_xzoom=zoom;
  }

FP g3_xzoom_get(void)
  {
  return(g3_xzoom);
  }

void g3_viewskl_set(SD viewskl)
  {
  g3_viewskl=viewskl;
  }

SD g3_viewskl_get(void)
  {
  return(g3_viewskl);
  }


void g3_yzoom_set(FP zoom)
  {
  g3_yzoom=zoom;
  }

FP g3_yzoom_get(void)
  {
  return(g3_yzoom);
  }

void g3_init(SD screenw,SD screenh,SD linesize,SD zoomshift,FP xratio,FP yratio,SD mode32000)
  {
  static SD callcount=0;
  /*----- if it s the first initialisation -----------*/
  if(callcount==0)
    {
    /*---- enable patch assembly code -----*/
    g3_virtual_unprotect((void *)_asm_begin,(UD)&_asm_end-(UD)&_asm_begin,&asmprotect);
    /*---- reset buffers adresss---------------*/

    //g3_ot=NULL;

    g3_ptrbank=NULL;
    g3_colorlight=NULL;

    /*---- set farplane and nearplane -----*/
    g3_nearplane_set(G3_FPVALUE(16));
    g3_farplane_set(G3_FPVALUE(4096));
    /*---- set fogplane and fog depth -----*/
    g3_fogplane_set(((G3_GRIDSIZE/2)-2)<<(G3_FPSHIFT+G3_LANDGRIDSTEPSHIFT));
    g3_fogdepth_set((2)<<(G3_FPSHIFT+G3_LANDGRIDSTEPSHIFT));
    /*------ init the sincos table -----*/
    {
    SD i;
    for(i=0;i<G3_NBANGLE;i++)
      {
      g3_sincos[i].sin=(FP)(sin(G3_RADIAN(i))*(G3_FPUNIT-1));
      g3_sincos[i].cos=(FP)(cos(G3_RADIAN(i))*(G3_FPUNIT-1));
      }
    }
    /*----- reset the graphcall table -------*/
    g3_graphcall_reset();
    /*---------- allocate bank buffer -----------*/
    g3_ptrbank=(PIXEL *)(g3_malloc
            (
            ((256*256)*(G3_MAXBANK)*sizeof(PIXEL))+        // for bank
            ((256*256)*sizeof(DPIXEL)*1)+                  // for colorlight tables 
            ((256*256)*sizeof(DPIXEL))                     // for alignment to 256*256*2
            ));
    {
      SD i;
      PIXEL *ptrbank;
      ptrbank=(PIXEL *)( ((SD)(g3_ptrbank)+(256*256*sizeof(DPIXEL))) & (0xFFFE0000));
      for(i=0;i<G3_MAXBANK;i++)
        {
        g3_bank[i].ptrpixel=ptrbank;
        ptrbank+=(256*256)*sizeof(PIXEL);
        }
      g3_colorlight=(DPIXEL *)ptrbank;
    }
    /*----- next time we do not enter ------*/
    callcount=314116;
    }

  g3_screenw=screenw;
  g3_screenh=screenh;
  g3_linesize=linesize;
  g3_zoomshift=zoomshift;
  g3_xzoom_set(xratio);
  g3_yzoom_set(yratio);

  {
    SD i;
    for(i=0;i<G3_OTMAX;i++) g3_ot[i]=NULL;
  }

  /*---------- init global variables ----------*/
  g3_polyminy=G3_FPMAX;
  g3_polymaxy=G3_FPMIN;
  /*----------- init the yoffset table */
  {
    SD y;
    SD offset;
    offset=0;
    for(y=0;y<g3_screenh;y++)
      {
      g3_yoffset[y]=offset;
      offset+=g3_linesize;
      }
  }
  g3_window_set(0,0,screenw,screenh);
  }

void g3_restore(void)
  {
  g3_free(g3_ptrbank);
  }

/**************************************************************************/
/*  window functions                                                      */
/**************************************************************************/

void g3_window_set(SD x,SD y,SD w,SD h)
  {
  if(x<0||y<0||w<4||h<4)
  {
          g3_error("G3_WINDOW_SET : bad parameters");
  }
  g3_windowx=x;
  g3_windowy=y;
  g3_windoww=w;
  g3_windowh=h;
  g3_xmax=(w-1);
  g3_ymax=(h-1);
  g3_xmaxp1=w;
  g3_ymaxp1=h;
  g3_xmid=(w>>1);
  g3_ymid=(h>>1);
  }


SD g3_window_getx(void)
  {
  return(g3_windowx);
  }

SD g3_window_gety(void)
  {
  return(g3_windowy);
  }

SD g3_window_getw(void)
  {
  return(g3_windoww);
  }

SD g3_window_geth(void)
  {
  return(g3_windowh);
  }

PIXEL *g3_window_getaddr(void)
  {
  PIXEL *ptrpixel;
  ptrpixel=g3_scrbuffer+g3_yoffset[g3_windowy];
  ptrpixel+=g3_windowx*2;
  return(ptrpixel);
  }

/**************************************************************************/
/*  screen functions                                                      */
/**************************************************************************/

SD g3_screen_getw(void)
  {
  return(g3_screenw);
  }

SD g3_screen_geth(void)
  {
  return(g3_screenh);
  }

SD g3_screen_getbytesperline(void)
  {
  return(g3_linesize);
  }

PIXEL *g3_screen_getaddr(void)
  {
  return(g3_scrbuffer);
  }

void g3_screen_setaddr(PIXEL *buffer)
  {
  g3_scrbuffer=buffer;
  g3_cscrbuffer=g3_scrbuffer+g3_yoffset[g3_windowy];
  g3_cscrbuffer+=g3_windowx*2;
  }

SD g3_screen_getyoffset(SD y)
  {
  return(g3_yoffset[y]);
  }

PIXEL *g3_screen_computeaddr(SD x,SD y)
  {
  return(g3_scrbuffer+g3_yoffset[y]+x);
  }

void g3_screen_clear(SD color)
  {
  memset(g3_scrbuffer,(PIXEL)color,g3_linesize*g3_screenh*sizeof(PIXEL));
  }

void g3_screen_black(void)
  {
  memset(g3_scrbuffer,0,g3_linesize*g3_screenh);
  }

G3_COLORREGISTER *g3_palette_getaddr(void)
  {
  return(g3_colorregister);
  }

#if 0
G3_COLORREGISTER *g3_palette_save(void)
  {
  G3_COLORREGISTER *palette;
  palette=(G3_COLORREGISTER *)(g3_malloc(G3_MAXCOLOR*sizeof(G3_COLORREGISTER)));
  if(palette!=NULL)
    {
    memcpy(palette,g3_palette_getaddr(),G3_MAXCOLOR*sizeof(G3_COLORREGISTER));
    }
  return(palette);
  }


void g3_palette_restore(G3_COLORREGISTER *palette)
  {
  memcpy(g3_palette_getaddr(),palette,G3_MAXCOLOR*sizeof(G3_COLORREGISTER));
  }
#endif


/**************************************************************************/
/*  skl management                                                        */
/**************************************************************************/

void g3_skl_sinmotion(FP *value,SD start,SD speed,SD anglestart,SD angleamplitude)
  {
  SD angle;
  angle=(g3_timer_get()*speed+start)&(G3_NBANGLE-1);
  *value=g3_sincos[angle].sin*angleamplitude+anglestart;
  }

void g3_skl_motion(FP *value,SD motiontime,SD start,SD anglestart,SD angleamplitude)
  {
  SD angle;
  angle=(motiontime+start)&(G3_NBANGLE-1);
  *value=g3_sincos[angle].sin*angleamplitude+anglestart;
  }


G3_SKL *g3_skl_select(G3_SKL *ptrskl,SD px,SD py)
  {
  SD mindist,dist;
  G3_SKL *ptrsklcurrent,*ptrsklselect;
  ptrsklselect=ptrsklcurrent=ptrskl;
  mindist=G3_FPMAX;
  while(ptrsklcurrent->recurse!=G3_ENDMARK)
    {
    dist=(ptrsklcurrent->ddtarget.px-px)*(ptrsklcurrent->ddtarget.px-px)+(ptrsklcurrent->ddtarget.py-py)*(ptrsklcurrent->ddtarget.py-py);
    if(dist<mindist)
      {
      mindist=dist;
      ptrsklselect=ptrsklcurrent;
      }
    ptrsklcurrent++;
    }
  return(ptrsklselect);
  }



/**************************************************************************/
/*  sklpos management                                                     */
/**************************************************************************/

G3_SKLPOS *g3_sklpos_new(G3_OBJECT *ptrobject)
  {
  G3_SKLPOS *ptrsklpos,*sklpos;
  G3_SKL *ptrskl;
  ptrskl=ptrobject->ptrskl;
  sklpos=(G3_SKLPOS *)g3_malloc(sizeof(G3_SKLPOS)*g3_skl_count(ptrskl));
  ptrsklpos=sklpos;
  while(ptrskl->recurse!=G3_ENDMARK)
  {
    ptrsklpos->rotation=ptrskl->rotation;
    ptrsklpos->translation=ptrskl->translation;
    ptrskl++;
    ptrsklpos++;
  }
  return(sklpos);
  }

  /*---------------------- copy a skl pos to a skl -------------*/
void g3_sklpos_put(G3_SKLPOS *ptrsklpos,G3_SKL *ptrskl)
  {
  while(ptrskl->recurse!=G3_ENDMARK)
    {
    ptrskl->rotation=ptrsklpos->rotation;
    ptrskl->translation=ptrsklpos->translation;
    ptrsklpos++;
    ptrskl++;
    }
  }

  /*----------------------- copy a skl current pos to a sklpos ------------*/

void g3_sklpos_get(G3_SKL *ptrskl,G3_SKLPOS *ptrsklpos)
  {
  while(ptrskl->recurse!=G3_ENDMARK)
    {
    ptrsklpos->rotation=ptrskl->rotation;
    ptrsklpos->translation=ptrskl->translation;
    ptrsklpos++;
    ptrskl++;
    }
  }


/**************************************************************************/
/*  path management                                                       */
/**************************************************************************/

void g3_path_set(char *path)
  {
  char drive[ _MAX_DRIVE ];
  char dir[ _MAX_DIR ];
  char fname[ _MAX_FNAME ];
  char ext[ _MAX_EXT ];

  _splitpath(path,drive,dir,fname,ext);
  _makepath(g3_path,drive,dir,"","");
  }

char *g3_path_get()
  {
  return(g3_path);
  }



/**************************************************************************/
/*  bank management                                                       */
/**************************************************************************/

SD g3_bank_load(char *filename,SD rshift,SD gshift,SD bshift,SD rcount,SD gcount,SD bcount)
  {
  SD numbank;
  SD w,h;
  PIXEL *ptrpixel;

  char drive[ _MAX_DRIVE ];
  char dir[ _MAX_DIR ];
  char fname[ _MAX_FNAME ];
  char ext[ _MAX_EXT ];
  char bnkname[_MAX_PATH];
  for(numbank=G3_MAXBANK-1;numbank>=0;numbank--)
    {
    /* try to load the pixel buffer */
    sprintf(bnkname,"%s%d",filename,numbank);
    _splitpath(bnkname,drive,dir,fname,ext);
    _makepath(bnkname,drive,dir,fname,".PCX");
    ptrpixel=g3_pcx_load(bnkname,&w,&h);
    if(ptrpixel==NULL) return(G3_FALSE);
    memcpy(g3_bank[numbank].ptrpixel,ptrpixel,256*256);
    g3_free(ptrpixel);
    /* verify size of the bank */
    if(w!=256 || h!=256) return(G3_FALSE);
    }
  g3_colorlight_load(filename,rshift,gshift,bshift,rcount,gcount,bcount);
  return(G3_TRUE);
  }

PIXEL *g3_bank_getaddr(SD numbank)
  {
  return(g3_bank[numbank].ptrpixel);
  }

SD g3_bank_save()
  {
  SD numbank;
  for(numbank=0;numbank<G3_MAXBANK;numbank++)
    {
    }
  return(G3_TRUE);
  }


/**************************************************************************/
/*  polygon management                                                    */
/**************************************************************************/

SD g3_polygon_getbank(G3_POLYGON *ptrpolygon)
{
  return(ptrpolygon->bank);
}

void g3_polygon_setbank(G3_POLYGON *ptrpolygon,SD bank)
{
  ptrpolygon->bank=(bank % G3_MAXBANK);
}


void g3_polygon_rotate(G3_POLYGON *ptrpolygon)
  {
  SD tmp;
    tmp=ptrpolygon->pts1;
    ptrpolygon->pts1=ptrpolygon->pts2;
    ptrpolygon->pts2=ptrpolygon->pts3;
    ptrpolygon->pts3=tmp;
  }


void g3_polygon_invert(G3_POLYGON *ptrpolygon)
  {
  G3_EXCHANGESD(ptrpolygon->pts2,ptrpolygon->pts3);
  }

SD g3_polygon_typeget(G3_POLYGON *ptrpolygon)
  {
  return(ptrpolygon->type);
  }

char *g3_polygon_typenameget(G3_POLYGON *ptrpolygon)
  {
  if(ptrpolygon->type>=0 && ptrpolygon->type<G3_GRAPHCALL_COUNT)
    return(g3_polygontypename[ptrpolygon->type]);
  return(g3_polygonbadtypename);
  }

void g3_polygon_typeput(G3_POLYGON *ptrpolygon,SD type)
  {
  ptrpolygon->type=type % G3_GRAPHCALL_COUNT;
  }

SD g3_polygon_colorget(G3_POLYGON *ptrpolygon)
  {
  return(ptrpolygon->color & 0xFF);
  }

void g3_polygon_colorput(G3_POLYGON *ptrpolygon,SD color)
  {
  color &= 0xFF;
  color=G3_EXTPIXEL(color);
  ptrpolygon->color=color;
  }

G3_POLYGON *g3_polygon_selectinot(SD px,SD py)
  {
  G3_OTITEM *ptrotitem;
  G3_POLYGON *ptrpolygon,*ptrnearestpolygon;
  G3_DVERTEX *objectdvertex,*ptra,*ptrb,*ptrc;
  SD dist,mindist;
  SD dx,dy;
  SD i;
  mindist=G3_FPMAX;
  ptrnearestpolygon=NULL;
  for(i=g3_otmax;i>=g3_otmin;i--)
    {
    ptrotitem=g3_ot[i];
    if(ptrotitem!=NULL)
      {
      do
        {
        objectdvertex=ptrotitem->ptrdvertex;
        ptrpolygon=(G3_POLYGON *)ptrotitem->ptr;
        ptra=G3_OBJECT_DVERTEXADR(objectdvertex,ptrpolygon->pts1);
        ptrb=G3_OBJECT_DVERTEXADR(objectdvertex,ptrpolygon->pts2);
        ptrc=G3_OBJECT_DVERTEXADR(objectdvertex,ptrpolygon->pts3);
        if(g3_isintriangle(ptra->px,ptra->py,ptrb->px,ptrb->py,ptrc->px,ptrc->py,px,py))
          {
          dx=(ptra->px+ptrb->px+ptrc->px)/3;
          dy=(ptra->py+ptrb->py+ptrc->py)/3;
          dx=px-dx;
          dy=py-dy;
          dist=dx*dx+dy*dy;
          if(dist<mindist)
            {
            dist=mindist;
            ptrnearestpolygon=ptrpolygon;
            }
          }
        ptrotitem=(G3_OTITEM *)ptrotitem->next;
        }
      while(ptrotitem!=NULL);
      g3_ot[i]=NULL;
      }
    }
  return(ptrnearestpolygon);
  }

/**************************************************************************/
/*  ot management                                                         */
/**************************************************************************/

void g3_ot_clear(void)
  {
  g3_otmin=G3_FPMAX;             /* set the ot min */
  g3_otmax=G3_FPMIN;             /* set the ot max */
  g3_ptrotitem=&g3_otitem[0];    /* set current otitem to add */
  g3_objectdvertex=g3_dvertex;    /* reinit dvertex buffer pointer */
  g3_ptrsubpoly=g3_subpoly;      /* reinit subpoly buffer */
  g3_otcount=0;
  }


void g3_ot_display(void)
  {
  G3_OTITEM *ptrotitem;
  SD i;
  g3_cscrbuffer=g3_window_getaddr();
  for(i=g3_otmax;i>=g3_otmin;i--)
    {
    ptrotitem=g3_ot[i];
    if(ptrotitem!=NULL)
      {
      do
        {
        g3_objectdvertex=ptrotitem->ptrdvertex;
        (g3_graphcall[ptrotitem->ptr->type])(ptrotitem->ptr);
        ptrotitem=(G3_OTITEM *)ptrotitem->next;
        g3_otcount++;
        }
      while(ptrotitem!=NULL);
      g3_ot[i]=NULL;
      }
    }
  }

SD g3_ot_getcount(void)
  {
    return(g3_otcount);
  }


/**************************************************************************/
/*              OTHER FUNCTION                                            */
/**************************************************************************/

#if 0
SD g3_makefont(PIXEL *data,char *filename)
  {
  SD w,h;
  SD size;
  SD x,y;
  SD cw,ch;
  SD cx,cy;
  G3_SPRITE car[128-32];
  SD offset[128-32];
  SD ncar;
  PIXEL *ptr,*sptr,*dptr,*buffer;
  SD *dwordptr;
  SW *wordptr;
  PIXEL *ptrdata,*ptrpixel;
  G3_FILE *handle;
  wordptr=(SW *)data;
  w=*wordptr++;
  h=*wordptr++;
  ptrdata=(PIXEL *)wordptr;
  ncar=0;
  y=0;
  /* scan the font in sprite */

  while(y<h)
    {
    x=0;
    while(x<w)
      {
      ptr=ptrdata+x+y*w;
      if(*ptr!=0 && *(ptr-1)==0 && *(ptr-w)==0 && *(ptr-w-1)==0)
        {

        /* get sprite width */
        sptr=ptr;
        cw=0;
        while(*sptr!=0)
          {
          cw++;
          sptr++;
          }
        /* get sprite height */
        sptr=ptr;
        ch=0;
        while(*sptr!=0)
          {
          ch++;
          sptr+=w;
          }

        /* store sprite data */


        car[ncar].w=cw;
        car[ncar].h=ch;
        dptr=(PIXEL *)(g3_malloc(cw*ch));
        car[ncar].data=dptr;
        for(cy=0;cy<ch;cy++)
          for(cx=0;cx<cw;cx++)
            {
            sptr=ptr+cx+cy*w;
            if(*sptr==1) *sptr=0;
            *dptr++=*sptr;
            *sptr=2;
            }
        /* scan the next sprite */
        ncar++;
        if(ncar==128) goto stop;
        }
      x++;
      }
    y++;
    }

  stop:

  /* calculate offset of each sprite */

  size=0;
  for(ncar=32;ncar<128;ncar++)
    {
    offset[ncar-32]=size;
    size+=car[ncar-32].w*car[ncar-32].h+2;
    }

  /* allocate font buffer */

  buffer=(PIXEL *)g3_malloc(size+sizeof(SD)*(128-32));

  /* build font buffer offset */

  dwordptr=(SD *)buffer;
  for(ncar=32;ncar<128;ncar++)
    {
    *dwordptr=offset[ncar-32];
    dwordptr++;
    }

  /* build font buffer */

  ptrpixel=(PIXEL *) dwordptr;
  for(ncar=32;ncar<128;ncar++)
    {
    *ptrpixel++=car[ncar-32].w;
    *ptrpixel++=car[ncar-32].h;
    size=car[ncar-32].w*car[ncar-32].h;
    for(w=0;w<size;w++)
      {
      *ptrpixel++=*(car[ncar-32].data+w);
      }
    g3_free(car[ncar-32].data);
    }
  handle=g3_fopen(filename,"wb");
  if(handle!=NULL)
    {
    g3_fwrite(buffer,1,(UD)(ptrpixel)-(UD)(buffer),handle);
    g3_fclose(handle);
    }
  g3_free(buffer);
  return(0);
  }

#endif


