/************************************************************************/
/*  G3_LIB  : 3D GRAPHIC LIBRARY                                        */
/************************************************************************/

#ifndef _G3LIB_H_
#define _G3LIB_H_

/*------------------------------------------------------------------------------*/
/* DEFINITION FOR LIBRARY COMPILING                                             */
/*------------------------------------------------------------------------------*/

#ifdef _G3LIB_C_
  #define _G3LIB_EXTERN_
#else
  #define _G3LIB_EXTERN_ extern
#endif


/*------------------------------------------------------------------------------*/
/* KEYBOARD DEFINITIONS                                                         */
/*------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------*/
/* POLYGON TYPES                                                                */
/*------------------------------------------------------------------------------*/

  #define G3_GRAPHCALL_LEVEL_FULL             0
  #define G3_GRAPHCALL_LEVEL_ALLWIRE          1
  #define G3_GRAPHCALL_LEVEL_ALLFILLEDWIRE    2
  #define G3_GRAPHCALL_LEVEL_ALLFLAT          3
  #define G3_GRAPHCALL_LEVEL_MAX              4


/*------------------------------------------------------------------------------*/
/*  GLOBAL VARIABLES                                                            */
/*------------------------------------------------------------------------------*/

  /*---------- path variables ----------*/
  _G3LIB_EXTERN_ char          g3_path[256];

  /*---------- ARRAYS FOR COLOR-LIGHT MANAGEMENT ---------------*/




  /*---------- ARRAY FOR SINCOS STORAGE -----------*/

  _G3LIB_EXTERN_ G3_SINCOS g3_sincos[G3_NBANGLE];


  /*---------- ot manangment ----------------------------------*/

  _G3LIB_EXTERN_ SD g3_otmax;                         /* value of g3_otmax */
  _G3LIB_EXTERN_ SD g3_otmin;                         /* value of g3_otmin */
  _G3LIB_EXTERN_ G3_OTITEM *g3_ptrotitem;
  _G3LIB_EXTERN_ G3_OTITEM *g3_ot[G3_OTMAX];
  _G3LIB_EXTERN_ G3_OTITEM g3_otitem[G3_MAXPOLY];

  /*---------- graphic engine variables ----------*/


  _G3LIB_EXTERN_ SD g3_viewskl;

  /*---------- bank storage ----------*/

  _G3LIB_EXTERN_ PIXEL         *g3_ptrbank;
  _G3LIB_EXTERN_ SD             g3_bankcount;
  _G3LIB_EXTERN_ G3_BANK        g3_bank[G3_MAXBANK];
  _G3LIB_EXTERN_ G3_DVERTEX    g3_dvertex[G3_MAXVERTEX];


  _G3LIB_EXTERN_ SD            g3_otcount;
  _G3LIB_EXTERN_ G3_POLYGON    *g3_ptrsubpoly;
  _G3LIB_EXTERN_ G3_POLYGON    g3_subpoly[G3_MAXSUBPOLY];

  /*---------- ARRAYS FOR GRAPHIC COMPUTE ---------------*/


  _G3LIB_EXTERN_ FP g3_nearplane;                     /* value of near clipping plane */
  _G3LIB_EXTERN_ FP g3_farplane;                      /* value of far clipping plane */
  _G3LIB_EXTERN_ FP g3_fogplane;
  _G3LIB_EXTERN_ FP g3_fogdepth;




  /*---------- camera storage ----------*/

  _G3LIB_EXTERN_ G3_CAMERA     g3_camera;
  /*---------- shape storage ------------*/


  _G3LIB_EXTERN_ G3_DVERTEX    *g3_objectdvertex;

  /*---------- font variables ----------*/

  _G3LIB_EXTERN_ G3_COLORREGISTER g3_colorregister[G3_MAXCOLOR]; /* current palette color */

/*------------------------------------------------------------------------------*/
/*  ASSEMBLY PROTOTYPE                                                          */
/*------------------------------------------------------------------------------*/

  void _g3_red(void);
  void _g3_blue(void);
  void _g3_black(void);
  void _g3_polyf_fill(SD color);
  void _g3_polyt_fill(PIXEL *ptrtexture);
  void _g3_lpolyg_fill(UD color);
  void _g3_lpolytg_fill(PIXEL *ptrtexture);
  void _g3_lpolyt_fill(PIXEL *ptrtexture,SD light);
  void _g3_polyf_edge(SD xa,SD ya,SD xb,SD yb);

/*------------------------------------------------------------------------------*/
/*  C PROTOTYPES                                                                */
/*------------------------------------------------------------------------------*/

  void g3_particle_init(void);
  void g3_particle_restore(void);
  void g3_particle_polyadd(FP x,FP y,FP z,SD color,SD life);
  void g3_particle_manage(void);
  void g3_particle_compute(void);


/*---------- keyboard functions ----------*/

  void __interrupt __far g3_newint9();
  void g3_keyboard_init(void);
  void g3_keyboard_restore(void);

/*---------- mathematical functions ----------*/


  void g3_dvertex_project(G3_DVERTEX *ptrdvertex);
  SD g3_vector_norm(G3_VECTOR *v);
  void g3_vector_normalize(G3_VECTOR *v);
  SD g3_dwordsqrt(FP v);
  SD g3_angle2angle(SD a,SD b);
  SD g3_clipping3d(FP xinf,FP zinf,FP xsup,FP zsup);
  SD g3_pvmaxdistance(G3_DVERTEX *ptra,G3_DVERTEX *ptrb,G3_DVERTEX *ptrc);
  SD g3_pvmindistance(G3_DVERTEX *ptra,G3_DVERTEX *ptrb,G3_DVERTEX *ptrc);
  SD g3_pv2d(SD x1,SD y1,SD x2,SD y2,SD x3,SD y3);
  SD g3_isintriangle(SD x1,SD y1,SD x2,SD y2,SD x3,SD y3,SD x,SD y);
  SD g3_isinquad(SD x1,SD y1,SD x2,SD y2,SD x3,SD y3,SD x4,SD y4,SD x,SD y);

  void g3_vertex_mult_matrix(G3_VERTEX *src,G3_MATRIX *m,G3_DVERTEX *dst);
  void g3_vertex_set(FP x,FP y,FP z,G3_VERTEX *rv);
  void g3_vertex_sub(G3_VERTEX *va,G3_VERTEX *vb,G3_VERTEX *rv);
  void g3_vertex_add(G3_VERTEX *va,G3_VERTEX *vb,G3_VERTEX *rv);
  void g3_vertex_mul(G3_VERTEX *va,G3_VERTEX *vb,G3_VERTEX *rv);

  void g3_matrix_null(G3_MATRIX *m);
  void g3_matrix_rotx(SD angle,G3_MATRIX *m);
  void g3_matrix_roty(SD angle,G3_MATRIX *m);
  void g3_matrix_rotz(SD angle,G3_MATRIX *m);
  void g3_matrix_rotzrotxroty(G3_ROTATION *rot,G3_MATRIX *m);
  void g3_matrix_rotzrotxrotytranslate(G3_ROTATION *rot,G3_VECTOR *trs,G3_MATRIX *m);
  void g3_matrix_theta(G3_VECTOR *v,G3_MATRIX *m);
  void g3_matrix_phi(G3_VECTOR *v,G3_MATRIX *m);
  void g3_matrix_translate(G3_VECTOR *v,G3_MATRIX *m);
  void g3_matrix_invtranslate(G3_VECTOR *v,G3_MATRIX *m);
  void g3_matrix_scale(G3_VERTEX *v,G3_MATRIX *rm);
  void g3_matrix_mult(G3_MATRIX *a,G3_MATRIX *b,G3_MATRIX *rm);
  void g3_matrix_identity(G3_MATRIX *m);
  SD g3_matrix_getrotation(G3_MATRIX *m,G3_ROTATION *r);
  void g3_matrix_invert(G3_MATRIX *sm,G3_MATRIX *rm);
  void g3_matrix_copy(G3_MATRIX *s,G3_MATRIX *d);
  void g3_matrix_xyzoom(G3_MATRIX *m);

/*---------- graphic management functions ----------*/


  void g3_init(SD screenw,SD screenh,SD linewidth,SD zoomshift,FP xratio,FP yratio,SD mode32000);
  void g3_restore(void);
  void g3_setcolorregister(SD color,UB red,UB green,UB blue);
  void g3_setallcolorregister(void);
  void g3_pixel_put(SD xa,SD ya,PIXEL color);
  void g3_marker(SD x,SD y,PIXEL color);
  void g3_clipxyz(FP ax,FP ay,FP az,FP bx,FP by,FP bz,FP *nx,FP *ny,FP *nz);
  void g3_plot(FP x,FP y,FP z,G3_MATRIX *m);
  void g3_ot_display(void);
  SD g3_ot_getcount(void);
  void g3_path_set(char *path);
  char *g3_path_get();
  void g3_ot_clear(void);
  void g3_nearplane_set(FP nearplane);
  FP g3_nearplane_get( void );
  void g3_fogplane_set(FP fogplane);
  FP g3_fogplane_get(void);
  void g3_fogdepth_set(FP fogdepth);
  FP g3_fogdepth_get(void);
  void g3_videomode_switch(void);
  SD g3_videomode_get(void);
  void g3_screen_clear(SD color);
  SD g3_screen_getw(void);
  SD g3_screen_geth(void);
  SD g3_screen_getyoffset(SD y);
  PIXEL *g3_screen_computeaddr(SD x,SD y);
  PIXEL *g3_screen_getaddr(void);
  void g3_screen_setaddr(PIXEL *buffer);
  void g3_window_set(SD x,SD y,SD w,SD h);
  PIXEL *g3_window_getaddr(void);
  void g3_window_clear(SD color);
  SD g3_window_getx(void);
  SD g3_window_gety(void);
  SD g3_window_getw(void);
  SD g3_window_geth(void);
  void g3_viewskl_set(SD viewskl);
  SD g3_viewskl_get(void);
  SD g3_bank_load(char *filename,SD rshift,SD gshift,SD bshift,SD rcount,SD gcount,SD bcount);
  PIXEL *g3_bank_getaddr(SD numbank);
  G3_DVERTEX *g3_ot_addsubdivide(SD dist,G3_POLYGON *ptrpoly,G3_DVERTEX *ptrdvertex);


/*---------- object management functions ----------*/

  G3_POLYGON *g3_object_envaddr(G3_OBJECT *ptrobject);
  void g3_object_prepare(G3_OBJECT *ptrobject);
  void g3_object_compute(G3_OBJECT *ptrobject);
  void g3_object_motionprepare(G3_OBJECT *ptrobject);
  void g3_object_motionprogress(G3_OBJECT *ptrobject);
  SD g3_object_vtxload(char *name,G3_OBJECT *ptrobject);
  SD g3_object_vtxsave(G3_OBJECT *ptrobject,char *name);
  SD g3_object_vtxcount(G3_OBJECT *ptrobject);
  SD g3_object_envload(char *name,G3_OBJECT *ptrobject);
  SD g3_object_envsave(G3_OBJECT *ptrobject,char *name);
  SD g3_object_envcount(G3_OBJECT *ptrobject);
  SD g3_object_sklcount(G3_OBJECT *ptrobject);
  SD g3_object_sklsave(G3_OBJECT *ptrobject,char *name);
  SD g3_object_sklload(char *name,G3_OBJECT *ptrobject);
  SD g3_object_load(char *name,G3_OBJECT *ptrobject);
  SD g3_object_save(G3_OBJECT *ptrobject,char *name);
  G3_SKL *g3_object_sklselect(G3_OBJECT *ptrobject,SD px,SD py);

/*----------- editor functions -----------*/

  G3_SKL *g3e_sklselectbypxpy(G3_SKL *ptrskl,SD px,SD py);
  SD g3e_skl_count(G3_SKL *ptrskl);
  G3_SKLPOS *g3_sklpos_new(G3_OBJECT *ptrobject);
  G3_ANIFRAME *g3e_aniframe_new(G3_OBJECT *ptrobject);
  void g3_sklpos_put(G3_SKLPOS *ptrsklpos,G3_SKL *ptrskl);
  void g3_sklpos_get(G3_SKL *ptrskl,G3_SKLPOS *ptrsklpos);
  SD g3_skl_count(G3_SKL *ptrskl);
  void g3_skl_sinmotion(SD *value,SD start,SD speed,SD anglestart,SD angleamplitude);
  G3_SKL *g3_skl_select(G3_SKL *ptrskl,SD px,SD py);

/*----------- other functions ----------*/

  UD g3_byteflipdword(UD number);
  UW g3_byteflipword(UW number);
  void g3_setpolarcamera(G3_ROTATION *rotation,G3_VECTOR *translation);


  G3_POLYGON *g3_polygon_selectinot(SD px,SD py);
  char *g3_polygon_typenameget(G3_POLYGON *ptrpolygon);
  void g3_polygon_invert(G3_POLYGON *ptrpolygon);
  void g3_polygon_colorput(G3_POLYGON *ptrpolygon,SD color);
  void g3_polygon_typeput(G3_POLYGON *ptrpolygon,SD type);
  SD g3_polygon_typeget(G3_POLYGON *ptrpolygon);
  SD g3_polygon_colorget(G3_POLYGON *ptrpolygon);
  void g3_polygon_rotate(G3_POLYGON *ptrpolygon);


  void g3_camera_set(SD rx,SD ry,SD rz,FP tx,FP ty,FP tz);
  void g3_camera_progress(SD drx,SD dry,SD drz,FP dtx,FP dty,FP dtz);
  void g3_camera_lockonobject(G3_OBJECT *ptrobject);
  FP g3_camera_getx(void);
  FP g3_camera_gety(void);
  FP g3_camera_getz(void);
  void g3_camera_setx(FP x);
  void g3_camera_sety(FP y);
  void g3_camera_setz(FP z);
  SD g3_camera_getrx(void);
  SD g3_camera_getry(void);
  SD g3_camera_getrz(void);
  void g3_camera_setrx(SD x);
  void g3_camera_setry(SD y);
  void g3_camera_setrz(SD z);
  void g3_camera_polar();
  void g3_camera_lookat(G3_VECTOR *viewpoint);


  void g3_graphcall_setlevel(SD level);
  SD g3_graphcall_getlevel(void);



  #define G3_GRAPHCALL_POLYF       0            // polygon flat
  #define G3_GRAPHCALL_POLYT       1            // polygon textured
  #define G3_GRAPHCALL_POLYW       2            // polygon wireframe
  #define G3_GRAPHCALL_LPOLYF      3            // lightened polygons flat
  #define G3_GRAPHCALL_LPOLYG      4            // lightened polygon gouraud
  #define G3_GRAPHCALL_LPOLYTG     5            // lightened polygon textured gouraud
  #define G3_GRAPHCALL_LPOLYT      6            // lightened polygon textured flat
  #define G3_GRAPHCALL_POLYTT      7            // polygon transparent textured
  #define G3_GRAPHCALL_LPOLYTGRY   8            // lightened polygon textured gourauld reflection by y
  #define G3_GRAPHCALL_LPOLYTGRZ   9            // lightened polygon textured gourauld reflection by z
  #define G3_GRAPHCALL_LTPOLYG     10           // lightened transparent gouraud
  #define G3_GRAPHCALL_POLYFW      11           // lightened polygon textured gourauld
  #define G3_GRAPHCALL_COUNT       12

  void g3_polyf_edge(SD xa,SD ya,SD xb,SD yb);
  void g3_polyf_fill(SD color);
  void g3_lpolyg_fill(SD color);

  void g3_polyf(G3_POLYGON *polybuffer);
  void g3_lpolyf(G3_POLYGON *polybuffer);
  void g3_polyft_fill(SD color);
  void g3_polyft(G3_POLYGON *polybuffer);
  void g3_polyfts_fill();
  void g3_polyfts(G3_POLYGON *polybuffer);
  void g3_polyftl_fill();
  void g3_polyftl(G3_POLYGON *polybuffer);
  void g3_polyw(G3_POLYGON *polybuffer);
  void g3_polyfw(G3_POLYGON *ptrpoly);

  void g3_lpolyg(G3_POLYGON *polybuffer);
  void g3_ltpolyg(G3_POLYGON *polybuffer);

  void g3_polyt_edge(SD xa,SD ya,SD xb,SD yb,SD txa,SD tya,SD txb,SD tyb);
  void g3_polyt_fill(PIXEL *ptrtexture);
  void g3_polytt_fill(PIXEL *ptrtexture);
  void g3_polyt(G3_POLYGON *polybuffer);

  void g3_lpolytg_edge(SD xa,SD ya,SD xb,SD yb,SD txa,SD tya,SD txb,SD tyb,SD lighta,SD lightb);
  void g3_lpolytg(G3_POLYGON *ptrpoly);
  void g3_lpolytg_fill(PIXEL *ptrtexture);

  void g3_lpolytg(G3_POLYGON *ptrpoly);
  void g3_polytt(G3_POLYGON *ptrpoly);

  void g3_lpolytgry(G3_POLYGON *ptrpoly);
  void g3_lpolytgrz(G3_POLYGON *ptrpoly);
  void g3_lpolyt_fill(PIXEL *ptrtexture,SD color);
  void g3_lpolyt(G3_POLYGON *ptrpoly);


  void g3_line(SD x1,SD y1,SD x2,SD y2,SD color);
  void g3_rectfill(SD x1,SD y1,SD x2,SD y2,PIXEL color);
  void g3_lineh(SD x1,SD y1,SD x2,PIXEL color);
  void g3_linev(SD x1,SD y1,SD y2,PIXEL color);
  void g3_rect(SD x1,SD y1,SD x2,SD y2,PIXEL color);
  void g3_rectboxed(SD x1,SD y1,SD x2,SD y2,PIXEL color,PIXEL edge);
  void g3_rectlightened(SD x1,SD y1,SD x2,SD y2,PIXEL color,PIXEL edgelightened,PIXEL edgeshadowed);

  void g3_sprite(PIXEL *srcdata,SD srcx,SD srcy,SD srcw,SD srch,SD dstx,SD dsty);
  void g3_sprite_zoom(PIXEL *srcdata,SD srcx,SD srcy,SD srcw,SD srch,SD dstx,SD dsty, SD dstw,SD dsth);

  UB *g3_font_load(char *filename);
  SD g3_char_put(PIXEL *font,SD x,SD y,char c,PIXEL color);
  SD g3_string_put(PIXEL *font,SD x,SD y,char *string,PIXEL color);
  SD g3_char_getwh(PIXEL *font,char c,SD *h,SD *w);
  void g3_string_getwh(PIXEL *font,char *string,SD *tw,SD *th);


  void g3_object_setplayer(G3_OBJECT *ptrobject);
  G3_OBJECT *g3_object_getplayer(void);
  int g3_object_getflag(G3_OBJECT *ptrobject,UB flag);
  void g3_object_setflag(G3_OBJECT *ptrobject,UB flag,int value);
  G3_OBJECT *g3_object_getplayer(void);

  G3_OBJECT *g3_object_add(char *name,G3_FUNCOBJECT function,G3_OBJECT *creator,G3_OBJECT *goal,G3_VECTOR *translation,G3_ROTATION *rotation,SD life,SD power);
  void g3_vector_mult_matrix(G3_VECTOR *src,G3_MATRIX *m,G3_VECTOR *dst);

void g3_allobject_compute(void);
void g3_allobject_init(void);
void g3_allobject_restore(void);
void g3_allobject_manage(void);


  #define G3_POLYF_EDGE                           g3_polyf_edge
  #define G3_POLYF_FILL                           _g3_polyf_fill

  #define G3_POLYF                                g3_polyf
  #define G3_LPOLYF                               g3_lpolyf

  #define G3_LTPOLYG_FILL                         g3_ltpolyg_fill
  #define G3_LTPOLYG                              g3_ltpolyg

  #define G3_LPOLYG_EDGE                          g3_lpolyg_edge
  #define G3_LPOLYG_FILL                          g3_lpolyg_fill
  #define G3_LPOLYG                               g3_lpolyg
  #define G3_POLYT_EDGE                           g3_polyt_edge
  #define G3_POLYT_FILL                           g3_polyt_fill
  #define G3_POLYTT                               g3_polytt
  #define G3_POLYTT_FILL                          g3_polytt_fill
  #define G3_POLYT                                g3_polyt

  #define G3_LPOLYT_FILL                          g3_lpolyt_fill
  #define G3_LPOLYT                               g3_lpolyt
  #define G3_LPOLYTG_EDGE                         g3_lpolytg_edge


  #define G3_LPOLYTG_FILL                         _g3_lpolytg_fill

  #define G3_LPOLYTG                              g3_lpolytg

  #define G3_LPOLYTGRY                            g3_lpolytgry
  #define G3_LPOLYTGRZ                            g3_lpolytgrz
  #define G3_POLYW                                g3_polyw
  #define G3_POLYFW                               g3_polyfw


#endif


