
#include <stdlib.h>
#include <math.h>

#include "g3def.h"
#include "g3math.h"
#include "g3error.h"



SD g3_dwordsqrt(SD v)
  {
  SD i;
  UD result,tmp;
  UD low,high;
  if(v<0)
  {
    g3_error("g3_dwordsqrt : argument < 0 !");
    v=-v;
  }
  low=v;
  high=result=0;
  i=16;
  do
    {
    result+=result;
    high=(high << 2) | ((low>>30) & 0x3);
    low<<=2;
    tmp=result+result+1;
    if (high >= tmp)
      {
      result++;
      high-=tmp;
      }
    }
  while(--i);
  if (v-(result*result)>=(result-1)) result++;
  return(result);
  }


/*
a'=a+va*t;
b'=b+vb*t;

a'-b'=ra+rb <=>

ra+rb=a+va*t-b-vb*t;
(ra+rb-a+b)/(va-vb)=t;
*/


SD g3_segment_timecollid
  (
  FP a,FP ra,FP va,
  FP b,FP rb,
  SD tmax,
  SD *timemin,SD *timemax
  )
{
  SD t1,t2;
  // we are in collid on this axis at time 0 ?
  if(va!=0)
  {
    // compute times of collid
    t1=(ra+rb-a+b)/(va);
    t2=(rb+ra-b+a)/(-va);
    // sort times of collid
    if(t1>t2)
    {
      SD tmp;
      tmp=t1;
      t1=t2;
      t2=tmp;
    }
    //intersect with time interval
    t1=G3_MAX(t1,0);
    t2=G3_MIN(t2,tmax);
    //if time interval invalid -> no collid possible
    if(t1>t2)
      {
        return(G3_FALSE);
      }
    //collid occurs in interval  
    *timemin=t1;
    *timemax=t2;  
    return(G3_TRUE);
  }
  else
  {
    if(abs(a-b)<=ra+rb)
    {
      //we are always in collid and speeds are equals -> we are in collid all time  
      *timemin=0;
      *timemax=tmax;  
      return(G3_TRUE);
    }
    else
    {  
      //we are not in collid and speeds are equal -> it's impossible to collid
      return(G3_FALSE);
    }
  }
}

SD g3_cube_timecollid
(
  G3_VECTOR *p1,FP r1,G3_VECTOR *v1,
  G3_VECTOR *p2,FP r2,
  SD tmax,
  SD *timecollid
  )
{
  SD timemin,timeminx,timeminy,timeminz;
  SD timemax,timemaxx,timemaxy,timemaxz;
  SD axis;
  if(
      g3_segment_timecollid
        (
        p1->x,r1,v1->x,
        p2->x,r2,
        tmax,
        &timeminx,&timemaxx
        )
      &&
      g3_segment_timecollid
        (
        p1->y,r1,v1->y,
        p2->y,r2,
        tmax,
        &timeminy,&timemaxy
        )
      &&
      g3_segment_timecollid
        (
        p1->z,r1,v1->z,
        p2->z,r2,
        tmax,
        &timeminz,&timemaxz
        )
      )
  {
    //seach the maximum of minimum global time of intersect
    timemin=timeminx;
    timemax=timemaxx;

    axis=G3_XAXIS;
    if(timeminy>timemin)
    {
      timemin=timeminy;
      axis=G3_YAXIS;
    }
    if(timemaxy<timemax)
    {
      timemax=timemaxy;
    }
    if(timemin>timemax) return(0);

    if(timeminz>timemin)
    {
      timemin=timeminz;
      axis=G3_ZAXIS;
    }
    if(timemaxz<timemax)
    {
      timemax=timemaxz;
    }
    if(timemin>timemax) return(0);

    *timecollid=timemin;
    return(axis);
  }
  return(0);
}

g3_matrix_fromtriangle(G3_VECTOR *ptra,G3_VECTOR *ptrb,G3_VECTOR *ptrc,G3_MATRIX *matrix)
{
    float nx1,ny1,nz1;
    float nx2,ny2,nz2;
    float nx3,ny3,nz3;
    float v1x,v1y,v1z;
    float v2x,v2y,v2z;
    float lenght1,lenght2,lenght3;

    // compute normal of side
    v1x=(ptrb->x-ptra->x)/(1<<G3_FPSHIFT);
    v1y=(ptrb->y-ptra->y)/(1<<G3_FPSHIFT);
    v1z=(ptrb->z-ptra->z)/(1<<G3_FPSHIFT);
    v2x=(ptrc->x-ptra->x)/(1<<G3_FPSHIFT);
    v2y=(ptrc->y-ptra->y)/(1<<G3_FPSHIFT);
    v2z=(ptrc->z-ptra->z)/(1<<G3_FPSHIFT);
    nx1=(v1y*v2z)-(v2y*v1z);
    ny1=(v2x*v1z)-(v1x*v2z);
    nz1=(v1x*v2y)-(v2x*v1y);
    lenght1=(nx1*nx1)+(ny1*ny1)+(nz1*nz1);
    lenght1=sqrt(lenght1);
    nx1=nx1/lenght1;
    ny1=ny1/lenght1;
    nz1=nz1/lenght1;
    // normalize vector A->B
    lenght2=(v1x*v1x)+(v1y*v1y)+(v1z*v1z);
    lenght2=sqrt(lenght2);
    nx2=v1x/lenght2;
    ny2=v1y/lenght2;
    nz2=v1z/lenght2;
    // find normal of (nx,ny,nz) (v1x,v1y,v1z)
    nx3=(ny1*nz2)-(ny2*nz1);//20:10
    ny3=(nx2*nz1)-(nx1*nz2);
    nz3=(nx1*ny2)-(nx2*ny1);
    lenght3=(nx3*nx3)+(ny3*ny3)+(nz3*nz3);
    lenght3=sqrt(lenght3);
    nx3=(nx3/lenght3);
    ny3=(ny3/lenght3);
    nz3=(nz3/lenght3);

    matrix->m13=(FP)(nx1*(1<<G3_FPSHIFT));
    matrix->m23=(FP)(ny1*(1<<G3_FPSHIFT));
    matrix->m33=(FP)(nz1*(1<<G3_FPSHIFT));

    matrix->m11=(FP)(nx2*(1<<G3_FPSHIFT));
    matrix->m21=(FP)(ny2*(1<<G3_FPSHIFT));
    matrix->m31=(FP)(nz2*(1<<G3_FPSHIFT));
    
    matrix->m12=(FP)(nx3*(1<<G3_FPSHIFT));
    matrix->m22=(FP)(ny3*(1<<G3_FPSHIFT));
    matrix->m32=(FP)(nz3*(1<<G3_FPSHIFT));
    
    matrix->m41=0;
    matrix->m42=0;
    matrix->m43=0;
}
