

#include "g3def.h"
#include "g3error.h"
#include "g3malloc.h"
#include "g3lib.h"
#include "g3graph.h"
#include "g3file.h"
#include "g3shape.h"
#include "g3env.h"
#include "g3vtx.h"
#include "g3skl.h"
#include "g3objgrd.h"
#include "g3shape.h"
#include "g3land.h"
#include "g3obj.h"

  /*---------- object storage ------------*/
  SD            g3_objectcount;
  SD            g3_objectactivecount;
  G3_OBJECT     g3_object[G3_MAXOBJECT];
  UB            g3_objectflag[G3_MAXOBJECT];
  G3_OBJECT     *g3_objectplayer;

/**************************************************************************/
/* object management                                                      */
/**************************************************************************/

G3_OBJECT *g3_object_getaddr(SD id)
{
return(&g3_object[id]);
}

SD g3_object_getcount(void)
  {
  return(g3_objectcount);
  }

SD g3_object_getactivecount(void)
  {
  return(g3_objectactivecount);
  }

G3_SKL *g3_object_sklprepare(G3_OBJECT *ptrobject,G3_SKL *ptrskl,G3_MATRIX *m)
  {
  G3_MATRIX mrt;
  G3_MATRIX *mskl;
  SD recurse;
  recurse=ptrskl->recurse;
  /*---------------- while sklitem are on the same level -----*/
  while(ptrskl->recurse==recurse)
    {
    /*--------------compute matrix of current skl item -------*/
    g3_matrix_rotzrotxrotytranslate(&ptrskl->rotation,&ptrskl->translation,&mrt);
    mskl=&ptrskl->matrix;
    /*--------------combine with parent matrix ----------------*/
    g3_matrix_mult(&mrt,m,mskl);
    /*--------------calculate target y --------------------------*/
    ptrskl->dtarget.x=(ptrskl->target.x*mskl->m11)+(ptrskl->target.y*mskl->m21)+(ptrskl->target.z*mskl->m31)+mskl->m41;
    ptrskl->dtarget.y=(ptrskl->target.x*mskl->m12)+(ptrskl->target.y*mskl->m22)+(ptrskl->target.z*mskl->m32)+mskl->m42;
    ptrskl->dtarget.z=(ptrskl->target.x*mskl->m13)+(ptrskl->target.y*mskl->m23)+(ptrskl->target.z*mskl->m33)+mskl->m43;
    ptrskl->starget.x=mskl->m41;
    ptrskl->starget.y=mskl->m42;
    ptrskl->starget.z=mskl->m43;
    /*------------ compute min and max -------------------*/
    if(ptrskl->dtarget.x<ptrobject->mintarget.x) ptrobject->mintarget.x=ptrskl->dtarget.x;
    if(ptrskl->dtarget.y<ptrobject->mintarget.y) ptrobject->mintarget.y=ptrskl->dtarget.y;
    if(ptrskl->dtarget.z<ptrobject->mintarget.z) ptrobject->mintarget.z=ptrskl->dtarget.z;
    if(ptrskl->dtarget.x>ptrobject->maxtarget.x) ptrobject->maxtarget.x=ptrskl->dtarget.x;
    if(ptrskl->dtarget.y>ptrobject->maxtarget.y) ptrobject->maxtarget.y=ptrskl->dtarget.y;
    if(ptrskl->dtarget.z>ptrobject->maxtarget.z) ptrobject->maxtarget.z=ptrskl->dtarget.z;
    /*--------------compute vertex from current sklitem-------*/
    ptrskl++;
    if(ptrskl->recurse>recurse) ptrskl=g3_object_sklprepare(ptrobject,ptrskl,mskl);
    }
  return(ptrskl);
  }

void g3_object_prepare(G3_OBJECT *ptrobject)
  {
  G3_VECTOR localtranslation;
  G3_MATRIX mat1,mat2,inv;
  memcpy(&inv,&g3_camera.invmatrix,sizeof(G3_MATRIX));
  inv.m41=0;
  inv.m42=0;
  inv.m43=0;
  ptrobject->mintarget.x=G3_FPMAX;              /* init of object parameters     */
  ptrobject->mintarget.y=G3_FPMAX;
  ptrobject->mintarget.z=G3_FPMAX;
  ptrobject->maxtarget.x=G3_FPMIN;
  ptrobject->maxtarget.y=G3_FPMIN;
  ptrobject->maxtarget.z=G3_FPMIN;
  G3_TORE_RELATIVEXYZ(
    g3_camera.matrix.m41,
    g3_camera.matrix.m42,
    g3_camera.matrix.m43,
    ptrobject->translation.x,
    ptrobject->translation.y,
    ptrobject->translation.z,
    localtranslation.x,
    localtranslation.y,
    localtranslation.z
    )
  //localtranslation.x=ptrobject->translation.x-g3_camera.matrix.m41;
  //localtranslation.y=ptrobject->translation.y-g3_camera.matrix.m42;
  //localtranslation.z=ptrobject->translation.z-g3_camera.matrix.m43;
  g3_matrix_rotzrotxrotytranslate(&ptrobject->rotation,&localtranslation,&mat1);
  g3_matrix_mult(&mat1,&inv,&mat2);
  g3_matrix_rotzrotxrotytranslate(&ptrobject->rotation,&ptrobject->translation,&ptrobject->matrix);
  if(ptrobject->ptrskl==NULL) return;
  g3_object_sklprepare(ptrobject,ptrobject->ptrskl,&mat2);
  }



G3_POLYGON *g3_object_envaddr(G3_OBJECT *ptrobject)
  {
  return(ptrobject->ptrpoly);
  }


SD g3_object_envcount(G3_OBJECT *ptrobject)
  {
  return(g3_env_count(ptrobject->ptrpoly));
  }


SD g3_object_envsave(G3_OBJECT *ptrobject,char *name)
  {
    return g3_env_save(ptrobject->ptrpoly,name);
  }


SD g3_object_envload(char *name,G3_OBJECT *ptrobject)
  {
  ptrobject->ptrpoly=g3_env_load(name);
  return(ptrobject->ptrpoly!=NULL);
  }

SD g3_object_vtxcount(G3_OBJECT *ptrobject)
  {
  return g3_vtx_count(ptrobject->ptrvertex);
  }

SD g3_object_vtxsave(G3_OBJECT *ptrobject,char *name)
  {
  return g3_vtx_save(ptrobject->ptrvertex,name);
  }

SD g3_object_vtxload(char *name,G3_OBJECT *ptrobject)
  {
  ptrobject->ptrvertex=g3_vtx_load(name);
  return(ptrobject->ptrvertex!=NULL);
  }

SD g3_object_sklcount(G3_OBJECT *ptrobject)
  {
  return(g3_skl_count(ptrobject->ptrskl));
  }

SD g3_object_sklsave(G3_OBJECT *ptrobject,char *name)
  {
  return g3_skl_save(ptrobject->ptrskl,name);
  }

SD g3_object_sklload(char *name,G3_OBJECT *ptrobject)
  {
  ptrobject->ptrskl=g3_skl_load(name);
  return(ptrobject->ptrskl!=NULL);
  }

void g3_object_load(char *name,G3_OBJECT *ptrobject)
  {
  G3_SHAPE *ptrshape;
  if(name==NULL)
  {
    ptrobject->ptrvertex=NULL;
    ptrobject->ptrskl=NULL;
    ptrobject->ptrpoly=NULL;
    return;
  }
  ptrshape=g3_shape_search(name);
  if(ptrshape==NULL)
  {
    ptrshape=g3_shape_load(name);
  }
  ptrobject->ptrvertex=ptrshape->ptrvertex;
  ptrobject->ptrskl=ptrshape->ptrskl;
  ptrobject->ptrpoly=ptrshape->ptrpoly;
  }

SD g3_object_save(G3_OBJECT *ptrobject,char *name)
  {
  g3_object_sklsave(ptrobject,name);
  g3_object_vtxsave(ptrobject,name);
  g3_object_envsave(ptrobject,name);
  return(G3_TRUE);
  }

G3_SKL *g3_object_sklselect(G3_OBJECT *ptrobject,SD px,SD py)
  {
  return(g3_skl_select(ptrobject->ptrskl,px,py));
  }

void g3_object_compute(G3_OBJECT *ptrobject)
  {
  G3_MATRIX matrix;
  G3_VERTEX *ptrvertex;
  G3_DVERTEX *ptrdvertex;
  G3_SKL *ptrskl;
  G3_POLYGON *ptrpoly;
  G3_DVERTEX *ptra,*ptrb,*ptrc;
  SD m11m21,m12m22,m13m23;
  SD dist;
  SD lightoffset;

  if(ptrobject->ptrvertex==NULL) return;
  ptrvertex=ptrobject->ptrvertex;
  /* first compute vertex of the object */
  ptrdvertex=g3_objectdvertex;
  ptrskl=ptrobject->ptrskl;
  if(ptrobject->scale!=G3_FPVALUE(1.0))
  {
    while(ptrskl->recurse!=G3_ENDMARK)             /* for each skl, compute vertex list */
      {
      /*--------------compute vertex from current sklitem-------*/
      memcpy(&matrix,&ptrskl->matrix,sizeof(G3_MATRIX));
      g3_matrix_scalexyz(&matrix,ptrobject->scale);
      g3_matrix_xyzoom(&matrix);
      do
        {
        ptrdvertex->x=(G3_FPMULS(matrix.m11,ptrvertex->x)+G3_FPMULS(matrix.m21,ptrvertex->y)+G3_FPMULS(matrix.m31,ptrvertex->z))+matrix.m41;
        ptrdvertex->y=(G3_FPMULS(matrix.m12,ptrvertex->x)+G3_FPMULS(matrix.m22,ptrvertex->y)+G3_FPMULS(matrix.m32,ptrvertex->z))+matrix.m42;
        ptrdvertex->z=(G3_FPMULS(matrix.m13,ptrvertex->x)+G3_FPMULS(matrix.m23,ptrvertex->y)+G3_FPMULS(matrix.m33,ptrvertex->z))+matrix.m43;
        {
          FP pz;                                                                                    
          pz=ptrdvertex->z;                                                                                  
          if(pz>=g3_nearplane)                                                                      
          {                                                                                       
            pz=G3_FPSHIFTR(pz,g3_zoomshift);                                                        
            ptrdvertex->px = g3_xmid+(SD)(ptrdvertex->x/pz);                                                          
            ptrdvertex->py = g3_ymid-(SD)(ptrdvertex->y/pz);                                                          
          }
        }
        ptrdvertex->light=-1;
        ptrdvertex->ptrskl=ptrskl;
        ptrdvertex->ptrvertex=ptrvertex;
        ptrvertex++;
        ptrdvertex++;
        }
      while(ptrvertex->flag>=0);
      ptrskl++;
      }
  }
  else
  {
    while(ptrskl->recurse!=G3_ENDMARK)             /* for each skl, compute vertex list */
      {
      /*--------------compute vertex from current sklitem-------*/
      memcpy(&matrix,&ptrskl->matrix,sizeof(G3_MATRIX));
      g3_matrix_xyzoom(&matrix);
      m11m21=G3_FPSHIFTR(matrix.m11*matrix.m21,G3_FPSHIFT);
      m12m22=G3_FPSHIFTR(matrix.m12*matrix.m22,G3_FPSHIFT);
      m13m23=G3_FPSHIFTR(matrix.m13*matrix.m23,G3_FPSHIFT);
      do
        {
        //ptrdvertex->x=(((matrix.m11*ptrvertex->x)+(matrix.m21*ptrvertex->y)+(matrix.m31*ptrvertex->z))>>G3_FPSHIFT)+matrix.m41;
        //ptrdvertex->y=(((matrix.m12*ptrvertex->x)+(matrix.m22*ptrvertex->y)+(matrix.m32*ptrvertex->z))>>G3_FPSHIFT)+matrix.m42;
        //ptrdvertex->z=(((matrix.m13*ptrvertex->x)+(matrix.m23*ptrvertex->y)+(matrix.m33*ptrvertex->z))>>G3_FPSHIFT)+matrix.m43;
#if 1
        ptrdvertex->x  =  G3_FPMULS(ptrvertex->x+matrix.m21,ptrvertex->y+matrix.m11)-
                          m11m21-
                          ptrvertex->xy+
                          G3_FPMULS(matrix.m31,ptrvertex->z)+
                          (matrix.m41);
        ptrdvertex->y  =  G3_FPMULS(ptrvertex->x+matrix.m22,ptrvertex->y+matrix.m12)-
                          m12m22-
                          ptrvertex->xy+
                          G3_FPMULS(matrix.m32,ptrvertex->z)+
                          (matrix.m42);
        ptrdvertex->z  =  G3_FPMULS(ptrvertex->x+matrix.m23,ptrvertex->y+matrix.m13)-
                          m13m23-
                          ptrvertex->xy+
                          G3_FPMULS(matrix.m33,ptrvertex->z)+
                          (matrix.m43);
#endif
        {
          FP pz;                                                                                    
          pz=ptrdvertex->z;                                                                                  
          if(pz>=g3_nearplane)                                                                      
          {                                                                                       
            pz=G3_FPSHIFTR(pz,g3_zoomshift);                                                        
            ptrdvertex->px = g3_xmid+(SD)(ptrdvertex->x/pz);                                                          
            ptrdvertex->py = g3_ymid-(SD)(ptrdvertex->y/pz);                                                          
          }
        }
        ptrdvertex->light=-1;
        ptrdvertex->ptrskl=ptrskl;
        ptrdvertex->ptrvertex=ptrvertex;
        ptrvertex++;
        ptrdvertex++;
        }
      while(ptrvertex->flag>=0);
      ptrskl++;
      }
  }

  // light offset related to altitude
//  if(ptrobject->translation.y>=G3_FPVALUE(G3_LANDWAVEHEIGHT-G3_LANDWAVEAMP))
//    lightoffset=G3_MAXLIGHTING;
//  else
    lightoffset=0;
  // then execute culling and store visible polygons
  ptrpoly=ptrobject->ptrpoly;
  while(ptrpoly->type!=G3_ENDMARK)
    {
    ptra=G3_OBJECT_DVERTEXADR(g3_objectdvertex,ptrpoly->pts1);     /* read a point       */
    ptrb=G3_OBJECT_DVERTEXADR(g3_objectdvertex,ptrpoly->pts2);     /* read b point       */
    ptrc=G3_OBJECT_DVERTEXADR(g3_objectdvertex,ptrpoly->pts3);     /* read c point       */

    dist=g3_pvmaxdistance(ptra,ptrb,ptrc);
    if(dist>0)
      {
      if(ptra->light==-1)
        {
        G3_PTRDVERTEX_LIGHT(ptra,lightoffset);
        }
      if(ptrb->light==-1)
        {
        G3_PTRDVERTEX_LIGHT(ptrb,lightoffset);
        }
      if(ptrc->light==-1)
        {
        G3_PTRDVERTEX_LIGHT(ptrc,lightoffset);
        }
      dist>>=G3_FPSHIFT;
      G3_OT_ADD(dist,ptrpoly);
      }
    ptrpoly++;
    }
  #if 0
  if(g3_viewskl)
    {
    ptrskl=ptrobject->ptrskl;
    while(ptrskl->recurse!=G3_ENDMARK)             /* for each skl, compute vertex list */
      {
      g3_line(ptrskl->ddtarget.px,ptrskl->ddtarget.py,ptrskl->dstarget.px,ptrskl->dstarget.py,0x7);
      g3_marker(ptrskl->ddtarget.px,ptrskl->ddtarget.py,7);
      ptrskl++;
      }
    }
  #endif
  g3_objectdvertex=ptrdvertex;
  }

void g3_object_motionprepare(G3_OBJECT *ptrobject)
  {
  G3_SKL *ptrskl;
  G3_SKLPOS *ptrsklpos;
  ptrskl=ptrobject->ptrskl;
  ptrsklpos=ptrobject->sklpos;
  while(ptrskl->recurse!=G3_ENDMARK)
    {
    ptrskl->drotation.x=g3_angle2angle(ptrskl->rotation.x,ptrsklpos->rotation.x);
    ptrskl->drotation.y=g3_angle2angle(ptrskl->rotation.y,ptrsklpos->rotation.y);
    ptrskl->drotation.z=g3_angle2angle(ptrskl->rotation.z,ptrsklpos->rotation.z);
    ptrskl->drotation.x/=ptrobject->framecount;
    ptrskl->drotation.y/=ptrobject->framecount;
    ptrskl->drotation.z/=ptrobject->framecount;
    ptrskl->dtranslation.x=ptrsklpos->translation.x-ptrskl->translation.x;
    ptrskl->dtranslation.y=ptrsklpos->translation.y-ptrskl->translation.y;
    ptrskl->dtranslation.z=ptrsklpos->translation.z-ptrskl->translation.z;
    ptrskl->dtranslation.x/=ptrobject->framecount;
    ptrskl->dtranslation.y/=ptrobject->framecount;
    ptrskl->dtranslation.z/=ptrobject->framecount;
    ptrsklpos++;
    ptrskl++;
    }
  }

void g3_object_motionprogress(G3_OBJECT *ptrobject)
  {
  G3_SKL *ptrskl;
  G3_SKLPOS *ptrsklpos;
  ptrskl=ptrobject->ptrskl;
  ptrsklpos=ptrobject->sklpos;
  while(ptrskl->recurse!=G3_ENDMARK)
    {
    ptrskl->rotation.x+=ptrskl->drotation.x;
    ptrskl->rotation.y+=ptrskl->drotation.y;
    ptrskl->rotation.z+=ptrskl->drotation.z;
    ptrskl->translation.x+=ptrskl->dtranslation.x;
    ptrskl->translation.y+=ptrskl->dtranslation.y;
    ptrskl->translation.z+=ptrskl->dtranslation.z;
    ptrsklpos++;
    ptrskl++;
    }
  ptrobject->framecount-=1;
  }

void g3_object_set(G3_OBJECT *ptrobject,SD rx,SD ry,SD rz,SD tx,SD ty,SD tz)
  {
  ptrobject->rotation.x=G3_ANGLE(rx);
  ptrobject->rotation.y=G3_ANGLE(ry);
  ptrobject->rotation.z=G3_ANGLE(rz);
  ptrobject->translation.x=tx;
  ptrobject->translation.y=ty;
  ptrobject->translation.z=tz;
  g3_matrix_rotzrotxrotytranslate(&ptrobject->rotation,&ptrobject->translation,&ptrobject->matrix);
  }


void g3_object_progress(G3_OBJECT *ptrobject,SD drx,SD dry,SD drz,SD dtx,SD dty,SD dtz)
  {
  G3_ROTATION rotation;
  G3_VECTOR translation;
  G3_MATRIX updatingmatrix,finalmatrix;

  /* compute progression of the object */
  ptrobject->matrix.m41+=G3_FPMULS(ptrobject->matrix.m11,dtx)+G3_FPMULS(ptrobject->matrix.m21,dty)+G3_FPMULS(ptrobject->matrix.m31,dtz);
  ptrobject->matrix.m42+=G3_FPMULS(ptrobject->matrix.m12,dtx)+G3_FPMULS(ptrobject->matrix.m22,dty)+G3_FPMULS(ptrobject->matrix.m32,dtz);
  ptrobject->matrix.m43+=G3_FPMULS(ptrobject->matrix.m13,dtx)+G3_FPMULS(ptrobject->matrix.m23,dty)+G3_FPMULS(ptrobject->matrix.m33,dtz);

  /* store translation to object translation */

  ptrobject->translation.x=g3_camera.matrix.m41;
  ptrobject->translation.y=g3_camera.matrix.m42;
  ptrobject->translation.z=g3_camera.matrix.m43;

  /* matrix is now a rotation matrix */

  ptrobject->matrix.m41=0;
  ptrobject->matrix.m42=0;
  ptrobject->matrix.m43=0;

  /* compute the updating matrix */

  rotation.x=G3_ANGLE(drx);
  rotation.y=G3_ANGLE(dry);
  rotation.z=G3_ANGLE(drz);
  translation.x=0;
  translation.y=0;
  translation.z=0;
  g3_matrix_rotzrotxrotytranslate(&rotation,&translation,&updatingmatrix);

  g3_matrix_mult(&updatingmatrix,&ptrobject->matrix,&finalmatrix);
  g3_matrix_copy(&finalmatrix,&ptrobject->matrix);

  /* adjust matrix */

  g3_matrix_getrotation(&ptrobject->matrix,&ptrobject->rotation);
  g3_matrix_rotzrotxrotytranslate(&ptrobject->rotation,&ptrobject->translation,&ptrobject->matrix);
  }


void g3_object_setplayer(G3_OBJECT *ptrobject)
{
  g3_objectplayer=ptrobject;
}

G3_OBJECT *g3_object_getplayer(void)
{
  return(g3_objectplayer);
}
  
G3_OBJECT *g3_object_add(char *name,G3_FUNCOBJECT function,G3_OBJECT *creator,G3_OBJECT *goal,G3_VECTOR *translation,G3_ROTATION *rotation,SD life,SD power)
  {
  SD i;  
  G3_OBJECT *ptrobject;
  UB *ptrobjectflag;
  ptrobject=g3_object;
  ptrobjectflag=g3_objectflag;
  i=G3_MAXOBJECT;
  do
  {
    if((*ptrobjectflag)==0) goto ok;
    ptrobjectflag++;
    ptrobject++;
  }
  while(--i);
  g3_error("g3_object_add : no more object space");
  return(NULL);
ok:
  *ptrobjectflag=G3_OBJECTFLAG_VALID|G3_OBJECTFLAG_CANRUN|G3_OBJECTFLAG_VISIBLE;
  memset(ptrobject,0,sizeof(G3_OBJECT));
  ptrobject->scale=G3_FPVALUE(1.0);
  ptrobject->canrun=G3_TRUE;
  ptrobject->function=function;
  if(creator==NULL)
    ptrobject->creator=ptrobject;
  else
    ptrobject->creator=creator;
  ptrobject->goal=goal;
  G3_TORE_MASKXYZ(
    translation->x,
    translation->y,
    translation->z,
    ptrobject->translation.x,
    ptrobject->translation.y,
    ptrobject->translation.z
    )
  G3_TORE_MASKXYZ(
    translation->x,
    translation->y,
    translation->z,
    ptrobject->oldtranslation.x,
    ptrobject->oldtranslation.y,
    ptrobject->oldtranslation.z
    )
  ptrobject->rotation.x=rotation->x;
  ptrobject->rotation.y=rotation->y;
  ptrobject->rotation.z=rotation->z;
  ptrobject->life=life;
  ptrobject->power=power;
  g3_object_load(name,ptrobject);
  g3_matrix_rotzrotxrotytranslate(&ptrobject->rotation,&ptrobject->translation,&ptrobject->matrix);
  // add object to grid
  g3_objectgrid_put(ptrobject,G3_OBJECTGRIDCONVERT(ptrobject->translation.x),G3_OBJECTGRIDCONVERT(ptrobject->translation.z));
  // call creation user function
  if(ptrobject->function)
  {
    (ptrobject->function)(ptrobject,G3_STM_CREATE);
  }
  // return pointer on created object
  g3_objectcount++;
  return(ptrobject);
  }

void g3_allobject_load(void)
  {
  }

G3_OBJECT *g3_object_searchcollid(G3_OBJECT *ptrobject,SD timemax,SD *timecollid,SD *axis)
{
        SD collidtimemin;
        G3_OBJECT *collidptrobject;
        SD axi;
        SD timecol;
        SD collidaxis;
        G3_OBJECTGRIDITEM *item;
        G3_OBJECTGRIDFINDER finder;
        SD xx,zz;
        G3_OBJECT *ptrobj;

          collidptrobject=NULL;
          collidaxis=0;
          //compute grid coord of object
          xx=G3_OBJECTGRIDCONVERT(ptrobject->translation.x);
          zz=G3_OBJECTGRIDCONVERT(ptrobject->translation.z);
          collidtimemin=0x7FFFFFFF;
          item=g3_objectgrid_findfirst(&finder,xx-1,zz-1,xx+1,zz+1);
          while(item!=NULL)
          {
            ptrobj=item->ptrdata;
            //be sure ptrobj!=ptrobject (HE CAN'T COLLID WITH HIMSELF !!!! I HOPE SO !!...)       
            if(ptrobj!=ptrobject)
            //an object can't collid his soon
            if(ptrobj->creator!=ptrobject)
            //two soon of the same (NON NULL)creator can't collid between
            if(ptrobj->creator!=ptrobject->creator)
            {
              //verify if collid occurs, if yes, what time ?
              G3_VECTOR trsnull;
              G3_VECTOR trsrel;
              trsnull.x=0;
              trsnull.y=0;
              trsnull.z=0;
              G3_TORE_RELATIVEXYZ(
                ptrobject->translation.x,
                ptrobject->translation.y,
                ptrobject->translation.z,
                ptrobj->translation.x,
                ptrobj->translation.y,
                ptrobj->translation.z,
                trsrel.x,
                trsrel.y,
                trsrel.z
                )              
              axi=g3_cube_timecollid
              (
                &trsnull,G3_FPVALUE(16),&ptrobject->wspeed,
                &trsrel,G3_FPVALUE(16),
                timemax,&timecol
              );
              if(axi)
              {
                if(timecol<collidtimemin)
                {
                  collidptrobject=ptrobj;
                  collidtimemin=timecol;
                  collidaxis=axi;
                }
              }
            }
            item=g3_objectgrid_findnext(&finder);
          }
  *timecollid=collidtimemin;
  *axis=collidaxis;      
  return(collidptrobject);
}

SD        g3_collidaxis;
G3_OBJECT *g3_collidptrobject;

SD g3_collid_axis(void)
{
  return(g3_collidaxis);
}

G3_OBJECT *g3_collid_ptrobject(void)
{
  return(g3_collidptrobject);
}

void g3_allobject_manage(void)
{
  SD i;
  SD timemax;
  G3_OBJECT *ptrobject;
  UB *ptrobjectflag;
  i=G3_MAXOBJECT;
  ptrobject=g3_object;
  ptrobjectflag=g3_objectflag;
  timemax=g3_timer_sampledelay();
  g3_objectactivecount=0;
  do
  {
      if(
        ((*ptrobjectflag)&G3_OBJECTFLAG_VALID )&&
        ((*ptrobjectflag)&G3_OBJECTFLAG_CANRUN)
        )
      {
        g3_objectactivecount++;
        //store old position
        ptrobject->oldrotation.x=ptrobject->rotation.x;
        ptrobject->oldrotation.y=ptrobject->rotation.y;
        ptrobject->oldrotation.z=ptrobject->rotation.z;
        // store old rotations
        ptrobject->oldtranslation.x=ptrobject->translation.x;
        ptrobject->oldtranslation.y=ptrobject->translation.y;
        ptrobject->oldtranslation.z=ptrobject->translation.z;
        //if there is a function to manage it then do it...
        if(ptrobject->function)
        {
          (ptrobject->function)(ptrobject,G3_STM_RUN);
        }
        G3_TORE_MASKXYZ(
          ptrobject->translation.x,
          ptrobject->translation.y,
          ptrobject->translation.z,
          ptrobject->translation.x,
          ptrobject->translation.y,
          ptrobject->translation.z
          )
        //if object life is over then unvalidate it
        if(ptrobject->life<=0)
        {
          if(ptrobject->function)
          {
            (ptrobject->function)(ptrobject,G3_STM_DESTROY);
          }
          g3_objectgrid_remove(
            ptrobject,
            G3_OBJECTGRIDCONVERT(ptrobject->oldtranslation.x),
            G3_OBJECTGRIDCONVERT(ptrobject->oldtranslation.z)
            );
          *ptrobjectflag=0;
          g3_objectcount--;
        }
        else
        //if object is living then manage its displacement
        {
          G3_OBJECT *objcollid;
          SD timecollid;
          SD axis;

          G3_TORE_RELATIVEXYZ(
            ptrobject->oldtranslation.x,
            ptrobject->oldtranslation.y,
            ptrobject->oldtranslation.z,
            ptrobject->translation.x,
            ptrobject->translation.y,
            ptrobject->translation.z,
            ptrobject->wspeed.x,
            ptrobject->wspeed.y,
            ptrobject->wspeed.z
            )
          //ptrobject->wspeed.x=ptrobject->translation.x-ptrobject->oldtranslation.x;
          //ptrobject->wspeed.y=ptrobject->translation.y-ptrobject->oldtranslation.y;
          //ptrobject->wspeed.z=ptrobject->translation.z-ptrobject->oldtranslation.z;
          if(ptrobject->wspeed.x!=0 || ptrobject->wspeed.y!=0 || ptrobject->wspeed.z!=0)
          {
            ptrobject->wspeed.x/=timemax;
            ptrobject->wspeed.y/=timemax;
            ptrobject->wspeed.z/=timemax;
            objcollid=g3_object_searchcollid(ptrobject,timemax,&timecollid,&axis);
            if(objcollid!=NULL)
            {
              ptrobject->translation.x=ptrobject->oldtranslation.x;
              ptrobject->translation.y=ptrobject->oldtranslation.y;
              ptrobject->translation.z=ptrobject->oldtranslation.z;
              g3_collidaxis=axis;
              if(ptrobject->function)
              {
                g3_collidptrobject=objcollid;
                (ptrobject->function)(ptrobject,G3_STM_COLLID);
              }
              if(objcollid->function)
              {
                g3_collidptrobject=ptrobject;
                (objcollid->function)(objcollid,G3_STM_COLLIDEDBY);
              }  
            }
            g3_objectgrid_move(
              ptrobject,
              G3_OBJECTGRIDCONVERT(ptrobject->oldtranslation.x),
              G3_OBJECTGRIDCONVERT(ptrobject->oldtranslation.z),
              G3_OBJECTGRIDCONVERT(ptrobject->translation.x),
              G3_OBJECTGRIDCONVERT(ptrobject->translation.z)
              );
          }
        }
        
      }
      ptrobjectflag++;
      ptrobject++;
    }
  while(--i);
}

void g3_allobject_compute(void)
  {
  G3_OBJECT *ptrobject;
  G3_OBJECTGRIDITEM *item;
  G3_OBJECTGRIDFINDER finder;
  SD xx,zz;
  //compute grid coord of visible object around camera
  xx=G3_OBJECTGRIDCONVERT(g3_camera.matrix.m41);
  zz=G3_OBJECTGRIDCONVERT(g3_camera.matrix.m43);
  item=g3_objectgrid_findfirst(&finder,xx-3,zz-3,xx+3,zz+3);
  while(item!=NULL)
  {
    ptrobject=item->ptrdata;
    if(ptrobject->function)
    {
      (ptrobject->function)(ptrobject,G3_STM_DISPLAY);
    }
    g3_object_prepare(ptrobject);
    g3_object_compute(ptrobject);
    item=g3_objectgrid_findnext(&finder);
  }
}

void g3_allobject_init(void)
{
  SD i;
  G3_OBJECT *ptrobject;
  UB *ptrobjectflag;
  // reset object and objectvalid tables
  g3_objectcount=0;
  i=G3_MAXOBJECT;
  ptrobject=g3_object;
  ptrobjectflag=g3_objectflag;
  do
  {
    memset(ptrobject,0,sizeof(G3_OBJECT));
    *ptrobjectflag=0;
    ptrobjectflag++;
    ptrobject++;
  }
  while(--i);
  // initialize object grid
  g3_objectgrid_begin();
}

void g3_allobject_restore(void)
{
  SD i;
  G3_OBJECT *ptrobject;
  UB *ptrobjectflag;
  i=G3_MAXOBJECT;
  ptrobject=g3_object;
  ptrobjectflag=g3_objectflag;
  do
  {
    if((*ptrobjectflag)&G3_OBJECTFLAG_VALID)
    {
      //if there is a function to manage it then do it...
      if(ptrobject->function)
      {
        (ptrobject->function)(ptrobject,G3_STM_DESTROY);
      }
      *ptrobjectflag=0;
      g3_objectgrid_remove
        (
        ptrobject,
        G3_OBJECTGRIDCONVERT(ptrobject->translation.x),
        G3_OBJECTGRIDCONVERT(ptrobject->translation.z)
        );
    }
    ptrobjectflag++;
    ptrobject++;
  }
  while(--i);
  // restore object grid
  g3_objectgrid_end();
}

int g3_object_getflag(G3_OBJECT *ptrobject,UB flag)
{
  SD index;
  index=((UD)(ptrobject)-(UD)(g3_object))/sizeof(G3_OBJECT);
  return((g3_objectflag[index]|flag)!=0);
}

void g3_object_setflag(G3_OBJECT *ptrobject,UB flag,int value)
{
  SD index;
  index=((UD)(ptrobject)-(UD)(g3_object))/sizeof(G3_OBJECT);
  if(value)
    g3_objectflag[index]=g3_objectflag[index]|flag;
  else
    g3_objectflag[index]=g3_objectflag[index]&(0xFF-flag);
}

