//**********************************************************************
// OBJECTGRID MANAGEMENT
//**********************************************************************

#include <string.h>

#include "g3def.h"
#include "g3land.h"
#include "g3error.h"
#include "g3objgrd.h"

typedef G3_OBJECTGRIDITEM *G3_OBJECTPTRGRIDITEM;
G3_OBJECTPTRGRIDITEM    g3_objectgrid[G3_OBJECTGRIDSIZE][G3_OBJECTGRIDSIZE];// pointer on grid entry
G3_OBJECTGRIDITEM       g3_objectgriditem[G3_MAXOBJECT];          // pointer on grid item array
G3_OBJECTPTRGRIDITEM    g3_objectgriditemend;       // pointer on grid item array end
G3_OBJECTPTRGRIDITEM    g3_objectgriditemfree;      // pointer on current free griditem

//---------------------------------
// find the next free grid item
//---------------------------------
G3_OBJECTGRIDITEM *_g3_objectgrid_findfree(void)
{
  G3_OBJECTGRIDITEM *ptrgriditem;
  ptrgriditem=g3_objectgriditemfree;
  // search for a free grid item
  while(ptrgriditem<g3_objectgriditemend)
  {
    if(ptrgriditem->ptrdata==NULL) goto found;
    ptrgriditem++;
  }
  // restart at beginning of item array
  ptrgriditem=g3_objectgriditem;
  while(ptrgriditem<g3_objectgriditemend)
  {
    if(ptrgriditem->ptrdata==NULL) goto found;
    ptrgriditem++;
  }
  // here there is an error cause there are no grid item free
  return(NULL);
found:
  g3_objectgriditemfree=ptrgriditem+1;
  return(ptrgriditem);
}

//-----------------------------
// init the grid system
//-----------------------------
void g3_objectgrid_begin(void)
{
  g3_objectgriditemfree=g3_objectgriditem;
  g3_objectgriditemend=g3_objectgriditem+G3_MAXOBJECT;
  // reset entry on grid items
  memset(g3_objectgrid,0,sizeof(G3_OBJECTPTRGRIDITEM)*G3_OBJECTGRIDSIZE*G3_OBJECTGRIDSIZE);
  // mark all griditem as free
  memset(g3_objectgriditem,0,sizeof(G3_OBJECTGRIDITEM)*G3_OBJECTGRIDMAXITEM);
  // pointer on grid item at beginning of array
}

//---------------------------------
// release the grid system
//---------------------------------
void g3_objectgrid_end(void)
{
  int x,y;
  G3_OBJECTGRIDITEM *ptrgriditem;
  // control the integrity of grid
  for(y=0;y<G3_OBJECTGRIDSIZE;y++)
  {
    for(x=0;x<G3_OBJECTGRIDSIZE;x++)
    {
      if(g3_objectgrid[x][y]!=NULL)
      {
        g3_error("G3_OBJECTGRID_END : grid[][] not NULL");
      }
    }
  }
  // control the integrity of griditem
  ptrgriditem=g3_objectgriditem;
  while(ptrgriditem<g3_objectgriditemend)
  {
    if(ptrgriditem->ptrdata!=NULL)
    {
      g3_error("G3_OBJECTGRID_END : griditem not NULL");
    }
    ptrgriditem++;
  }
}

//-----------------------------
// put an item in a grid
//-----------------------------
void g3_objectgrid_put(void *ptrdata,int x,int y)
{
  G3_OBJECTPTRGRIDITEM ptrgriditem;
  G3_OBJECTPTRGRIDITEM *ptrptrgrid;
  x&=(G3_OBJECTGRIDSIZE-1);
  y&=(G3_OBJECTGRIDSIZE-1);
  ptrptrgrid=&g3_objectgrid[x][y];
  // search a free grid item
  ptrgriditem=_g3_objectgrid_findfree();
  // the new grid item point a data
  ptrgriditem->ptrdata=ptrdata;
  // the new griditem point on the first from entry
  ptrgriditem->ptrnextgriditem=*ptrptrgrid;
  // the entry points on the new griditem
  *ptrptrgrid=ptrgriditem;
}

//-----------------------------
// remove an item from the grid
//-----------------------------
void g3_objectgrid_remove(void *ptrdata,int x,int y)
{
  G3_OBJECTPTRGRIDITEM ptrgrid;
  G3_OBJECTPTRGRIDITEM *ptrptrgrid;
  x&=(G3_OBJECTGRIDSIZE-1);
  y&=(G3_OBJECTGRIDSIZE-1);
  ptrptrgrid=&g3_objectgrid[x][y];
  ptrgrid=*ptrptrgrid;
  while(ptrgrid!=NULL)
  {
    if(ptrgrid->ptrdata==ptrdata)
    {
      // jump over removed link
      *ptrptrgrid=ptrgrid->ptrnextgriditem;
      // the item removed is now free
      ptrgrid->ptrdata=NULL;
      g3_objectgriditemfree=ptrgrid;     //best for next g3_grid_put !!!!
      return;
    }
    ptrptrgrid=&ptrgrid->ptrnextgriditem;
    ptrgrid=ptrgrid->ptrnextgriditem;
  }
  g3_error("G3_OBJECTGRID_REMOVE : data not found for removing");
}

//---------------------------------
// move an item in the grid
//---------------------------------
void g3_objectgrid_move(void *ptrdata,int x1,int y1,int x2,int y2)
{
  if((x1!=x2) || (y1!=y2))
  {
    g3_objectgrid_remove(ptrdata,x1,y1);
    g3_objectgrid_put(ptrdata,x2,y2);
  }
}

//--------------------------------------------
// find next item according to grid finder
//--------------------------------------------
G3_OBJECTGRIDITEM *g3_objectgrid_findnext(G3_OBJECTGRIDFINDER *finder)
{
  // goto next grid item in current list
  finder->ptrgriditem=finder->ptrgriditem->ptrnextgriditem;
  if(finder->ptrgriditem!=NULL) return(finder->ptrgriditem);
  // find the next grid entry if no more griditem in current chain
  while(finder->ptrgriditem==NULL)
  {
    //increment x
    finder->x++;
    if(finder->x>finder->x2)
    {
      finder->x=finder->x1;     // restore x counter
      finder->y++;          // goto next y
      if(finder->y>finder->y2)    // the final y exceed ?
      {
        return(NULL);       // this is the real end !!!!
      }
    }
    finder->ptrgriditem=g3_objectgrid[finder->x & (G3_OBJECTGRIDSIZE-1)][finder->y & (G3_OBJECTGRIDSIZE-1)];
  }
  // restore all like if the chain was the same
  return(finder->ptrgriditem);
}

//--------------------------------------------
// initialise grid finder
//--------------------------------------------

G3_OBJECTGRIDITEM *g3_objectgrid_findfirst(G3_OBJECTGRIDFINDER *finder,int x1,int y1,int x2,int y2)
{
  finder->ptrgriditem=g3_objectgrid[x1 & (G3_OBJECTGRIDSIZE-1)][y1 & (G3_OBJECTGRIDSIZE-1)];
  finder->x1=x1;        // remind the start pos
  finder->y1=y1;
  finder->x2=x2;        // remind the current pos
  finder->y2=y2;
  finder->x=x1;       // remind the final pos
  finder->y=y1;
  // find the next grid entry if no more griditem in current chain
  while(finder->ptrgriditem==NULL)
  {
    //increment x
    finder->x++;
    if(finder->x>finder->x2)
    {
      finder->x=finder->x1;     // restore x counter
      finder->y++;          // goto next y
      if(finder->y>finder->y2)    // the final y exceed ?
      {
        return(NULL);       // this is the real end !!!!
      }
    }
    finder->ptrgriditem=g3_objectgrid[finder->x & (G3_OBJECTGRIDSIZE-1)][finder->y & (G3_OBJECTGRIDSIZE-1)];
  }
  // restore all like if the chain was the same
  return(finder->ptrgriditem);
}

//--------------------------------------------
// x of item in grid finder
//--------------------------------------------
int g3_objectgrid_findx(G3_OBJECTGRIDFINDER *finder)
{
  return(finder->x);
}

//--------------------------------------------
// y of item in grid finder
//--------------------------------------------
int g3_objectgrid_findy(G3_OBJECTGRIDFINDER *finder)
{
  return(finder->x);
}

