
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#include "g3def.h"
#include "g3lib.h"
#include "g3graph.h"
#include "g3part.h"
#include "g3error.h"


  typedef struct _g3_particle
  {
  SD              type;  
  SD              life;         /* life                      */
  G3_VECTOR       translation;  /* position of object        */
  G3_ROTATION     rotation;     /* rotation of object        */
  G3_VECTOR       vtranslation;
  G3_ROTATION     vrotation;
  G3_VERTEX       vertex[4];
  G3_POLYGON      poly[2];      /* polygon                   */
  }
  G3_PARTICLE;

  /*---------- particle storage ------------*/
  SD            g3_particlecount;
  G3_PARTICLE   g3_particle[G3_MAXPARTICLE];
  G3_PARTICLE   *g3_ptrparticle;

/**************************************************************************/
/* particle management                                                    */
/**************************************************************************/


void g3_particle_init(void)
  {
  SD i;
  // clear life of each particle
  for(i=0;i<G3_MAXPARTICLE;i++)
  {
    g3_particle[i].life=0;
  }
  // set pointer on the first particle
  g3_ptrparticle=g3_particle;
  }

void g3_particle_restore(void)
  {
  }

G3_PARTICLE *_g3_particle_findfree(void)
  {
  G3_PARTICLE *ptrparticle;
    if(g3_particlecount==G3_MAXPARTICLE) return(NULL);
    ptrparticle=g3_ptrparticle;
    do
    {
      if(ptrparticle->life<=0) goto found;
      ptrparticle++;
    }
    while(ptrparticle<&g3_particle[G3_MAXPARTICLE]);
    ptrparticle=g3_particle;
    do
    {
      if(ptrparticle->life<=0) goto found;
      ptrparticle++;
    }
    while(ptrparticle<&g3_particle[G3_MAXPARTICLE]);
  return(NULL);
  found:
    g3_ptrparticle=ptrparticle;
    return(ptrparticle);
  }

SD g3_particle_getcount(void)
  {
    return(g3_particlecount);
  }

void g3_particle_addtriangle(SD x,SD y,SD z,SD color,SD life)
  {
    G3_PARTICLE *ptrparticle;
    ptrparticle=_g3_particle_findfree();
    if(!ptrparticle) return;

    

    ptrparticle->poly[0].type=G3_GRAPHCALL_POLYF;
    ptrparticle->poly[0].color=color;

    ptrparticle->poly[0].pts1=(sizeof(G3_DVERTEX))*0;
    ptrparticle->poly[0].pts2=(sizeof(G3_DVERTEX))*1;
    ptrparticle->poly[0].pts3=(sizeof(G3_DVERTEX))*2;

    ptrparticle->poly[1].type=G3_GRAPHCALL_POLYF;
    ptrparticle->poly[1].color=color;
    ptrparticle->poly[1].pts1=(sizeof(G3_DVERTEX))*2;
    ptrparticle->poly[1].pts2=(sizeof(G3_DVERTEX))*1;
    ptrparticle->poly[1].pts3=(sizeof(G3_DVERTEX))*0;

    ptrparticle->translation.x=x;
    ptrparticle->translation.y=y;
    ptrparticle->translation.z=z;

    ptrparticle->rotation.x=0;
    ptrparticle->rotation.y=0;
    ptrparticle->rotation.z=0;

    ptrparticle->vrotation.x=rand();
    ptrparticle->vrotation.y=rand();
    ptrparticle->vrotation.z=rand();

    ptrparticle->vtranslation.x=-(rand() & (0x1FFF>>2)) + (0xFFF>>2);
    ptrparticle->vtranslation.y=(rand() & (0xFFF>>2)) ;
    ptrparticle->vtranslation.z=-(rand() & (0x1FFF>>2)) + (0xFFF>>2);

    ptrparticle->life=life;
    ptrparticle->type=G3_PARTICLE_TRIANGLE;
  }


void g3_particle_addbulle(SD x,SD y,SD z,SD life)
{
    G3_PARTICLE *ptrparticle;
    ptrparticle=_g3_particle_findfree();
    if(!ptrparticle) return;

    ptrparticle->poly[0].type=G3_GRAPHCALL_POLYTT;
    ptrparticle->poly[0].color=0x0F0F0F0F;
    ptrparticle->poly[0].pts1=(sizeof(G3_DVERTEX))*0;
    ptrparticle->poly[0].pts2=(sizeof(G3_DVERTEX))*1;
    ptrparticle->poly[0].pts3=(sizeof(G3_DVERTEX))*2;
    ptrparticle->poly[0].xa=G3_GFPVALUE(0);
    ptrparticle->poly[0].ya=G3_GFPVALUE(0);
    ptrparticle->poly[0].xb=G3_GFPVALUE(31);
    ptrparticle->poly[0].yb=G3_GFPVALUE(0);
    ptrparticle->poly[0].xc=G3_GFPVALUE(15);
    ptrparticle->poly[0].yc=G3_GFPVALUE(31);
    ptrparticle->poly[0].bank=8;

    ptrparticle->translation.x=x;
    ptrparticle->translation.y=y;
    ptrparticle->translation.z=z;

    ptrparticle->translation.x+=(rand() & (G3_FPVALUE(8)-1))-G3_FPVALUE(4);
    ptrparticle->translation.z+=(rand() & (G3_FPVALUE(8)-1))-G3_FPVALUE(4);

    ptrparticle->vtranslation.y=(rand() & (0xFFF>>5))+G3_FPVALUE(0.08);

    ptrparticle->life=life;
    ptrparticle->type=G3_PARTICLE_BULLE;
}


void g3_particle_manage(void)
  {
  SD time;
  G3_PARTICLE *ptrparticle;

    time=g3_timer_sampledelay();
    ptrparticle=&g3_particle[G3_MAXPARTICLE-1];
    do
    {
      if(ptrparticle->life>0)
      {
        switch(ptrparticle->type)
        {
          case G3_PARTICLE_TRIANGLE:
          {
            // compute displacement of particle
            ptrparticle->translation.x+=ptrparticle->vtranslation.x*time;
            ptrparticle->translation.y+=ptrparticle->vtranslation.y*time;
            ptrparticle->translation.z+=ptrparticle->vtranslation.z*time;
            ptrparticle->rotation.x+=ptrparticle->vrotation.x*time;
            ptrparticle->rotation.y+=ptrparticle->vrotation.y*time;
            ptrparticle->rotation.z+=ptrparticle->vrotation.z*time;
          }
          break;
          case G3_PARTICLE_BULLE:
          {
            ptrparticle->translation.y+=ptrparticle->vtranslation.y*time;
          }
          break;
        }
        // particle life decrease
        ptrparticle->life-=time;
      }
      ptrparticle--;
    }
    while(ptrparticle!=g3_particle);
  }

void g3_particle_compute(void)
  {
  G3_MATRIX matrix;
  G3_PARTICLE *ptrparticle;
  G3_DVERTEX *ptrdvertex;
  SD m41,m42,m43;
  G3_MATRIX zoominvmatrix;

    memcpy(&zoominvmatrix,&g3_camera.invmatrix,sizeof(G3_MATRIX));
    g3_matrix_xyzoom(&zoominvmatrix);

    ptrparticle=&g3_particle[G3_MAXPARTICLE-1];
    do
    {
      if(ptrparticle->life>0)
      {
        switch(ptrparticle->type)
        {
          case G3_PARTICLE_TRIANGLE:
          //---------------------------------------------------------------------------------
          // TRIANGLE
          //---------------------------------------------------------------------------------
          #define G3_PARTICLE_XASHIFT   0
          #define G3_PARTICLE_YASHIFT   0
          #define G3_PARTICLE_XBSHIFT   4
          #define G3_PARTICLE_YBSHIFT   0
          #define G3_PARTICLE_XCSHIFT   3
          #define G3_PARTICLE_YCSHIFT   4
          {
            // compute main matrix of particle
            g3_matrix_rotzrotxrotytranslate(&ptrparticle->rotation,&ptrparticle->translation,&matrix);
            m41=      G3_FPMULS(matrix.m41,g3_camera.invmatrix.m11)
                      +G3_FPMULS(matrix.m42,g3_camera.invmatrix.m21)
                      +G3_FPMULS(matrix.m43,g3_camera.invmatrix.m31)
                      +(g3_camera.invmatrix.m41);
            m42=      G3_FPMULS(matrix.m41,g3_camera.invmatrix.m12)
                      +G3_FPMULS(matrix.m42,g3_camera.invmatrix.m22)
                      +G3_FPMULS(matrix.m43,g3_camera.invmatrix.m32)
                      +(g3_camera.invmatrix.m42);
            m43=      G3_FPMULS(matrix.m41,g3_camera.invmatrix.m13)
                      +G3_FPMULS(matrix.m42,g3_camera.invmatrix.m23)
                     +G3_FPMULS(matrix.m43,g3_camera.invmatrix.m33)
                     +(g3_camera.invmatrix.m43);
            matrix.m41=m41;
            matrix.m42=m42;
            matrix.m43=m43;
            g3_matrix_xyzoom(&matrix);
            // compute vertex of particle
            ptrdvertex=g3_objectdvertex;
            // compute vertex A
            ptrdvertex[0].x  =
                  (matrix.m11<<G3_PARTICLE_XASHIFT)
                + (matrix.m21<<G3_PARTICLE_YASHIFT)
                + (matrix.m41);
            ptrdvertex[0].y  =
                  (matrix.m12<<G3_PARTICLE_XASHIFT)
                + (matrix.m22<<G3_PARTICLE_YASHIFT)
                + (matrix.m42);
            ptrdvertex[0].z  =
                  (matrix.m13<<G3_PARTICLE_XASHIFT)
                + (matrix.m23<<G3_PARTICLE_YASHIFT)
                + (matrix.m43);
            G3_DVERTEX_PROJECT(ptrdvertex[0]);
            // compute vertex B
            ptrdvertex[1].x  =
                  (matrix.m11<<G3_PARTICLE_XBSHIFT)
                + (matrix.m21<<G3_PARTICLE_YBSHIFT)
                + (matrix.m41);
            ptrdvertex[1].y  =
                  (matrix.m12<<G3_PARTICLE_XBSHIFT)
                + (matrix.m22<<G3_PARTICLE_YBSHIFT)
                + (matrix.m42);
            ptrdvertex[1].z  =
                  (matrix.m13<<G3_PARTICLE_XBSHIFT)
                + (matrix.m23<<G3_PARTICLE_YBSHIFT)
                + (matrix.m43);
            G3_DVERTEX_PROJECT(ptrdvertex[1]);
            // compute vertex C
            ptrdvertex[2].x  =
                  (matrix.m11<<G3_PARTICLE_XCSHIFT)
                + (matrix.m21<<G3_PARTICLE_YCSHIFT)
                + (matrix.m41);
            ptrdvertex[2].y  =
                  (matrix.m12<<G3_PARTICLE_XCSHIFT)
                + (matrix.m22<<G3_PARTICLE_YCSHIFT)
                + (matrix.m42);
            ptrdvertex[2].z  =
                  (matrix.m13<<G3_PARTICLE_XCSHIFT)
                + (matrix.m23<<G3_PARTICLE_YCSHIFT)
                + (matrix.m43);
            G3_DVERTEX_PROJECT(ptrdvertex[2]);
            // now store polygons of particle in OT
            {
              SD dist;
              dist=g3_pvmindistance(&ptrdvertex[0],&ptrdvertex[1],&ptrdvertex[2]);
              if(dist>0)
              {
                dist>>=G3_FPSHIFT;
                G3_OT_ADD(dist,&ptrparticle->poly[0]);
              }
              else
              {
                dist=g3_pvmindistance(&ptrdvertex[2],&ptrdvertex[1],&ptrdvertex[0]);
                if(dist>0)
                {
                  dist>>=G3_FPSHIFT;
                  G3_OT_ADD(dist,&ptrparticle->poly[1]);
                }
              }
            }
            // close object vertex group
            g3_objectdvertex=ptrdvertex+3;
          }
          #undef G3_PARTICLE_XASHIFT
          #undef G3_PARTICLE_YASHIFT
          #undef G3_PARTICLE_XBSHIFT
          #undef G3_PARTICLE_YBSHIFT
          #undef G3_PARTICLE_XCSHIFT
          #undef G3_PARTICLE_YCSHIFT
          break;
          //---------------------------------------------------------------------------------
          // BULLE
          //---------------------------------------------------------------------------------
          case G3_PARTICLE_BULLE:
          {
            FP lx,ly,lz,rx,ry,rz,w;
            G3_TORE_RELATIVEXYZ(
              g3_camera.matrix.m41,
              g3_camera.matrix.m42,
              g3_camera.matrix.m43,
              ptrparticle->translation.x,
              ptrparticle->translation.y,
              ptrparticle->translation.z,
              lx,
              ly,
              lz
              );
            G3_FPXYZ_ROTATE(lx,ly,lz,g3_camera.invmatrix,rx,ry,rz);
            if(rz>g3_nearplane)
            {
              SD px,py,pz;
              pz=G3_FPSHIFTR(rz,g3_zoomshift);                                                        
              px = g3_xmid+(SD)(rx/pz);
              py = g3_ymid-(SD)(ry/pz);
              // compute vertex of particle
              ptrdvertex=g3_objectdvertex;
              ptrdvertex[0].z=rz;
              ptrdvertex[0].px=px;
              ptrdvertex[0].py=py;
              ptrdvertex[1].z=rz;
              ptrdvertex[1].px=px+31;
              ptrdvertex[1].py=py;
              ptrdvertex[2].z=rz;
              ptrdvertex[2].px=px+15;
              ptrdvertex[2].py=py+31;
              // add polygons in ot
              rz>>=G3_FPSHIFT;
              G3_OT_ADD(rz,&ptrparticle->poly[0]);
              // close object vertex group
              g3_objectdvertex=ptrdvertex+3;
            }
          }
          break;
        }
      }
      ptrparticle--;
    }
    while(ptrparticle!=g3_particle);
  }

