//**********************************************************************
//          PLANE SYSTEM  
//**********************************************************************
// File                 : g3plane.h
// Author               : C.Levastre
// Date                 : 1996
//**********************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>

#include "g3def.h"
#include "g3lib.h"
#include "g3plane.h"

void g3_plane_normal(G3_VECTOR *v1,G3_VECTOR *v2,G3_VECTOR *normal)
{
        FP nx,ny,nz,lenght;
        nx=G3_FPMULS(v1->y,v2->z)-G3_FPMULS(v2->y,v1->z);//20:10
        ny=G3_FPMULS(v2->x,v1->z)-G3_FPMULS(v1->x,v2->z);
        nz=G3_FPMULS(v1->x,v2->y)-G3_FPMULS(v2->x,v1->y);
        lenght=G3_FPMULSPREC(nx,nx,30)+G3_FPMULSPREC(ny,ny,30)+G3_FPMULSPREC(nz,nz,30);
        lenght=G3_FPSQRT(lenght);
        normal->x=G3_FPSHIFTR(nx/lenght,5);
        normal->y=G3_FPSHIFTR(ny/lenght,5);
        normal->z=G3_FPSHIFTR(nz/lenght,5);
}

void g3_plane_init(G3_VECTOR *va,G3_VECTOR *vb,G3_VECTOR *vc,G3_PLANE *plane)
{
        G3_VECTOR v1,v2;
        plane->min.x=G3_MIN3(va->x,vb->x,vc->x);
        plane->min.y=G3_MIN3(va->y,vb->y,vc->y);
        plane->min.z=G3_MIN3(va->z,vb->z,vc->z);
        plane->max.x=G3_MAX3(va->x,vb->x,vc->x);
        plane->max.y=G3_MAX3(va->y,vb->y,vc->y);
        plane->max.z=G3_MAX3(va->z,vb->z,vc->z);
        //------------------------------------
        // plane equation
        //------------------------------------
        v1.x=vb->x-va->x;
        v1.y=vb->y-va->y;
        v1.z=vb->z-va->z;
        v2.x=vc->x-va->x;
        v2.y=vc->y-va->y;
        v2.z=vc->z-va->z;
        // compute normal of face
        g3_plane_normal(&v1,&v2,&plane->normal);
        // compute coef of plane
        plane->d=-G3_FPSHIFTR(((plane->normal.x*va->x)+(plane->normal.y*va->y)+(plane->normal.z*va->z)),G3_FPSHIFT);
        //------------------------------------
        // shift normal on left
        //------------------------------------
        plane->normal.x=G3_FPSHIFTL(plane->normal.x,G3_FPSHIFT);
        plane->normal.y=G3_FPSHIFTL(plane->normal.y,G3_FPSHIFT);
        plane->normal.z=G3_FPSHIFTL(plane->normal.z,G3_FPSHIFT);
        //------------------------------------
        // plane A,B,NORMAL equation
        //------------------------------------
        v1.x=vb->x-va->x;
        v1.y=vb->y-va->y;
        v1.z=vb->z-va->z;
        v2.x=plane->normal.x;
        v2.y=plane->normal.y;
        v2.z=plane->normal.z;
        // compute normal of face
        g3_plane_normal(&v1,&v2,&plane->nab);
        // compute coef of plane
        plane->dab=-(
                        (plane->nab.x*va->x)+
                        (plane->nab.y*va->y)+
                        (plane->nab.z*va->z)
                        );
        //------------------------------------
        // plane B,C,NORMAL equation
        //------------------------------------
        v1.x=vc->x-vb->x;
        v1.y=vc->y-vb->y;
        v1.z=vc->z-vb->z;
        v2.x=plane->normal.x;
        v2.y=plane->normal.y;
        v2.z=plane->normal.z;
        // compute normal of face
        g3_plane_normal(&v1,&v2,&plane->nbc);
        // compute coef of plane
        plane->dbc=-(
                        (plane->nbc.x*vb->x)+
                        (plane->nbc.y*vb->y)+
                        (plane->nbc.z*vb->z)
                        );
        //------------------------------------
        // plane C,A,NORMAL equation
        //------------------------------------
        v1.x=va->x-vc->x;
        v1.y=va->y-vc->y;
        v1.z=va->z-vc->z;
        v2.x=plane->normal.x;
        v2.y=plane->normal.y;
        v2.z=plane->normal.z;
        // compute normal of face
        g3_plane_normal(&v1,&v2,&plane->nca);
        // compute coef of plane
        plane->dca=-(
                        (plane->nca.x*vc->x)+
                        (plane->nca.y*vc->y)+
                        (plane->nca.z*vc->z)
                        );
        //------------------------------------
        //shift normal on right
        //------------------------------------
        plane->normal.x=G3_FPSHIFTR(plane->normal.x,G3_FPSHIFT);
        plane->normal.y=G3_FPSHIFTR(plane->normal.y,G3_FPSHIFT);
        plane->normal.z=G3_FPSHIFTR(plane->normal.z,G3_FPSHIFT);
}



G3_PLANE *g3_plane_intersect(G3_PLANE *plane,int nbplane,G3_VECTOR *p1,G3_VECTOR *p2,G3_VECTOR *res,FP *ratio)
{
        G3_PLANE *ptrplane;
        G3_VECTOR min,max;
        G3_VECTOR inter;
        FP s1,s2,t,mint;
        int i;
        // if no plane to test then stop and return no collision
        if(nbplane<=0) return(NULL);
        i=nbplane;
        ptrplane=NULL;
        // initialize bounding box of ray p1->p2
        min.x=G3_MIN(p1->x,p2->x);
        min.y=G3_MIN(p1->y,p2->y);
        min.z=G3_MIN(p1->z,p2->z);
        max.x=G3_MAX(p1->x,p2->x);
        max.y=G3_MAX(p1->y,p2->y);
        max.z=G3_MAX(p1->z,p2->z);
        // the min ratio is very high
        mint=G3_FPVALUE(32768);
        // for each plane in the list.... test intersection
        do
        {
                // test for box collision
                if(G3_MIN(max.x,plane->max.x)<G3_MAX(min.x,plane->min.x)) goto nextplane;
                if(G3_MIN(max.y,plane->max.y)<G3_MAX(min.y,plane->min.y)) goto nextplane;
                if(G3_MIN(max.z,plane->max.z)<G3_MAX(min.z,plane->min.z)) goto nextplane;
                // collision can exist only from front to back of plane....
                // give sign of p1 according to plane 
                s1 = G3_FPMULS(p1->x,plane->normal.x)+
                         G3_FPMULS(p1->y,plane->normal.y)+
                         G3_FPMULS(p1->z,plane->normal.z)+
                         plane->d;
                // if point 1 not in front then next plane
                if(s1<0) goto nextplane;
                // give sign of p2 according to plane
                s2 = G3_FPMULS(p2->x,plane->normal.x)+
                         G3_FPMULS(p2->y,plane->normal.y)+
                         G3_FPMULS(p2->z,plane->normal.z)+
                         plane->d;
                // if point 2 not in back then next plane
                if(s2>0) goto nextplane;
                // compute ratio of collid dist relative to segment P1-P2
                t = G3_FPDIVS
                        (
                                -(
                                G3_FPMULS(plane->normal.x,p1->x)+
                                G3_FPMULS(plane->normal.y,p1->y)+
                                G3_FPMULS(plane->normal.z,p1->z)+plane->d
                                )
                                ,
                                s2-s1
                        );
                // if it's not the nearest intersection then goto next plane
                if(t>mint) goto nextplane;
                // compute intersection with plane
                inter.x=p1->x+G3_FPSHIFTR(t*(p2->x-p1->x),G3_FPSHIFT);
                inter.y=p1->y+G3_FPSHIFTR(t*(p2->y-p1->y),G3_FPSHIFT);
                inter.z=p1->z+G3_FPSHIFTR(t*(p2->z-p1->z),G3_FPSHIFT);
                goto collid;
                // sign of res according to ab
                s1 = G3_FPMULS(inter.x,plane->nab.x)+
                         G3_FPMULS(inter.y,plane->nab.y)+
                         G3_FPMULS(inter.z,plane->nab.z)+
                         plane->dab;
                if(s1>0) goto nextplane;
                // sign of res according to bc
                s1 = G3_FPMULS(inter.x,plane->nbc.x)+
                         G3_FPMULS(inter.y,plane->nbc.y)+
                         G3_FPMULS(inter.z,plane->nbc.z)+
                         plane->dbc;
                if(s1>0) goto nextplane;
                // sign of res according to ca
                s1 = G3_FPMULS(inter.x,plane->nca.x)+
                         G3_FPMULS(inter.y,plane->nca.y)+
                         G3_FPMULS(inter.z,plane->nca.z)+
                         plane->dca;
                if(s1>0) goto nextplane;
                // store the coordinate of intersection, the current ratio, the addr of plane
        collid:
                res->x=inter.x;
                res->y=inter.y;
                res->z=inter.z;
                *ratio=t;
                mint=t;
                ptrplane=plane;
        nextplane:
                plane++;
        }
        while(--i);
        return(ptrplane);
}

                
                
                
                
