
#include <windows.h>
#include <windowsx.h>
#include "dsound.h"

#include "g3file.h"
#include "g3wav.h"
#include "g3snd.h"

#define G3_SND_FREQUENCY 22050
#define G3_SNDFILEMAX 128
#define G3_SNDMAX 16

LPDIRECTSOUND g3_ds;
HWND g3_dshwnd=NULL;

int g3_sndfilecount=0;


typedef struct
{
  char name[_MAX_FNAME];
  unsigned char *buffer;
  int buffersize;
  LPDIRECTSOUNDBUFFER dsbuffer;
}
G3_SNDFILE;

G3_SNDFILE g3_sndfile[G3_SNDFILEMAX];    //array of loaded sounds

LPDIRECTSOUNDBUFFER g3_snd[G3_SNDMAX];


//---------------------------------------------------------------------------
// G3_SND_VALID returns TRUE if DS system is valid
//---------------------------------------------------------------------------
int g3_snd_valid(HWND hwnd)
{
  return(g3_ds!=NULL);
}

//---------------------------------------------------------------------------
// G3_SND_RELEASE release DS system 
//---------------------------------------------------------------------------
void g3_snd_release(void)
{
  int i;
  for(i=0;i<g3_sndfilecount;i++)
  {
    IDirectSoundBuffer_Release(g3_sndfile[i].dsbuffer);
  }
  if(g3_ds!=NULL) IDirectSound_Release(g3_ds);
  g3_dshwnd=NULL;
  g3_ds=NULL;
  g3_sndfilecount=0;
}


//---------------------------------------------------------------------------
// G3_SND_VALID inits DS system 
//---------------------------------------------------------------------------
int g3_snd_init(HWND hwnd)
{
  g3_dshwnd=hwnd;
  g3_sndfilecount=0;
  if(DirectSoundCreate(NULL,&g3_ds,NULL)==DS_OK)
  {
    if(IDirectSound_SetCooperativeLevel(g3_ds,g3_dshwnd,DSSCL_NORMAL)==DS_OK)
    {
      return(TRUE);
    }
  }
  g3_snd_release();
  return(FALSE);
}

int g3_snd_add(char *path)
{
  char filename[_MAX_PATH];
  PCMWAVEFORMAT pcmwf;
  DSBUFFERDESC dsbdesc;
  HRESULT hr;
  LPVOID lpvPtr1;
  DWORD dwBytes1; 
  LPVOID lpvPtr2;
  DWORD dwBytes2;

  g3_fextput(path,".WAV",filename);
  if(g3_sndfilecount>=G3_SNDFILEMAX)
  {
    return(-1);
  }
  g3_sndfile[g3_sndfilecount].buffer=(unsigned char *)g3_wav_load(filename,&g3_sndfile[g3_sndfilecount].buffersize);
  if(g3_sndfile[g3_sndfilecount].buffer==NULL)
  {
    return(-1);
  }
  // store filename
  g3_fnameget(path,&g3_sndfile[g3_sndfilecount].name);
  // Set up wave format structure.
  memset(&pcmwf,0,sizeof(PCMWAVEFORMAT));
  pcmwf.wf.wFormatTag=WAVE_FORMAT_PCM;
  pcmwf.wf.nChannels=1;
  pcmwf.wf.nSamplesPerSec=G3_SND_FREQUENCY;
  pcmwf.wf.nBlockAlign=1;
  pcmwf.wf.nAvgBytesPerSec=pcmwf.wf.nSamplesPerSec*pcmwf.wf.nBlockAlign;
  pcmwf.wBitsPerSample=8;
  // Set up DSBUFFERDESC structure.
  memset(&dsbdesc,0,sizeof(DSBUFFERDESC)); // Zero it out.
  dsbdesc.dwSize=sizeof(DSBUFFERDESC);
  dsbdesc.dwFlags=DSBCAPS_CTRLDEFAULT; // Need default controls (pan, volume, frequency).
  dsbdesc.dwBufferBytes=g3_sndfile[g3_sndfilecount].buffersize; // 3 second buffer.
  dsbdesc.lpwfxFormat=(LPWAVEFORMATEX)&pcmwf;
  // Create buffer.
  hr=IDirectSound_CreateSoundBuffer(g3_ds,&dsbdesc,&g3_sndfile[g3_sndfilecount].dsbuffer,NULL);
  if(hr==DS_OK)
  {
    hr=IDirectSoundBuffer_Lock(g3_sndfile[g3_sndfilecount].dsbuffer,0,g3_sndfile[g3_sndfilecount].buffersize,&lpvPtr1,&dwBytes1,&lpvPtr2,&dwBytes2,0);
    if(hr==DSERR_BUFFERLOST)
    {
      IDirectSoundBuffer_Restore(g3_sndfile[g3_sndfilecount].dsbuffer);
      hr=IDirectSoundBuffer_Lock(g3_sndfile[g3_sndfilecount].dsbuffer,0,g3_sndfile[g3_sndfilecount].buffersize,&lpvPtr1,&dwBytes1,&lpvPtr2,&dwBytes2,0);
    }
  }
  if(hr!=DS_OK)
  {
    free(g3_sndfile[g3_sndfilecount].buffer);
    return(-1);
  }
  memcpy(lpvPtr1,g3_sndfile[g3_sndfilecount].buffer,dwBytes1);
  if(lpvPtr2!=NULL)
  {
    memcpy(lpvPtr2,g3_sndfile[g3_sndfilecount].buffer+dwBytes1,dwBytes2);
  }
  hr=IDirectSoundBuffer_Unlock(g3_sndfile[g3_sndfilecount].dsbuffer,lpvPtr1,dwBytes1,lpvPtr2,dwBytes2);
  if(hr!=DS_OK)
  {
    free(g3_sndfile[g3_sndfilecount].buffer);
    return(-1);
  }      
  g3_sndfilecount++;
  return(TRUE);
}

int _g3_snd_search(char *path)
{
  char fname[_MAX_FNAME];
  int index;
  index=0;
  g3_fnameget(path,fname);
  while(index<g3_sndfilecount && strcmp(fname,g3_sndfile[index].name))
  {
    index++;
  }
  if(index>=g3_sndfilecount) return(-1);
  return(index);
}

int _g3_snd_findfree()
{
  int index=0;
  int status;
  do
  {
    if(g3_snd[index]==NULL) goto found;
    IDirectSoundBuffer_GetStatus(g3_snd[index],&status);
    if(!(status & DSBSTATUS_PLAYING))
      {
        //release the free directsoundbuffer
        IDirectSoundBuffer_Release(g3_snd[index]);
        goto found;
      }
    index++;
  }
  while(index<G3_SNDMAX);
  return(-1);
found:
  //mark the current sound buffer free
  g3_snd[index]=NULL;
  return(index);
}

int g3_snd_play(char *name,int pan,int volume)
{
  int index;
  int id;
  HRESULT hr;
  id=-1;
  if(volume==0) return(id);
  index=_g3_snd_search(name);
  if(index<0)
  {
    index=g3_snd_add(name);
  }
  if(index>=0)
  {
    id=_g3_snd_findfree();
    if(id>=0)
    {   
    IDirectSound_DuplicateSoundBuffer(g3_ds,g3_sndfile[index].dsbuffer,&g3_snd[id]);
    hr=IDirectSoundBuffer_SetPan(g3_snd[id],(pan*1000)>>G3_FPSHIFT);
    hr=IDirectSoundBuffer_SetVolume(g3_snd[id],((volume-G3_FPVALUE(1.0))*1000)>>G3_FPSHIFT);
    hr=IDirectSoundBuffer_Play(g3_snd[id],NULL,NULL,NULL);
    }
  }
  return(id);
}

int g3_snd_stop()
{
  
}


